Package: PeerPerformance
Version: 2.4.0
Title: Luck-Corrected Peer Performance Analysis in R
Authors@R: c(person("David", "Ardia", role = c("aut", "cre", "cph", "fnd"),
             email = "david.ardia.ch@gmail.com",
             comment = c(ORCID = "0000-0003-2823-782X")),
             person("Nabil", "Bouamara", role = "ctb"),
             person("Kris", "Boudt", role = "aut",
             comment = c(ORCID = "0000-0002-1000-5142")),
             person("Sebastien", "Legros", role = "ctb"),
             person("Murilo Andre", "Peres Pereira", role = "ctb",
             comment = c(ORCID = "0009-0007-0086-4801")),
             person("Benjamin", "Seguin", role = "ctb"))
Depends: R (>= 3.5.0)
Imports: compiler, graphics, grDevices, lmtest, parallel, sandwich,
        stats
Suggests: knitr, rmarkdown, testthat
VignetteBuilder: knitr
Description: Provides functions to perform the peer performance analysis of
    funds' returns as described in Ardia and Boudt (2018)
    <doi:10.1016/j.jbankfin.2017.10.014>. For each fund, the package estimates
    the proportion of peers it outperforms, is equalled by, and is outperformed
    by, correcting for luck with the false discovery approach of Storey (2002)
    <doi:10.1111/1467-9868.00346>. Screenings can be based on factor-model
    alphas, Sharpe ratios, or modified Sharpe ratios, the latter using the
    equality test of Ardia and Boudt (2015) <doi:10.1016/j.frl.2015.02.008>.
    Funds can be screened within a universe or against a separate peer group,
    over rolling windows, and results come with bootstrap confidence intervals,
    summary, plot, and tidy data frame methods.
BugReports: https://github.com/ArdiaD/PeerPerformance/issues
URL: https://github.com/ArdiaD/PeerPerformance
License: GPL (>= 2)
LazyData: true
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-08-02 20:33:38 UTC; ardiad
Author: David Ardia [aut, cre, cph, fnd] (ORCID:
    <https://orcid.org/0000-0003-2823-782X>),
  Nabil Bouamara [ctb],
  Kris Boudt [aut] (ORCID: <https://orcid.org/0000-0002-1000-5142>),
  Sebastien Legros [ctb],
  Murilo Andre Peres Pereira [ctb] (ORCID:
    <https://orcid.org/0009-0007-0086-4801>),
  Benjamin Seguin [ctb]
Maintainer: David Ardia <david.ardia.ch@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-02 20:50:02 UTC

# Generated by roxygen2: do not edit by hand

export(alphaScreening)
export(alphaTesting)
export(msharpe)
export(msharpeScreening)
export(msharpeTesting)
export(sharpe)
export(sharpeScreening)
export(sharpeTesting)
import(compiler)
import(lmtest)
import(sandwich)
importFrom(compiler,cmpfun)
importFrom(parallel,clusterApplyLB)
importFrom(parallel,makeCluster)
importFrom(parallel,stopCluster)
importFrom(stats,qnorm)

## Set of various R functions

#@name processControl
#@title Control parameters processsing
processControl <- function(control) {
  if (!is.list(control)) {
    control <- list()
  }
  nam <- names(control)
  if (!("type" %in% nam) || is.null(control$type)) {
    control$type <- 1
  }
  if (!("ttype" %in% nam) || is.null(control$ttype)) {
    control$ttype <- 2
  }
  if (!("hac" %in% nam) || is.null(control$hac)) {
    control$hac <- FALSE
  }
  if (!("nBoot" %in% nam) || is.null(control$nBoot)) {
    control$nBoot <- 499
  }
  if (!("bBoot" %in% nam) || is.null(control$bBoot)) {
    control$bBoot <- 1
  }
  if (!("pBoot" %in% nam) || is.null(control$pBoot)) {
    control$pBoot <- 1
  }
  if (!("nCore" %in% nam) || is.null(control$nCore)) {
    control$nCore <- 1
  }
  if (!("minObs" %in% nam) || is.null(control$minObs)) {
    control$minObs <- 10
  }
  if (!("minObsPi" %in% nam) || is.null(control$minObsPi)) {
    control$minObsPi <- 1
  }
  if (!("lambda" %in% nam) || is.null(control$lambda)) {
    control$lambda <- NULL
  }
  if (!("screen_beta" %in% nam) || is.null(control$screen_beta)) {
    control$screen_beta <- FALSE
  }
  if (!("fastAdjust" %in% nam) || is.null(control$fastAdjust)) {
    control$fastAdjust <- FALSE
  }
  if (!("gammaPos" %in% nam) || is.null(control$gammaPos)) {
    control$gammaPos <- 0.4
  }
  if (!("gammaNeg" %in% nam) || is.null(control$gammaNeg)) {
    control$gammaNeg <- 0.6
  }

  # basic validation of the (possibly user-supplied) values
  for (nm in c("hac", "screen_beta", "fastAdjust")) {
    v <- control[[nm]]
    if (length(v) != 1L || !is.logical(v) || is.na(v))
      stop("'control$", nm, "' must be a single TRUE/FALSE")
  }
  for (nm in c("type", "ttype", "pBoot", "nBoot", "bBoot", "nCore",
               "minObs", "minObsPi", "gammaPos", "gammaNeg")) {
    v <- control[[nm]]
    if (length(v) != 1L || !is.numeric(v) || !is.finite(v))
      stop("'control$", nm, "' must be a single finite number")
  }
  # the count-like controls must be whole numbers (they feed sample.int(),
  # makeCluster(), and array indexing)
  for (nm in c("nBoot", "bBoot", "nCore", "minObs", "minObsPi")) {
    v <- control[[nm]]
    if (v != round(v))
      stop("'control$", nm, "' must be a whole number")
    control[[nm]] <- as.integer(round(v))
  }
  if (!(control$type %in% c(1, 2))) stop("'control$type' must be 1 or 2")
  if (!(control$ttype %in% c(1, 2))) stop("'control$ttype' must be 1 or 2")
  if (!(control$pBoot %in% c(1, 2))) stop("'control$pBoot' must be 1 or 2")
  if (!is.logical(control$hac)) stop("'control$hac' must be logical")
  if (control$nBoot < 1) stop("'control$nBoot' must be >= 1")
  if (control$bBoot < 0) stop("'control$bBoot' must be >= 0")
  if (control$nCore < 1) stop("'control$nCore' must be >= 1")
  if (control$minObs < 1) stop("'control$minObs' must be >= 1")
  if (control$minObsPi < 0) stop("'control$minObsPi' must be >= 0")
  if (!is.null(control$lambda) && (any(control$lambda < 0) || any(control$lambda >= 1)))
    stop("'control$lambda' must lie in [0, 1)")
  if (control$gammaPos <= 0 || control$gammaPos >= 1) stop("'control$gammaPos' must lie in (0, 1)")
  if (control$gammaNeg <= 0 || control$gammaNeg >= 1) stop("'control$gammaNeg' must lie in (0, 1)")

  return(control)
}

#@name .coefNames
#@title Row labels for the screen_beta coefficients (alpha + factor betas)
.coefNames <- function(factors) {
  nm <- colnames(factors)
  if (is.null(nm)) {
    nm <- paste0("beta", seq_len(ncol(as.matrix(factors))))
  }
  c("alpha", nm)
}

#@name alphaFactor
#@title Compute alpha factor
alphaFactor <- function(X, factors = NULL) {
  fit <- stats::lm(X ~ 1 + factors)
  alpha <- as.vector(fit$coef[1, ])
  return(alpha)
}

#' @name sharpe
#' @title Compute Sharpe ratio
#' @description Function which computes the Sharpe ratio.
#' @details The Sharpe ratio (Sharpe 1994) is one industry standard for measuring the
#' absolute risk adjusted performance of hedge funds.
#' @param X Vector (of length \eqn{T}) or matrix (of size \eqn{T \times
#' N}{TxN}) of returns for \eqn{N} funds. \code{NA} values are allowed.
#' @param na.rm A logical value indicating whether \code{NA} values should be
#' stripped before the computation. Default \code{na.rm = TRUE}
#' @return A scalar or a vector (of size \eqn{N}) with the Sharpe ratios.
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{sharpeTesting}}, \code{\link{sharpeScreening}} and
#' \code{\link{msharpe}}.
#' @references
#' Ardia, D., Boudt, K. (2015).
#' Testing equality of modified Sharpe ratios.
#' \emph{Finance Research Letters} \bold{13}, pp.97--104.
#' \doi{10.1016/j.frl.2015.02.008}
#'
#' Ardia, D., Boudt, K. (2016).
#' The Peer Ratios Performance of Hedge Funds.
#' \emph{Working paper}.
#' \doi{10.2139/ssrn.2000901}
#'
#' Sharpe, W.F. (1994).
#' The Sharpe ratio.
#' \emph{Journal of Portfolio Management} \bold{21}(1), pp.49--58.
#' \doi{10.3905/jpm.1994.409501}
#' @keywords htest
#' @examples
#' ## Load the data
#' data('hfdata')
#'
#' ## Compute the Sharpe ratio
#' out = sharpe(hfdata)
#' print(out)
#'
#' out = sharpe(hfdata, na.rm = FALSE)
#' print(out)
#' @export
sharpe <- function(X, na.rm = TRUE) {
  X <- as.matrix(X)
  N <- ncol(X)
  tmp <- .sharpe(X, na.rm)
  out <- tmp$mu.hat/tmp$sig.hat
  return(out)
}

#@name .sharpe
#@title Compute Sharpe ratio
.sharpe <- function(X, na.rm) {
  nObs <- colSums(is.finite(X))
  mu.hat <- colMeans(X, na.rm = na.rm)
  X_ <- sweep(x = X, MARGIN = 2, STATS = mu.hat, FUN = "-")
  sig.hat <- sqrt(colSums(X_^2, na.rm = na.rm)/(nObs - 1))
  out <- list(mu.hat = mu.hat, sig.hat = sig.hat)
  return(out)
}

#' @name msharpe
#' @title Compute modified Sharpe ratio
#' @description Function which computes the modified Sharpe ratio
#' @details The modified Sharpe ratio (Favre and Galeano 2002) is one industry
#' standard for measuring the absolute risk adjusted performance of hedge
#' funds.
#' @param X Vector (of length \eqn{T}) or matrix (of size \eqn{T \times
#' N}{TxN}) of returns.  \code{NA} values are allowed.
#' @param level Modified Value-at-Risk level. Default: \code{level = 0.90}.
#' @param na.rm A logical value indicating whether \code{NA} values should be
#' stripped before the computation. Default \code{na.rm = TRUE}.
#' @param na.neg A logical value indicating whether \code{NA} values should be
#' returned if a negative modified Value-at-Risk is obtained.  Default
#' \code{na.neg = TRUE}.
#' @return Scalar or a vector (of size \eqn{N}) with the modified Sharpe
#' ratios.
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{msharpeTesting}}, \code{\link{msharpeScreening}} and
#' \code{\link{sharpe}}.
#' @references
#' Ardia, D., Boudt, K. (2015).
#' Testing equality of modified Sharpe ratios.
#' \emph{Finance Research Letters} \bold{13}, 97--104.
#'
#' Ardia, D., Boudt, K. (2018).
#' The peer performance ratios of hedge Funds.
#' \emph{Journal of Banking and Finance} \bold{87}, 351--368.
#'
#' Favre, L., Galeano, J.A. (2002).
#' Mean-modified Value-at-Risk Optimization with Hedge Funds.
#' \emph{Journal of Alternative Investments} \bold{5}(2), 21--25.
#'
#' Gregoriou, G. N., Gueyie, J.-P. (2003).
#' Risk-adjusted performance of funds of hedge funds using a modified Sharpe ratio.
#' \emph{Journal of Wealth Management} \bold{6}(3), 77--83.
#' @keywords htest
#' @examples
#' ## Load the data (randomized data of monthly hedge fund returns)
#' data('hfdata')
#'
#' out = msharpe(hfdata)
#' print(out)
#'
#' out = msharpe(hfdata, na.rm = FALSE)
#' print(out)
#' @export
#' @importFrom stats qnorm
msharpe <- function(X, level = 0.9, na.rm = TRUE, na.neg = TRUE) {
  X <- as.matrix(X)
  N <- ncol(X)
  tmp <- .msharpe(X, level, na.rm, na.neg)
  out <- tmp$m1/tmp$mVaR
  return(out)
}

.msharpe <- function(X, level, na.rm, na.neg) {
  m1 <- colMeans(X, na.rm = na.rm)
  X_ <- sweep(x = X, MARGIN = 2, STATS = m1, FUN = "-")
  m2 <- colMeans(X_^2, na.rm = na.rm)
  m3 <- colMeans(X_^3, na.rm = na.rm)
  m4 <- colMeans(X_^4, na.rm = na.rm)
  za <- stats::qnorm(1 - level)
  skew <- m3/m2^(3/2)
  kurt <- (m4/m2^2) - 3
  mVaR <- -m1 + sqrt(m2) * (-za - (1/6) * (za^2 - 1) * skew - (1/24) *
                              (za^3 - 3 * za) * kurt + (1/36) * (2 * za^3 - 5 * za) * skew^2)
  if (na.neg) {
    mVaR[mVaR < 0] <- NA
  }
  out <- list(m1 = m1, mVaR = mVaR)
  return(out)
}

# #' @name .infoFund
# #' @import compiler
.infoFund <- function(X, factors = NULL, level = NULL, na.rm = TRUE, na.neg = TRUE, screen_beta=FALSE) {
  X <- as.matrix(X)
  N <- ncol(X)
  nObs <- colSums(is.finite(X))
  muX <- colMeans(X, na.rm = na.rm)
  rX <- sweep(x = X, MARGIN = 2, STATS = muX, FUN = "-")
  sigX <- sqrt(colSums(rX^2, na.rm = na.rm)/(nObs - 1))
  sharpe_ <- muX/sigX

  # if (is.null(factors)) {
  #   fit <- stats::lm(X ~ 1)
  # } else {
  #   fit <- stats::lm(X ~ 1 + factors)
  # }
  # alpha_ <- as.vector(fit$coef[1, ]) #output is a vector.

  # FIX
  # preallocate space.
  if(screen_beta){
    alpha_ <- matrix(NA, nrow=ncol(factors)+1, ncol=ncol(X))
  }else{
    alpha_ <- rep(NA, ncol(X))
  }

  if (is.null(factors)) {
    for (col in 1:ncol(X)) {
      fit <- stats::lm(X[,col] ~ 1)
      alpha_[col] <- fit$coef[1]
    }
  } else {
    for (col in 1:ncol(X)) {
      fit <- stats::lm(X[,col] ~ 1 + factors)
      if(screen_beta){
        alpha_[, col] <- fit$coef
      }else{
        alpha_[col] <- fit$coef[1]
      }

    }
  }

  msharpe_ <- NULL
  if (!is.null(level)) {
    msharpe_ <- msharpe(X, level = level, na.rm = na.rm, na.neg = na.neg)
  }

  out <- list(nObs = nObs, mu = muX, sig = sigX, sharpe = sharpe_, alpha = alpha_,
              msharpe = msharpe_)
  return(out)
}
infoFund <- compiler::cmpfun(.infoFund)

# #' @name .bootIndicesByLen
# Pre-generate bootstrap indices for every distinct pairwise sample length, in
# the master process. The screening functions pass the resulting named list to
# the PSOCK workers, which select the matrix matching their pair's
# complete-case length. This keeps all randomness in the master (so seeded
# results do not depend on 'nCore') and guarantees a valid circular-block
# structure at every length: previously a single full-length index matrix was
# remapped by modulo inside the workers, which biased the iid resampling and
# broke the block contiguity on unbalanced panels. Lengths shorter than the
# block length are skipped (the corresponding pairs are left untested). On a
# balanced panel there is a single length and the draws are identical to the
# previous behavior.
.bootIndicesByLen <- function(lens, nBoot, bBoot) {
  lens <- sort(unique(as.integer(lens)), decreasing = TRUE)
  lens <- lens[lens >= 1L]
  keep <- lens >= max(1L, bBoot)
  if (!any(keep)) {
    stop("the bootstrap block length 'bBoot' cannot exceed the number of concordant observations of every pair")
  }
  if (!all(keep)) {
    warning("some pairs have fewer concordant observations than the block length 'bBoot' and are left untested")
  }
  lens <- lens[keep]
  out <- lapply(lens, function(L) bootIndices(L, nBoot, bBoot))
  names(out) <- as.character(lens)
  return(out)
}
bootIndicesByLen <- compiler::cmpfun(.bootIndicesByLen)

# #' @name .bootIndices
# #' @import compiler
.bootIndices <- function(T, nBoot, bBoot) {
  if (bBoot > T) {
    stop("the bootstrap block length 'bBoot' cannot exceed the number of observations")
  }
  idsBoot <- matrix(data = NA, nrow = T, ncol = nBoot)
  if (bBoot == 1) {
    idsBoot <- matrix(sample.int(T, size = T * nBoot, replace = TRUE),
                      nrow = T, ncol = nBoot)
  } else {
    for (i in 1:nBoot) {
      l <- floor(T/bBoot)
      ids <- c(1:T, 1:bBoot)
      seqb <- vector("integer", T)
      start.points <- sample.int(T, size = l, replace = TRUE)
      for (j in (1:l)) {
        start <- start.points[j]
        seqb[((j - 1) * bBoot + 1):(j * bBoot)] <- ids[start:(start +
                                                                bBoot - 1)]
      }
      idsBoot[, i] <- seqb
    }
  }
  return(idsBoot)
}
bootIndices <- compiler::cmpfun(.bootIndices)

context("Testing") 

f.test.pval <- function(pval) {
  out <- pval >= 0 && pval <= 1
  return(out)
}

data('hfdata')
N = dim(hfdata)[2]

x = hfdata[,1]
y = hfdata[,2]

test_that("Sharpe and modified Sharpe", { 
  tmp = sharpe(hfdata)
  expect_true(is.vector(tmp))
  expect_true(length(tmp) == N)
  
  tmp = msharpe(hfdata)
  expect_true(is.vector(tmp))
  expect_true(length(tmp) == N)
}) 

test_that("Sharpe testing", { 
  tmp = sharpeTesting(x, y, control = list())
  expect_true(f.test.pval(tmp$pval))
  set.seed(123)
  tmp = sharpeTesting(x, y, control = list(type = 2))
  expect_true(f.test.pval(tmp$pval))
  tmp = sharpeTesting(x, y, control = list(ttype = 1))
  expect_true(f.test.pval(tmp$pval))
  tmp = sharpeTesting(x, y, control = list(hac = TRUE))
  expect_true(f.test.pval(tmp$pval))
  tmp = sharpeTesting(x, y, control = list(type  = 2, pBoot = 2))
  expect_true(f.test.pval(tmp$pval))
})

test_that("Modified Sharpe testing", { 
  tmp = msharpeTesting(x, y, control = list())
  expect_true(f.test.pval(tmp$pval))
  set.seed(123)
  tmp = msharpeTesting(x, y, control = list(type = 2))
  expect_true(f.test.pval(tmp$pval))
  tmp = msharpeTesting(x, y, control = list(ttype = 1))
  expect_true(f.test.pval(tmp$pval))
  tmp = msharpeTesting(x, y, control = list(hac = TRUE))
  expect_true(f.test.pval(tmp$pval))
  tmp = msharpeTesting(x, y, control = list(type  = 2, pBoot = 2))
  expect_true(f.test.pval(tmp$pval))
})

## Cross-group screening: each fund in group X against every fund in group Y.
## The single-fund-against-a-group case is simply nX = 1 (X a vector/one column).
## ----------------------------------------------------------------------------
## alpha
## ----------------------------------------------------------------------------

# #' @importFrom stats lm na.omit
# #' @importFrom lmtest coeftest
# #' @importFrom sandwich vcovHAC
.alphaScreeningXYi <- function(i, X, Y, factors, T, nY, hac, minObs,
                               screen_beta = FALSE) {
  if (screen_beta && !is.null(factors)) {
    ncoef <- 1L + ncol(factors)
    row_return <- 1:ncoef
  } else {
    ncoef <- 1L
    row_return <- 1L
  }
  pvali <- dalphai <- tstati <- matrix(NA, nrow = ncoef, ncol = nY)
  xi <- X[, i]
  for (j in 1:nY) {
    dxy <- xi - Y[, j]
    # usable rows: account for missing factors as lm(na.omit) would
    if (is.null(factors)) {
      ok <- !is.na(dxy) & !is.nan(dxy)
    } else {
      ok <- stats::complete.cases(dxy, factors)
    }
    if (sum(ok) < minObs) {
      next
    }
    if (all(abs(dxy[ok]) < sqrt(.Machine$double.eps))) {
      next  # identical series (e.g. a fund compared with itself)
    }
    if (is.null(factors)) {
      fit <- stats::lm(dxy ~ 1, na.action = stats::na.omit)
    } else {
      fit <- stats::lm(dxy ~ 1 + factors, na.action = stats::na.omit)
    }
    sfit <- summary(fit)
    if (!is.finite(sfit$sigma) || sfit$sigma < sqrt(.Machine$double.eps)) {
      next  # (near) deterministic differential: skip to avoid spurious significance
    }
    if (!hac) {
      sm <- sfit$coef
    } else {
      sm <- lmtest::coeftest(fit, vcov. = sandwich::vcovHAC(fit))
    }
    pvali[row_return, j]   <- sm[row_return, 4]
    dalphai[row_return, j] <- sm[row_return, 1]
    tstati[row_return, j]  <- sm[row_return, 3]
  }
  list(dalphai = dalphai, pvali = pvali, tstati = tstati)
}
alphaScreeningXYi <- compiler::cmpfun(.alphaScreeningXYi)

.alphaScreeningXY <- function(X, Y, factors = NULL, control = list(),
                              screen_beta = FALSE) {
  ctr <- processControl(control)
  X <- as.matrix(X)
  Y <- as.matrix(Y)
  T <- nrow(X)
  if (nrow(Y) != T) {
    stop("'X' and 'Y' must have the same number of rows (time periods)")
  }
  nX <- ncol(X)
  nY <- ncol(Y)
  if (nX < 1L || nY < 1L) {
    stop("cross-group screening needs at least one fund in 'X' and one peer in 'Y'")
  }

  if (screen_beta && !is.null(factors)) {
    ncoef <- 1L + ncol(factors)
  } else {
    ncoef <- 1L
  }
  row_return <- 1:ncoef
  pval <- dalpha <- tstat <- array(NA, dim = c(ncoef, nX, nY))

  if (ctr$nCore == 1) {
    # serial path: no PSOCK cluster (avoids the per-call cluster overhead);
    # the anonymous wrapper avoids the clash with lapply's own 'X' argument
    z <- lapply(1:nX, function(ii)
      alphaScreeningXYi(ii, X = X, Y = Y, factors = factors, T = T, nY = nY,
                        hac = ctr$hac, minObs = ctr$minObs,
                        screen_beta = screen_beta))
  } else {
    cl <- parallel::makeCluster(ctr$nCore)
    on.exit(parallel::stopCluster(cl), add = TRUE)
    z <- parallel::clusterApplyLB(cl = cl, x = as.list(1:nX),
                                  fun = alphaScreeningXYi, X = X, Y = Y,
                                  factors = factors, T = T, nY = nY, hac = ctr$hac,
                                  minObs = ctr$minObs, screen_beta = screen_beta)
  }

  for (i in 1:nX) {
    out <- z[[i]]
    dalpha[row_return, i, ] <- out$dalphai[row_return, ]
    pval[row_return, i, ]   <- out$pvali[row_return, ]
    tstat[row_return, i, ]  <- out$tstati[row_return, ]
  }

  pi <- computePi(pval = pval, dalpha = dalpha, tstat = tstat,
                  lambda = ctr$lambda, nBoot = ctr$nBoot,
                  bpos = ctr$gammaPos, bneg = ctr$gammaNeg,
                  fast = ctr$fastAdjust)

  info <- infoFund(X, factors = factors, screen_beta = screen_beta)

  if (!screen_beta) {
    pval   <- matrix(pval[1, , ],   nrow = nX, ncol = nY)
    dalpha <- matrix(dalpha[1, , ], nrow = nX, ncol = nY)
    tstat  <- matrix(tstat[1, , ],  nrow = nX, ncol = nY)
    npeer  <- rowSums(!is.na(pval))
  } else {
    npeer <- apply(!is.na(pval), c(1, 2), sum)
    cn <- .coefNames(factors)
    rownames(pi$pizero) <- rownames(pi$pipos) <- rownames(pi$pineg) <- cn
    rownames(pi$lambda) <- cn
    rownames(info$alpha) <- cn
    rownames(npeer) <- cn
  }

  out <- list(n = info$nObs, npeer = npeer, ny = nY, alpha = info$alpha,
              dalpha = dalpha, pval = pval, tstat = tstat, lambda = pi$lambda,
              pizero = pi$pizero, pipos = pi$pipos, pineg = pi$pineg,
              cross = TRUE)
  class(out) <- "SCREENING"
  return(out)
}
alphaScreeningXY <- compiler::cmpfun(.alphaScreeningXY)

## ----------------------------------------------------------------------------
## Sharpe
## ----------------------------------------------------------------------------

.sharpeScreeningXYi <- function(i, X, Y, T, nY, nBoot, bsids, minObs, type,
                                hac, b, ttype, pBoot) {
  pvali <- dsharpei <- tstati <- rep(NA, nY)
  xi <- X[, i]
  for (j in 1:nY) {
    dxy <- xi - Y[, j]
    ok <- !is.na(dxy) & !is.nan(dxy)
    if (sum(ok) < minObs) {
      next
    }
    if (all(abs(dxy[ok]) < sqrt(.Machine$double.eps))) {
      next
    }
    rets <- cbind(xi[ok], Y[ok, j])
    if (type == 1) {
      tmp <- sharpeTestAsymptotic(rets, hac, ttype)
    } else {
      # indices pre-generated in the master for this pair's length
      bs <- bsids[[as.character(sum(ok))]]
      if (is.null(bs)) {
        next  # block length exceeds this pair's sample size
      }
      tmp <- sharpeTestBootstrap(rets, bs, b, ttype, pBoot)
    }
    if (!is.finite(tmp$tstat)) {
      next  # degenerate pair (e.g. zero-variance series)
    }
    dsharpei[j] <- tmp$dsharpe
    pvali[j]    <- tmp$pval
    tstati[j]   <- tmp$tstat
  }
  list(dsharpei = dsharpei, pvali = pvali, tstati = tstati)
}
sharpeScreeningXYi <- compiler::cmpfun(.sharpeScreeningXYi)

.sharpeScreeningXY <- function(X, Y, control = list()) {
  ctr <- processControl(control)
  if (ctr$bBoot == 0) {
    stop("'bBoot = 0' (data-driven block length) is not supported in screening; ",
         "set an explicit block length 'bBoot >= 1'.")
  }
  X <- as.matrix(X)
  Y <- as.matrix(Y)
  T <- nrow(X)
  if (nrow(Y) != T) {
    stop("'X' and 'Y' must have the same number of rows (time periods)")
  }
  nX <- ncol(X)
  nY <- ncol(Y)
  if (nX < 1L || nY < 1L) {
    stop("cross-group screening needs at least one fund in 'X' and one peer in 'Y'")
  }

  # pre-generate the bootstrap indices in the master, one matrix per distinct
  # cross-pair complete-case length (workers select by length)
  # only needed for the bootstrap test, and only for pairs passing 'minObs',
  # so that the asymptotic path never depends on 'bBoot'
  pairLens <- crossprod(1 * (!is.na(X) & !is.nan(X)), 1 * (!is.na(Y) & !is.nan(Y)))
  bsids <- NULL
  if (ctr$type == 2) {
    bsids <- bootIndicesByLen(pairLens[pairLens >= ctr$minObs],
                              ctr$nBoot, ctr$bBoot)
  }
  pval <- dsharpe <- tstat <- matrix(NA, nX, nY)

  if (ctr$nCore == 1) {
    # serial path: no PSOCK cluster (avoids the per-call cluster overhead);
    # the anonymous wrapper avoids the clash with lapply's own 'X' argument
    z <- lapply(1:nX, function(ii)
      sharpeScreeningXYi(ii, X = X, Y = Y, T = T, nY = nY, nBoot = ctr$nBoot,
                         bsids = bsids, minObs = ctr$minObs, type = ctr$type,
                         hac = ctr$hac, b = ctr$bBoot, ttype = ctr$ttype,
                         pBoot = ctr$pBoot))
  } else {
    cl <- parallel::makeCluster(ctr$nCore)
    on.exit(parallel::stopCluster(cl), add = TRUE)
    z <- parallel::clusterApplyLB(cl = cl, x = as.list(1:nX),
                                  fun = sharpeScreeningXYi, X = X, Y = Y, T = T,
                                  nY = nY, nBoot = ctr$nBoot, bsids = bsids,
                                  minObs = ctr$minObs, type = ctr$type,
                                  hac = ctr$hac, b = ctr$bBoot, ttype = ctr$ttype,
                                  pBoot = ctr$pBoot)
  }

  for (i in 1:nX) {
    out <- z[[i]]
    dsharpe[i, ] <- out$dsharpei
    pval[i, ]    <- out$pvali
    tstat[i, ]   <- out$tstati
  }

  pi <- computePi(pval = pval, dalpha = dsharpe, tstat = tstat,
                  lambda = ctr$lambda, nBoot = ctr$nBoot,
                  bpos = ctr$gammaPos, bneg = ctr$gammaNeg,
                  fast = ctr$fastAdjust)
  info <- infoFund(X)

  out <- list(n = info$nObs, npeer = rowSums(!is.na(pval)), ny = nY,
              sharpe = info$sharpe, dsharpe = dsharpe, pval = pval,
              tstat = tstat, lambda = pi$lambda, pizero = pi$pizero,
              pipos = pi$pipos, pineg = pi$pineg, cross = TRUE)
  class(out) <- "SCREENING"
  return(out)
}
sharpeScreeningXY <- compiler::cmpfun(.sharpeScreeningXY)

## ----------------------------------------------------------------------------
## modified Sharpe
## ----------------------------------------------------------------------------

.msharpeScreeningXYi <- function(i, X, Y, level, T, nY, nBoot, bsids, minObs,
                                 na.neg, type, hac, b, ttype, pBoot) {
  pvali <- dmsharpei <- tstati <- rep(NA, nY)
  xi <- X[, i]
  for (j in 1:nY) {
    dxy <- xi - Y[, j]
    ok <- !is.na(dxy) & !is.nan(dxy)
    if (sum(ok) < minObs) {
      next
    }
    if (all(abs(dxy[ok]) < sqrt(.Machine$double.eps))) {
      next
    }
    rets <- cbind(xi[ok], Y[ok, j])
    if (type == 1) {
      tmp <- msharpeTestAsymptotic(rets, level, na.neg, hac, ttype)
    } else {
      # indices pre-generated in the master for this pair's length
      bs <- bsids[[as.character(sum(ok))]]
      if (is.null(bs)) {
        next  # block length exceeds this pair's sample size
      }
      tmp <- msharpeTestBootstrap(rets, level, na.neg, bs, b, ttype, pBoot)
    }
    if (!is.finite(tmp$tstat)) {
      next  # degenerate pair or NA modified VaR (na.neg)
    }
    dmsharpei[j] <- tmp$dmsharpe
    pvali[j]     <- tmp$pval
    tstati[j]    <- tmp$tstat
  }
  list(dmsharpei = dmsharpei, pvali = pvali, tstati = tstati)
}
msharpeScreeningXYi <- compiler::cmpfun(.msharpeScreeningXYi)

.msharpeScreeningXY <- function(X, Y, level = 0.9, na.neg = TRUE, control = list()) {
  ctr <- processControl(control)
  if (ctr$bBoot == 0) {
    stop("'bBoot = 0' (data-driven block length) is not supported in screening; ",
         "set an explicit block length 'bBoot >= 1'.")
  }
  X <- as.matrix(X)
  Y <- as.matrix(Y)
  T <- nrow(X)
  if (nrow(Y) != T) {
    stop("'X' and 'Y' must have the same number of rows (time periods)")
  }
  nX <- ncol(X)
  nY <- ncol(Y)
  if (nX < 1L || nY < 1L) {
    stop("cross-group screening needs at least one fund in 'X' and one peer in 'Y'")
  }

  # pre-generate the bootstrap indices in the master, one matrix per distinct
  # cross-pair complete-case length (workers select by length)
  # only needed for the bootstrap test, and only for pairs passing 'minObs',
  # so that the asymptotic path never depends on 'bBoot'
  pairLens <- crossprod(1 * (!is.na(X) & !is.nan(X)), 1 * (!is.na(Y) & !is.nan(Y)))
  bsids <- NULL
  if (ctr$type == 2) {
    bsids <- bootIndicesByLen(pairLens[pairLens >= ctr$minObs],
                              ctr$nBoot, ctr$bBoot)
  }
  pval <- dmsharpe <- tstat <- matrix(NA, nX, nY)

  if (ctr$nCore == 1) {
    # serial path: no PSOCK cluster (avoids the per-call cluster overhead);
    # the anonymous wrapper avoids the clash with lapply's own 'X' argument
    z <- lapply(1:nX, function(ii)
      msharpeScreeningXYi(ii, X = X, Y = Y, level = level, T = T, nY = nY,
                          nBoot = ctr$nBoot, bsids = bsids,
                          minObs = ctr$minObs, na.neg = na.neg,
                          type = ctr$type, hac = ctr$hac, b = ctr$bBoot,
                          ttype = ctr$ttype, pBoot = ctr$pBoot))
  } else {
    cl <- parallel::makeCluster(ctr$nCore)
    on.exit(parallel::stopCluster(cl), add = TRUE)
    z <- parallel::clusterApplyLB(cl = cl, x = as.list(1:nX),
                                  fun = msharpeScreeningXYi, X = X, Y = Y,
                                  level = level, T = T, nY = nY, nBoot = ctr$nBoot,
                                  bsids = bsids, minObs = ctr$minObs,
                                  na.neg = na.neg, type = ctr$type, hac = ctr$hac,
                                  b = ctr$bBoot, ttype = ctr$ttype, pBoot = ctr$pBoot)
  }

  for (i in 1:nX) {
    out <- z[[i]]
    dmsharpe[i, ] <- out$dmsharpei
    pval[i, ]     <- out$pvali
    tstat[i, ]    <- out$tstati
  }

  pi <- computePi(pval = pval, dalpha = dmsharpe, tstat = tstat,
                  lambda = ctr$lambda, nBoot = ctr$nBoot,
                  bpos = ctr$gammaPos, bneg = ctr$gammaNeg,
                  fast = ctr$fastAdjust)
  info <- infoFund(X, level = level, na.neg = na.neg)

  out <- list(n = info$nObs, npeer = rowSums(!is.na(pval)), ny = nY,
              msharpe = info$msharpe, dmsharpe = dmsharpe, pval = pval,
              tstat = tstat, lambda = pi$lambda, pizero = pi$pizero,
              pipos = pi$pipos, pineg = pi$pineg, cross = TRUE)
  class(out) <- "SCREENING"
  return(out)
}
msharpeScreeningXY <- compiler::cmpfun(.msharpeScreeningXY)

## Set of R functions for Sharpe ratio testing

#@name .alphaTesting
#@description See alphaTesting
# .alphaTesting <- function(x, y, factors=NULL, control=list()){
# 	X <- cbind(x, y)
# 	screen <- alphaScreening(X, factors, control)
# 	out <- list(n = screen$n[[1]], alpha = screen$alpha, dalpha = screen$dalpha[1,2],
# 				tstat = screen$tstat[1,2], pval = screen$pval[1,2])
# 	class(out) <- "TESTING"
# 	return(out)
# }

.alphaTesting <- function(x, y, factors=NULL, control=list(), screen_beta=NULL){
	# process control
	ctr <- processControl(control)
	hac <- ctr$hac
	# screen_beta: argument takes precedence, else fall back to control
	if (is.null(screen_beta)) {
		screen_beta <- ctr$screen_beta
	}
	# screen_beta only makes sense with factors
	if (isTRUE(screen_beta) && is.null(factors)) {
		warning("'screen_beta = TRUE' requires 'factors'; it is ignored.")
		screen_beta <- FALSE
	}
	X <- as.matrix(x)
	Y <- as.matrix(y)
	dXY <- X - Y
	# usable observations = rows where the return differential and (when
	# supplied) the factors are all observed
	if (is.null(factors)) {
		ok <- stats::complete.cases(dXY)
	} else {
		ok <- stats::complete.cases(dXY, factors)
	}
	nObs <- sum(ok)
	if (nObs < ctr$minObs) {
		stop("fewer than 'control$minObs' usable observations to compare the two funds")
	}
	N <- 2


	if (screen_beta & !is.null(factors)) {
		n_coef <- ncol(factors) + 1
		row_return <- 1:n_coef
	} else {
		row_return <- 1
	}

	if (is.null(factors)) {
		fit <- stats::lm(dXY ~ 1, na.action = stats::na.omit)
		fitX <- stats::lm(X ~ 1, na.action = stats::na.omit)
		fitY <- stats::lm(Y ~ 1, na.action = stats::na.omit)
	} else {
		beta <- factors
		fit <- stats::lm(dXY ~ 1 + beta, na.action = stats::na.omit)
		fitX <- stats::lm(X ~ 1 + beta, na.action = stats::na.omit)
		fitY <- stats::lm(Y ~ 1 + beta, na.action = stats::na.omit)
	} # end of factors/no factors

	# HAC within loop.
	if (!hac) {
		sumfit <- summary(fit)
		sumfitX <- summary(fitX)
		sumfitY <- summary(fitY)
		alpha <- cbind(sumfitX$coef[row_return,1], sumfitY$coef[row_return,1])
		pval <- sumfit$coef[row_return, 4]
		dalpha <- sumfit$coef[row_return, 1]
		tstat <- sumfit$coef[row_return, 3]
	} else {
		sumfit <- lmtest::coeftest(fit, vcov. = sandwich::vcovHAC(fit))
		sumfitX <- lmtest::coeftest(fitX, vcov. = sandwich::vcovHAC(fitX))
		sumfitY <- lmtest::coeftest(fitY, vcov. = sandwich::vcovHAC(fitY))
		alpha <- cbind(sumfitX[row_return,1], sumfitY[row_return,1])
		pval <- sumfit[row_return, 4]
		dalpha <- sumfit[row_return, 1]
		tstat <- sumfit[row_return, 3]
	}

	out <- list(n = nObs,
				alpha = alpha,
				dalpha = dalpha,
				tstat = tstat,
				pval = pval,
				screen_beta = screen_beta)
	class(out) <- "TESTING"
	return(out)
}

#' @name alphaTesting
#' @title Testing the difference of alpha outperformance ratios
#' @description Function which performs the testing of the difference of alpha outperformance ratios.
#' @details The alpha measure (Treynor and Black 1973, Carhart 1997, Fung and Hsieh
#' 2004) is one industry standard for measuring the absolute risk adjusted
#' performance of hedge funds. This function performs
#' the testing of alpha outperformance ratio difference for two funds.
#'
#' For the testing, only the intersection of non-\code{NA} observations for the
#' two funds are used.
#'
##' The argument \code{control} is a list that can supply any of the following
#' components:
#' \itemize{
#' \item \code{'hac'} Heteroscedastic-autocorrelation consistent
#' standard errors. Default: \code{hac = FALSE}.
#' }
#' @param x Vector (of length \eqn{T}) of returns for the first fund. \code{NA}
#' values are allowed.
#' @param y Vector (of length \eqn{T}) returns for the second fund. \code{NA}
#' values are allowed.
#' @param factors Matrix \eqn{(T \times K)}{(TxK)} of \eqn{T} returns for the
#' \eqn{K} factors. \code{NA} values are allowed.
#' @param screen_beta Boolean to screen all factors' coefficients (beta).
#' Default: \code{screen_beta = NULL}, in which case the value is taken from
#' \code{control$screen_beta} (itself defaulting to \code{FALSE}). When supplied
#' directly, the argument takes precedence. \code{screen_beta = TRUE} requires
#' \code{factors}.
#' @param control Control parameters (see *Details*).
#' @return A list with the following components:\cr
#'
#' \code{n}: Number of non-\code{NA} concordant observations.\cr
#'
#' \code{alpha}: Vector (of length 2) of the two funds' unconditional alphas.\cr
#'
#' \code{dalpha}: Difference of the two funds' alphas.\cr
#'
#' \code{tstat}: t-stat of the alpha difference.\cr
#'
#' \code{pval}: p-value of the test of equal alpha.\cr
#'
#' When \code{screen_beta = TRUE}, \code{alpha} is a matrix (one row per
#' coefficient, the first being the alpha) and \code{dalpha}, \code{tstat} and
#' \code{pval} are vectors with one entry per coefficient; the \code{print}
#' method reports the alpha (first) coefficient only.
#' @note Further details on the methodology with an application to the hedge
#' fund industry is given in Ardia and Boudt (2018).
#'
#' Some internal functions where adapted from Michael Wolf MATLAB code.
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{alphaScreening}}.
#' @references
#' Ardia, D., Boudt, K. (2015).
#' Testing equality of modified Sharpe ratios.
#' \emph{Finance Research Letters} \bold{13}, 97--104.
#'
#' Ardia, D., Boudt, K. (2018).
#' The peer performance ratios of hedge funds.
#' \emph{Journal of Banking and Finance} \bold{87}, 351--368.
#'
#' Barras, L., Scaillet, O., Wermers, R. (2010).
#' False discoveries in mutual fund performance: Measuring luck in estimated alphas.
#' \emph{Journal of Finance} \bold{65}(1), 179--216.
#'
#' Sharpe, W.F. (1994).
#' The Sharpe ratio.
#' \emph{Journal of Portfolio Management} \bold{21}(1), 49--58.
#'
#' Ledoit, O., Wolf, M. (2008).
#' Robust performance hypothesis testing with the Sharpe ratio.
#' \emph{Journal of Empirical Finance} \bold{15}(5), 850--859.
#'
#' Storey, J. (2002).
#' A direct approach to false discovery rates.
#' \emph{Journal of the Royal Statistical Society B} \bold{64}(3), 479--498.
#' @keywords htest
#' @examples
#' ## Load the data (randomized data of monthly hedge fund returns)
#' data("hfdata")
#' x = hfdata[,1]
#' y = hfdata[,2]
#'
#' ## Run alpha testing
#' alphaTesting(x, y)

#' @export
#' @importFrom compiler cmpfun
alphaTesting <- compiler::cmpfun(.alphaTesting)

## ---------------------------------------------------------------------------
## Monte-Carlo validation of the PeerPerformance estimators
##
## Reproducible checks that the implementation behaves as the methodology of
## Ardia and Boudt (2018, JBF) and Ardia and Boudt (2015, FRL) predicts:
##
##   (A) Under a null of equal performance, a naive percentile rank spuriously
##       labels ~half of each fund's peers as out-performed (pure luck). The
##       luck correction collapses this: the estimated out-performance ratio
##       pi+ is small and most of the mass is (correctly) assigned to the
##       equal-performance ratio pi0. This is the correction the package
##       delivers.
##
##   (B) Under a planted skill structure, the estimator is deliberately
##       conservative: only statistically detectable differences are flagged,
##       so undetected true differences are pooled into pi0 and the estimated
##       pi+/pi- are lower bounds on their population values. It errs toward
##       "equal performance" rather than toward false discoveries.
##
##   (C) The modified Sharpe ratio equality test has approximately correct
##       size under the null, and power against a genuine difference.
##
## Run with:
##   source(system.file("scripts", "validation.R", package = "PeerPerformance"))
## Increase the replication counts below for publication-quality figures; the
## defaults are chosen to run in a couple of minutes on a laptop.
## ---------------------------------------------------------------------------

library("PeerPerformance")
set.seed(1234)

ctr <- list(nCore = 1)          # single core for reproducibility

## ===========================================================================
## (A) Equal performance: pi0 should be ~1, naive rank ~0.5 outperformance
## ===========================================================================
N  <- 30L                       # funds
TT <- 120L                      # months
RA <- 100L                      # replications

pi0A <- posA <- rankA <- numeric(RA)
for (r in seq_len(RA)) {
  X  <- matrix(stats::rnorm(TT * N, mean = 0.005, sd = 0.04), TT, N)
  sc <- alphaScreening(X, control = ctr)
  pi0A[r]  <- mean(sc$pizero, na.rm = TRUE)
  posA[r]  <- mean(sc$pipos,  na.rm = TRUE)
  ## naive percentile rank: fraction of peers each fund beats on mean return,
  ## averaged over funds (ignores statistical significance)
  mu <- colMeans(X)
  rankA[r] <- mean(sapply(seq_len(N), function(i) mean(mu[i] > mu[-i])))
}

cat("\n(A) Equal-performance null (true pi0 = 1, true out-performance = 0)\n")
cat(sprintf("    naive rank out-performance   : %.3f  <- spurious 'skill' (pure luck)\n",
            mean(rankA)))
cat(sprintf("    luck-corrected pi+           : %.3f  <- collapses toward 0\n",
            mean(posA)))
cat(sprintf("    luck-corrected pi0           : %.3f  (sd %.3f) <- mass restored to 'equal'\n",
            mean(pi0A), stats::sd(pi0A)))

## ===========================================================================
## (B) Planted skill: one third skilled, one third neutral, one third poor
## ===========================================================================
RB <- 100L
mu_true <- c(rep(0.012, N %/% 3), rep(0.005, N %/% 3),
             rep(-0.002, N - 2L * (N %/% 3)))
## true population ratios (per fund), from the planted means (ties -> pi0)
truePos <- mean(sapply(seq_len(N), function(i) mean(mu_true[i] > mu_true[-i])))
trueNeg <- mean(sapply(seq_len(N), function(i) mean(mu_true[i] < mu_true[-i])))

posB <- negB <- numeric(RB)
for (r in seq_len(RB)) {
  X  <- sapply(mu_true, function(m) stats::rnorm(TT, mean = m, sd = 0.04))
  sc <- alphaScreening(X, control = ctr)
  posB[r] <- mean(sc$pipos, na.rm = TRUE)
  negB[r] <- mean(sc$pineg, na.rm = TRUE)
}

cat("\n(B) Planted skill structure (universe-average ratios; estimates are\n",
    "    conservative lower bounds -- undetected differences fall into pi0)\n", sep = "")
cat(sprintf("    out-performance pi+ : estimated %.3f  vs true %.3f\n",
            mean(posB), truePos))
cat(sprintf("    under-performance pi-: estimated %.3f  vs true %.3f\n",
            mean(negB), trueNeg))

## ===========================================================================
## (C) Modified Sharpe equality test: size (null) and power (alternative)
## ===========================================================================
RC    <- 500L
alpha <- 0.05

## size: two funds drawn from the SAME distribution -> H0 true
rej_size <- 0L
for (r in seq_len(RC)) {
  a <- stats::rnorm(TT, 0.006, 0.04)
  b <- stats::rnorm(TT, 0.006, 0.04)
  rej_size <- rej_size + (msharpeTesting(a, b, level = 0.90)$pval < alpha)
}

## power: clearly different risk/return -> H0 false
rej_pow <- 0L
for (r in seq_len(RC)) {
  a <- stats::rnorm(TT, 0.010, 0.030)
  b <- stats::rnorm(TT, 0.002, 0.055)
  rej_pow <- rej_pow + (msharpeTesting(a, b, level = 0.90)$pval < alpha)
}

cat("\n(C) Modified Sharpe equality test (asymptotic, nominal 5%)\n")
cat(sprintf("    empirical size (H0 true) : %.3f  (target %.3f)\n",
            rej_size / RC, alpha))
cat(sprintf("    empirical power (H0 false): %.3f\n", rej_pow / RC))
cat("\nNote: the asymptotic test can be mildly oversized in short samples;\n",
    "use control = list(type = 2) for the studentized bootstrap.\n", sep = "")

## ===========================================================================
## Compact summary with Monte-Carlo standard errors (for reporting)
## ===========================================================================
binse <- function(p, R) sqrt(p * (1 - p) / R)   # binomial s.e. for a rate
cat("\n--- summary (estimate (MC s.e.)) ----------------------------------\n")
cat(sprintf("(A) naive-rank pi+        : %.3f (%.3f)\n", mean(rankA), stats::sd(rankA)/sqrt(RA)))
cat(sprintf("(A) luck-corrected pi+    : %.3f (%.3f)   [true 0]\n",     mean(posA), stats::sd(posA)/sqrt(RA)))
cat(sprintf("(A) luck-corrected pi0    : %.3f (%.3f)   [true 1]\n",     mean(pi0A), stats::sd(pi0A)/sqrt(RA)))
cat(sprintf("(B) pi+ (planted skill)   : %.3f (%.3f)   [true %.3f]\n",  mean(posB), stats::sd(posB)/sqrt(RB), truePos))
cat(sprintf("(B) pi- (planted skill)   : %.3f (%.3f)   [true %.3f]\n",  mean(negB), stats::sd(negB)/sqrt(RB), trueNeg))
cat(sprintf("(C) mSR test size         : %.3f (%.3f)   [target 0.05]\n", rej_size/RC, binse(rej_size/RC, RC)))
cat(sprintf("(C) mSR test power        : %.3f (%.3f)\n",                 rej_pow/RC,  binse(rej_pow/RC, RC)))
cat(sprintf("settings: N=%d, T=%d, reps A/B=%d, reps C=%d, seed=1234\n", N, TT, RA, RC))

## Set of R functions used to compute pi and lambda

# #' @name .computePi
# #' @title Compute pi0, pi+ and pi-
# #' @importFrom stats qnorm
# #' @import compiler
.computePi <- function(pval, dalpha, tstat, lambda = 0.5, nBoot = 499,
                       bpos = 0.4, bneg = 0.6, adjust = TRUE, fast = FALSE) {

  if (!is.matrix(pval) & !is.array(pval)) {
    pval <- matrix(pval, nrow = 1)
  }
  if (!is.matrix(dalpha) & !is.array(dalpha)) {
    dalpha <- matrix(dalpha, nrow = 1)
  }
  if (!is.matrix(tstat) & !is.array(tstat)) {
    tstat <- matrix(tstat, nrow = 1)
  }

  if(length(dim(pval))==2){
    m <- nrow(pval)
    factors_dim <- 1
    dim(pval) <- c(1, dim(pval))
    dim(dalpha) <- c(1, dim(dalpha))
    dim(tstat) <- c(1, dim(tstat))
  }else{
    m <- dim(pval)[2]
    factors_dim <- dim(pval)[1]
    }

  # n1 = ncol(pval) # n + 1 funds
  if (!is.null(lambda) && !(length(lambda) %in% c(1L, m))) {
    stop("'lambda' must be NULL, length 1, or length equal to the number of funds")
  }
  if (length(lambda) == 1) {
    lambda <- rep(lambda, m)
  }

  pizero <- pipos <- pineg <- lambda_ <- matrix(rep(NA, m*factors_dim), nrow=factors_dim)

  for(factor_dim in 1:factors_dim){
    for (i in 1:m) {
      pvali <- pval[factor_dim, i, ]
      dalphai <- dalpha[factor_dim, i, ]
      tstati <- tstat[factor_dim, i, ]
      if (all(is.na(pvali))) {
        next
      }
      if (is.null(lambda)) {
        lambdai <- computeOptLambda(pval = pvali, nBoot = nBoot,
                                    adjust = adjust, fast = fast)
      } else {
        lambdai <- lambda[i]
      }

      pizeroi <- computePizero(pvali, lambda = lambdai, adjust = adjust,
                               fast = fast)
      idxOK <- !is.na(pvali) & !is.na(dalphai) & !is.na(tstati)
      n <- sum(idxOK)  # number of peers
      if (n <= 1) {
        next
      }

      ni0 <- pizeroi * n
      hn <- round(0.5 * n)
      piposi <- pinegi <- 0

      # idxOKnPizeroi = idxOK & pvali >= lambdai we need to fix this
      qpos <- stats::qnorm(p = bpos)
      qneg <- stats::qnorm(p = bneg)

      if (sum(dalphai[idxOK] >= 0) >= hn) {
        piposi <- (1/n) * min((n - ni0), max(sum(tstati[idxOK] >= qpos) -
                                               ni0 * (1 - bpos), 0))
        pinegi <- min(max(1 - pizeroi - piposi, 0), 1)  # numerical stability
      } else {
        pinegi <- (1/n) * min((n - ni0), max(sum(tstati[idxOK] <= qneg) -
                                               ni0 * bneg, 0))
        piposi <- min(max(1 - pizeroi - pinegi, 0), 1)  # numerical stability
      }

      pizero[factor_dim, i] <- pizeroi
      pipos[factor_dim, i] <- piposi
      pineg[factor_dim, i] <- pinegi
      lambda_[factor_dim, i] <- lambdai
    }
  }

  if(factors_dim == 1){
  	pizero <- pizero[1, ]
  	pipos <- pipos[1, ]
  	pineg <- pineg[1, ]
  	lambda_ <- lambda_[1, ]
  }
  out <- list(pizero = pizero, pipos = pipos, pineg = pineg, lambda = lambda_)
  return(out)
}
computePi <- compiler::cmpfun(.computePi)

# #' @name .computePizero
# #' @import compiler
.computePizero <- function(pval, lambda = 0.5, adjust = TRUE, fast = FALSE) {
  if (!is.matrix(pval)) {
    pval <- matrix(pval, nrow = 1)
  }

  # Note: 'n' (number of trials in the truncated-binomial adjustment of adjustPi)
  # is the number of columns of 'pval'. This matches the original Ardia & Boudt
  # (2018) implementation; when some peers are NA it slightly over-counts the
  # effective number of peers. Kept as-is for consistency with published results.
  n <- ncol(pval)
  pizero <- apply(pval>=lambda, 1, mean, na.rm=TRUE)
  # pizero <- mean(pval >= lambda, na.rm = TRUE)
  pizero <- pizero * (1/(1 - lambda))
  pizero[pizero > 1] <- 1
  # adjust pi using truncated binomial
  if (adjust) {
    pizero <- adjustPi(pizero, n = n, lambda = lambda, fast = fast)
  }

  return(pizero)
}
computePizero <- compiler::cmpfun(.computePizero)

# #' @name .adjustPi
# #' @title Adjust estimated pi0 using quadratic fit
# #' @importFrom stats dnorm pnorm uniroot
# #' @import compiler
.adjustPi <- function(pi.hat, n = 100, lambda = 0.5, fast = FALSE) {

  # Fast path (control$fastAdjust): 'n' and 'lambda' are constant within a
  # call, so the same monotone map is inverted at every element. Instead of one
  # uniroot() per element, invert the whole vector at once by bisection. Forty
  # halvings of [1e-5, 1.5] locate the root to about 1e-12. Note that this is
  # the accuracy of the *inversion*, not the agreement with the default path:
  # uniroot() stops at its own tolerance (.Machine$double.eps^0.25, about
  # 1.2e-4), so the two paths typically differ by a few 1e-5, with the fast
  # path the more accurate of the two.
  if (isTRUE(fast)) {
    fwd <- function(pi0) {
      nlambda <- pi0 * n * (1 - lambda)
      out <- pi0
      i <- nlambda < n                       # else branch of asym.hatpi0
      if (any(i)) {
        s <- sqrt(nlambda[i] * (n - nlambda[i])/(n^3 * (1 - lambda)^2))
        zcrit <- (1 - pi0[i])/s
        out[i] <- pi0[i] + s * (-stats::dnorm(zcrit) +
                                  (1 - stats::pnorm(zcrit)) * zcrit)
      }
      out
    }
    m   <- length(pi.hat)
    out <- pi.hat        # fallback: keep the input, as the try-error path does
    if (m > 0L) {
      lo <- rep(1e-05, m)
      hi <- rep(1.5, m)
      # uniroot() fails when the target is NA/NaN or is not bracketed, and the
      # default path then keeps pi.hat; invert only the finite, bracketed
      # elements (which() also drops the NAs from the subscript)
      ok <- which(is.finite(pi.hat) & pi.hat >= fwd(lo) & pi.hat <= fwd(hi))
      if (length(ok) > 0L) {
        l <- lo[ok]; h <- hi[ok]; tgt <- pi.hat[ok]
        for (it in seq_len(40L)) {
          mid  <- (l + h)/2
          left <- fwd(mid) < tgt
          l[left]  <- mid[left]
          h[!left] <- mid[!left]
        }
        out[ok] <- (l + h)/2
      }
      big <- which(out > 1); if (length(big) > 0L) out[big] <- 1
      sml <- which(out < 0); if (length(sml) > 0L) out[sml] <- 0
    }
    return(out)
  }

  asym.hatpi0 <- function(pi0) {
    npi0 <- pi0 * n
    nlambda <- npi0 * (1 - lambda)
    if (nlambda >= n) {
      hatpi0 <- pi0
    } else {
      s2 <- nlambda * (n - nlambda)/(n^3 * (1 - lambda)^2)
      s <- sqrt(s2)
      zcrit <- (1 - pi0)/s
      hatpi0 <- pi0 + s * (-stats::dnorm(zcrit) + (1 - stats::pnorm(zcrit)) *
                             zcrit)
    }
    return(hatpi0)
  }

  asym.inverse <- function(f, lower = -100, upper = 100) {
    FUN <- function(y) stats::uniroot((function(x) f(x) - y), lower = lower,
                                      upper = upper)$root
    return(FUN)
  }

  asym.invpi0 <- asym.inverse(asym.hatpi0, lower = 1e-05, upper = 1.5)

  m <- length(pi.hat)
  out <- vector("double", m)
  for (i in 1:m) {
    tmp <- pi.hat[i]
    test.it <- try({
      tmp <- asym.invpi0(pi.hat[i])
    }, silent = TRUE)
    if (inherits(test.it, "try-error")) {
    	tmp <- pi.hat[i]
    }
    tmp[tmp > 1] <- 1
    tmp[tmp < 0] <- 0
    out[i] <- tmp
  }
  return(out)
}
adjustPi <- compiler::cmpfun(.adjustPi)

# #' @name .computeOptLambda
# #' @title Compute optimal lamba values
# #' @importFrom stats runif
# #' @import compiler
.computeOptLambda <- function(pval, nBoot = 499, adjust = TRUE, fast = FALSE) {
  if (!is.matrix(pval)) {
    pval <- matrix(pval, nrow = 1)
  }

  n <- nrow(pval)  # number of funds

  # Grid for the data-driven lambda. Ardia & Boudt (2018) describe a finer grid
  # {0.30, 0.32, ..., 0.70}; the coarser 0.1 step is kept here for consistency
  # with the published package results (and is much faster).
  vlambda <- seq(0.3, 0.7, 0.1)
  nvlambda <- length(vlambda)

  mpizero <- matrix(data = NA, nrow = n, ncol = nvlambda)
  for (i in 1:nvlambda) {
    mpizero[, i] <- computePizero(pval, lambda = vlambda[i], adjust = adjust,
                                  fast = fast)
  }
  # pi0hat
  vminpizero <- apply(mpizero, 1, "min")

  idx <- !is.nan(pval) & !is.na(pval)
  nObs <- rowSums(idx)

  bsunif <- matrix(stats::runif(max(nObs) * nBoot), max(nObs), nBoot)
  optlambda <- rep(0.5, n)

  for (i in 1:n) {
    nObsi <- nObs[i]
    if (nObsi > 0) {
      bsidx <- ceiling(bsunif[1:nObsi, 1:nBoot] * nObsi)
      pvalb <- stats::na.omit(pval[i, ])
      pvalb <- matrix(pvalb[bsidx], nrow = nBoot)

      mpizerob <- matrix(data = NA, nrow = nBoot, ncol = nvlambda)
      for (j in 1:nvlambda) {

        mpizerob[, j] <- computePizero(pvalb, lambda = vlambda[j],
                                       adjust = adjust, fast = fast)
      }
      vMSE <- colSums((mpizerob - vminpizero[i])^2)

      optlambda[i] <- vlambda[which.min(vMSE)]
    }
  }

  return(optlambda)
}
computeOptLambda <- compiler::cmpfun(.computeOptLambda)

context("Numerical regression")

data("hfdata")

test_that("alphaScreening peer ratios are stable and sum to one", {
  rets <- hfdata[, 15:20]
  res  <- alphaScreening(rets, control = list(nCore = 1))
  ## pinned values (verified against the article example)
  expect_equal(unname(round(res$pipos[1], 3)), 0.6)
  expect_equal(unname(round(res$pineg[1], 3)), 0.4)
  expect_equal(unname(res$npeer), rep(5, 6))
  ## the triple sums to one for every fund
  s <- res$pizero + res$pipos + res$pineg
  expect_true(all(abs(s - 1) < 1e-8))
})

test_that("gammaPos / gammaNeg are honoured and only affect the +/- split", {
  rets <- hfdata[, 1:30]
  ## fix lambda so that pizero is deterministic (the data-driven lambda uses an
  ## unseeded bootstrap); pizero must then be invariant to the gamma thresholds
  a <- alphaScreening(rets, control = list(nCore = 1, lambda = 0.5,
                                           gammaPos = 0.2, gammaNeg = 0.8))
  b <- alphaScreening(rets, control = list(nCore = 1, lambda = 0.5,
                                           gammaPos = 0.6, gammaNeg = 0.4))
  ## pizero is invariant to the gamma thresholds
  expect_equal(a$pizero, b$pizero, tolerance = 1e-10)
  ## the out/under split does respond
  expect_false(isTRUE(all.equal(a$pipos, b$pipos)))
  ## ratios remain valid
  for (res in list(a, b)) {
    expect_true(all(res$pipos  >= 0 & res$pipos  <= 1, na.rm = TRUE))
    expect_true(all(res$pineg  >= 0 & res$pineg  <= 1, na.rm = TRUE))
    expect_true(all(res$pizero >= 0 & res$pizero <= 1, na.rm = TRUE))
  }
})

test_that("alphaTesting returns a coherent htest-like list", {
  x <- hfdata[, 1]; y <- hfdata[, 2]
  out <- alphaTesting(x, y)
  expect_true(out$pval >= 0 && out$pval <= 1)
  expect_equal(length(out$alpha), 2)
  ## dalpha equals the difference of the two fund alphas
  expect_equal(unname(out$dalpha), unname(out$alpha[1] - out$alpha[2]), tolerance = 1e-8)
})

test_that("alphaScreening with screen_beta returns one row per coefficient", {
  set.seed(1)
  rets <- hfdata[, 1:5]
  fmat <- matrix(rnorm(nrow(rets)), ncol = 1)        # single factor
  res  <- alphaScreening(rets, factors = fmat, control = list(nCore = 1),
                         screen_beta = TRUE)
  ## first dimension = intercept (alpha) + 1 factor = 2
  expect_equal(nrow(res$pizero), 2L)
  expect_equal(ncol(res$pizero), 5L)
})

test_that("NA values are handled without error", {
  rets <- hfdata[, 1:6]
  rets[1:5, 1] <- NA
  res <- alphaScreening(rets, control = list(nCore = 1))
  expect_equal(length(res$pizero), 6L)
  expect_true(all(is.finite(res$npeer)))
})

test_that("audit-v2: output naming and control validation", {
  X <- hfdata[, 1:10]; colnames(X) <- paste0("HF", 1:10)
  ## as.data.frame keeps fund names; targetPeerPerformance labels its outputs
  tp <- targetPeerPerformance(X, funds = c("HF3", "HF7"), method = "sharpe",
                              control = list(nCore = 1))
  expect_equal(as.data.frame(tp)$fund, c("HF3", "HF7"))
  ## rollScreening: screen_beta requires factors (alpha) -> attr FALSE without
  suppressWarnings(
    rb <- rollScreening(hfdata[, 1:12], screen = "alpha", width = 40, by = 20,
                        control = list(nCore = 1, screen_beta = TRUE)))
  expect_false(isTRUE(attr(rb, "screen_beta")))
  ## input / control validation
  expect_error(targetPeerPerformance(X, funds = 1.9, control = list(nCore = 1)))
  expect_error(alphaScreening(X[, 1:4], control = list(nCore = 1, hac = c(TRUE, FALSE))))
  expect_error(alphaScreening(X[, 1:4], control = list(nCore = 1, lambda = c(0.4, 0.5))))
})

test_that("sharpe/msharpe screening on an unbalanced panel has no NaN (PR #14)", {
  set.seed(1)
  Tn <- 80
  X <- matrix(rnorm(Tn * 4, 0.01, 0.05), Tn, 4)
  X[1:25, 2] <- NA   # the FIRST peer of fund 1 is missing early
  ss <- sharpeScreening(X, control = list(nCore = 1))
  ms <- msharpeScreening(X, control = list(nCore = 1))
  ## with the old 'X[idx[, k], 1]' indexing the focal returns were NA-contaminated
  ## for some pairs, producing NaN p-values
  expect_false(any(is.nan(ss$pval)))
  expect_false(any(is.nan(ms$pval)))
  expect_true(all(abs(ss$pizero + ss$pipos + ss$pineg - 1) < 1e-8, na.rm = TRUE))
})

test_that("audit fixes: degenerate inputs are handled", {
  ## A1: screen_beta = TRUE without factors warns and falls back (no crash)
  expect_warning(r <- alphaScreening(hfdata[, 1:5], control = list(nCore = 1, screen_beta = TRUE)))
  expect_false(is.matrix(r$pizero))

  ## A2: bBoot = 0 in screening errors cleanly
  expect_error(sharpeScreening(hfdata[, 1:4], control = list(nCore = 1, type = 2, bBoot = 0)))
  expect_error(msharpeScreening(hfdata[, 1:4], control = list(nCore = 1, type = 2, bBoot = 0)))

  ## A3: a peer differing from the focal fund by a constant is excluded
  Y <- hfdata[, 11:14]; Y[, 2] <- hfdata[, 1] + 0.5
  s <- alphaScreening(hfdata[, 1], Y = Y, control = list(nCore = 1))
  expect_equal(s$npeer, 3L)
  expect_false(any(s$pval == 0, na.rm = TRUE))

  ## A4: rollScreening only stamps screen_beta for the alpha screen
  rb <- rollScreening(hfdata[, 1:20], screen = "sharpe", width = 40, by = 20,
                      control = list(nCore = 1, screen_beta = TRUE))
  expect_false(isTRUE(attr(rb, "screen_beta")))

  ## B5: invalid control values are rejected
  expect_error(alphaScreening(hfdata[, 1:4], control = list(nCore = 1, type = 3)))
  expect_error(alphaScreening(hfdata[, 1:4], control = list(nCore = 1, gammaPos = 1.5)))
})

test_that("cross-group screening (X vs Y) works and excludes self", {
  ## single focal fund against a peer group
  s <- alphaScreening(hfdata[, 1], Y = hfdata[, 11:30], control = list(nCore = 1))
  expect_equal(length(s$pizero), 1L)
  expect_equal(s$npeer, 20L)
  expect_true(abs(s$pizero + s$pipos + s$pineg - 1) < 1e-8)

  ## group X against group Y: ratios over the nY peers
  g <- alphaScreening(hfdata[, 1:5], Y = hfdata[, 11:30], control = list(nCore = 1))
  expect_equal(dim(g$pval), c(5L, 20L))
  expect_equal(g$ny, 20L)
  expect_true(all(abs(g$pizero + g$pipos + g$pineg - 1) < 1e-8))

  ## self-exclusion when X is a subset of Y
  s2 <- alphaScreening(hfdata[, 1], Y = hfdata[, 1:30], control = list(nCore = 1))
  expect_equal(s2$npeer, 29L)

  ## Sharpe and modified Sharpe cross-group
  sh <- sharpeScreening(hfdata[, 1:3], Y = hfdata[, 11:30], control = list(nCore = 1))
  expect_equal(dim(sh$pval), c(3L, 20L))
  ms <- msharpeScreening(hfdata[, 1:3], Y = hfdata[, 11:30], level = 0.95,
                         control = list(nCore = 1))
  expect_equal(dim(ms$pval), c(3L, 20L))
})

test_that("as.data.frame and exposureHeterogeneity work", {
  set.seed(1)
  w <- alphaScreening(hfdata[, 1:8], control = list(nCore = 1))
  df <- as.data.frame(w)
  expect_equal(nrow(df), 8L)
  expect_true(all(c("fund", "alpha", "pipos", "pizero", "pineg", "npeer") %in% names(df)))

  set.seed(1)
  fac <- matrix(rnorm(nrow(hfdata) * 2), ncol = 2,
                dimnames = list(NULL, c("MKT", "SMB")))
  ## lambda fixed: the assertions are on shapes and labels, not on lambda
  scb <- alphaScreening(hfdata[, 1:15], factors = fac, screen_beta = TRUE,
                        control = list(nCore = 1, lambda = 0.5))
  expect_equal(rownames(scb$pizero), c("alpha", "MKT", "SMB"))
  dfb <- as.data.frame(scb)
  expect_equal(nrow(dfb), 15L * 3L)
  expect_true("coefficient" %in% names(dfb))

  eh <- exposureHeterogeneity(scb)
  expect_s3_class(eh, "exposureHeterogeneity")
  expect_equal(nrow(eh), 3L)
  expect_equal(eh$heterogeneity, 1 - eh$equalExposure, tolerance = 1e-12)
  expect_true(all(eh$heterogeneity >= 0 & eh$heterogeneity <= 1))
})

test_that("screen_beta can be set via control and the argument overrides it", {
  set.seed(1)
  fac <- matrix(rnorm(nrow(hfdata) * 2), ncol = 2,
                dimnames = list(NULL, c("MKT", "SMB")))
  a <- alphaScreening(hfdata[, 1:8], factors = fac,
                      control = list(nCore = 1, screen_beta = TRUE))
  expect_true(is.matrix(a$pizero))
  expect_equal(rownames(a$pizero), c("alpha", "MKT", "SMB"))
  ## explicit argument wins over the control value
  b <- alphaScreening(hfdata[, 1:8], factors = fac, screen_beta = FALSE,
                      control = list(nCore = 1, screen_beta = TRUE))
  expect_false(is.matrix(b$pizero))
})

test_that("rollScreening returns a tidy time series", {
  ## this test checks the shape of the output, not the value of lambda, so the
  ## threshold is fixed: the data-driven selection dominates the run time and
  ## would be re-run for every window
  set.seed(1234)
  roll <- rollScreening(hfdata[, 1:15], screen = "alpha", width = 36, by = 8,
                        control = list(nCore = 1, lambda = 0.5))
  expect_s3_class(roll, "rollScreening")
  expect_true(all(c("window", "index", "pizero", "pipos", "pineg",
                    "heterogeneity") %in% names(roll)))
  expect_equal(roll$heterogeneity, 1 - roll$pizero, tolerance = 1e-12)
  expect_equal(nrow(roll), length(seq.int(1, nrow(hfdata) - 36 + 1, by = 8)))

  ## screen_beta -> one row per window/coefficient
  set.seed(1)
  fac <- matrix(rnorm(nrow(hfdata) * 2), ncol = 2,
                dimnames = list(NULL, c("MKT", "SMB")))
  rb <- rollScreening(hfdata[, 1:15], factors = fac, width = 36, by = 12,
                      control = list(nCore = 1, screen_beta = TRUE,
                                     lambda = 0.5))
  ## set comparison (locale-independent: avoids C vs UTF-8 sort order)
  expect_setequal(unique(rb$coefficient), c("alpha", "MKT", "SMB"))
  pf <- tempfile(fileext = ".pdf"); pdf(pf); plot(rb); dev.off(); unlink(pf)
})

test_that("plot works on a single focal fund (cross-group)", {
  s <- alphaScreening(hfdata[, 1], Y = hfdata[, 11:30], control = list(nCore = 1))
  pf <- tempfile(fileext = ".pdf"); pdf(pf)
  out <- plot(s)
  dev.off(); unlink(pf)
  expect_equal(length(out), 3L)
})

test_that("targetPeerPerformance equals screening with Y = X", {
  rets <- hfdata[, 1:10]
  focals <- c(2, 5, 7)
  tp <- targetPeerPerformance(rets, funds = focals, method = "alpha",
                              control = list(nCore = 1, lambda = 0.5))
  yx <- alphaScreening(rets[, focals], Y = rets,
                       control = list(nCore = 1, lambda = 0.5))
  expect_equal(unname(tp$pizero), unname(yx$pizero), tolerance = 1e-10)
  expect_equal(unname(tp$pipos),  unname(yx$pipos),  tolerance = 1e-10)
  expect_equal(unname(tp$npeer),  rep(9L, 3))
  expect_s3_class(tp, "SCREENING")
  ## selection by name works and rows are labelled
  colnames(rets) <- paste0("F", 1:10)
  tp2 <- targetPeerPerformance(rets, funds = c("F2", "F5"), method = "sharpe",
                               control = list(nCore = 1))
  expect_equal(rownames(tp2$pval), c("F2", "F5"))
})

test_that("summary.SCREENING produces a ranked table", {
  set.seed(1)
  sc <- alphaScreening(hfdata[, 1:12], control = list(nCore = 1))
  s <- summary(sc)
  expect_s3_class(s, "summary.SCREENING")
  expect_equal(nrow(s$table), 12L)
  expect_true(all(c("estimate", "pipos", "pineg", "wins", "losses") %in% names(s$table)))
  expect_identical(print(s), s)
})

test_that("print and plot methods dispatch and return invisibly", {
  set.seed(1234)
  sc <- alphaScreening(hfdata[, 1:12], control = list(nCore = 1))
  expect_s3_class(sc, "SCREENING")
  expect_identical(print(sc), sc)

  tt <- msharpeTesting(hfdata[, 1], hfdata[, 2], level = 0.95)
  expect_s3_class(tt, "TESTING")
  expect_identical(print(tt), tt)

  ## plot to a throwaway device; returns the plotted ratio matrix invisibly
  pf <- tempfile(fileext = ".pdf"); pdf(pf)
  out <- plot(sc, nblock = 6)
  dev.off(); unlink(pf)
  expect_equal(ncol(out), 3L)
  expect_true(all(abs(rowSums(out) - 1) < 1e-8))
})

test_that("seeded bootstrap tests are reproducible", {
  x <- hfdata[, 1]; y <- hfdata[, 2]
  set.seed(321); p1 <- sharpeTesting(x, y, control = list(type = 2, nBoot = 200))$pval
  set.seed(321); p2 <- sharpeTesting(x, y, control = list(type = 2, nBoot = 200))$pval
  expect_identical(p1, p2)

  set.seed(321); q1 <- msharpeTesting(x, y, control = list(type = 2, nBoot = 200))$pval
  set.seed(321); q2 <- msharpeTesting(x, y, control = list(type = 2, nBoot = 200))$pval
  expect_identical(q1, q2)
})

test_that("confint.SCREENING brackets the point estimate and is valid", {
  rets <- hfdata[, 1:12]
  sc   <- alphaScreening(rets, control = list(nCore = 1))
  set.seed(42)
  ci <- confint(sc, parm = "pipos", nBoot = 200)
  ## shape and naming
  expect_equal(nrow(ci), ncol(rets))
  expect_equal(ncol(ci), 2L)
  expect_equal(rownames(ci), colnames(rets))
  est <- attr(ci, "estimate")
  ok  <- !is.na(ci[, 1]) & !is.na(ci[, 2])
  ## bounds are ordered, inside [0, 1], and contain the point estimate
  expect_true(all(ci[ok, 1] <= ci[ok, 2]))
  expect_true(all(ci[ok, ] >= 0 & ci[ok, ] <= 1))
  expect_true(all(ci[ok, 1] <= est[ok] + 1e-8 & est[ok] <= ci[ok, 2] + 1e-8))
  ## all three ratios are supported: valid, ordered bounds that bracket the estimate
  for (p in c("pizero", "pineg")) {
    set.seed(42)
    cp  <- confint(sc, parm = p, nBoot = 200)
    ep  <- attr(cp, "estimate")
    okp <- !is.na(cp[, 1]) & !is.na(cp[, 2])
    expect_equal(dim(cp), c(ncol(rets), 2L))
    expect_true(all(cp[okp, 1] <= cp[okp, 2]))
    expect_true(all(cp[okp, ] >= 0 & cp[okp, ] <= 1))
    expect_true(all(cp[okp, 1] <= ep[okp] + 1e-8 & ep[okp] <= cp[okp, 2] + 1e-8))
  }
  ## screen_beta screenings are rejected
  scb <- alphaScreening(rets, factors = hfdata[, 50, drop = FALSE],
                        control = list(nCore = 1, screen_beta = TRUE))
  expect_error(confint(scb), "screen_beta")
})

test_that("alphaTesting screen_beta returns a (K+1) x 2 alpha matrix (HAC and not)", {
  x <- hfdata[, 1]; y <- hfdata[, 2]
  fac <- hfdata[, 50:51]                       # K = 2 factors
  for (use_hac in c(FALSE, TRUE)) {
    res <- alphaTesting(x, y, factors = fac,
                        control = list(hac = use_hac), screen_beta = TRUE)
    expect_equal(dim(res$alpha), c(ncol(fac) + 1L, 2L))   # rows = coef, cols = x/y
    expect_length(res$dalpha, ncol(fac) + 1L)
    ## print shows the two funds' *alphas* (row 1), not a beta
    expect_output(print(res), "Peer performance test")
  }
})

test_that("control validation rejects non-whole and out-of-range values", {
  rets <- hfdata[, 1:4]
  expect_error(alphaScreening(rets, control = list(nCore = 1, nBoot = 2.5)), "whole number")
  expect_error(alphaScreening(rets, control = list(nCore = 1, minObs = -1)))
  ## the block length is only relevant to the bootstrap test (type = 2)
  expect_error(sharpeScreening(rets, control = list(nCore = 1, type = 2,
                                                    bBoot = nrow(rets) + 1)),
               "cannot exceed")
  ## degenerate within-group input
  expect_error(alphaScreening(hfdata[, 1], control = list(nCore = 1)), "at least two funds")
})

test_that("bootstrap screening is valid and seeded-reproducible on unbalanced panels", {
  ## unbalanced panel: distinct complete-case lengths across pairs
  X <- hfdata[, 1:6]
  X[1:10, 2]  <- NA
  X[1:20, 3]  <- NA
  X[41:60, 4] <- NA
  ctr <- list(nCore = 1, type = 2, bBoot = 3, nBoot = 99)

  set.seed(99); s1 <- sharpeScreening(X, control = ctr)
  set.seed(99); s2 <- sharpeScreening(X, control = ctr)
  expect_identical(s1$pval, s2$pval)                    # all RNG in the master
  ok <- !is.na(s1$pval)
  expect_true(any(ok))
  expect_true(all(s1$pval[ok] >= 0 & s1$pval[ok] <= 1))
  s <- s1$pizero + s1$pipos + s1$pineg
  expect_true(all(abs(s[!is.na(s)] - 1) < 1e-8))

  ## cross-group path with the same panel
  set.seed(99); x1 <- sharpeScreening(X[, 1:2], Y = X[, 3:6], control = ctr)
  set.seed(99); x2 <- sharpeScreening(X[, 1:2], Y = X[, 3:6], control = ctr)
  expect_identical(x1$pval, x2$pval)
  okx <- !is.na(x1$pval)
  expect_true(any(okx))
  expect_true(all(x1$pval[okx] >= 0 & x1$pval[okx] <= 1))

  ## a block length larger than a short pair's sample leaves that pair NA
  ## (fund 7 overlaps fund 8 on only 5 observations < bBoot = 8)
  Z <- hfdata[, 7:9]
  Z[1:55, 2] <- NA
  set.seed(7)
  expect_warning(
    sz <- sharpeScreening(Z, control = list(nCore = 1, type = 2, bBoot = 8,
                                            nBoot = 99, minObs = 3)),
    "left untested")
  expect_true(is.na(sz$pval[1, 2]))                     # untestable pair
  expect_false(is.na(sz$pval[1, 3]))                    # full-length pair still tested
})

test_that("serial (nCore = 1) and cluster (nCore = 2) paths give identical results", {
  skip_on_cran()   # keep CRAN runs light; 2 cores are exercised locally/CI
  rets <- hfdata[, 1:8]
  ## fix lambda so the comparison does not depend on the data-driven bootstrap
  ctr1 <- list(nCore = 1, lambda = 0.5)
  ctr2 <- list(nCore = 2, lambda = 0.5)
  a1 <- alphaScreening(rets, control = ctr1)
  a2 <- alphaScreening(rets, control = ctr2)
  expect_equal(a1$pval,  a2$pval,  tolerance = 1e-12)
  expect_equal(a1$pipos, a2$pipos, tolerance = 1e-12)
  ## bootstrapped Sharpe screening: indices are drawn in the master, so the
  ## result must not depend on the number of workers
  set.seed(3); b1 <- sharpeScreening(rets, control = c(ctr1, type = 2, bBoot = 2, nBoot = 99))
  set.seed(3); b2 <- sharpeScreening(rets, control = c(ctr2, type = 2, bBoot = 2, nBoot = 99))
  expect_equal(b1$pval, b2$pval, tolerance = 1e-12)
})

test_that("control$fastAdjust matches the default path and is faster", {
  ## adjustPi: the fast vectorised inversion agrees with the uniroot path
  ## across the (n, lambda) grid, well inside uniroot's own default tolerance
  set.seed(1)
  ph <- runif(200, 0.3, 1.0)
  for (nn in c(5, 9, 29, 99)) {
    for (lam in c(0.3, 0.4, 0.5, 0.6, 0.7)) {
      a <- PeerPerformance:::adjustPi(ph, n = nn, lambda = lam, fast = FALSE)
      b <- PeerPerformance:::adjustPi(ph, n = nn, lambda = lam, fast = TRUE)
      expect_equal(a, b, tolerance = 1e-3)       # uniroot tol is ~1.2e-4
      expect_true(all(b >= 0 & b <= 1))
    }
  }
  ## the flag must not change semantics, only speed: NA/NaN inputs and targets
  ## outside the uniroot bracket must behave exactly as the default path
  edge <- c(0.5, NA, NaN, 0.9, -0.5, 0, 1.4, 3.0)
  e0 <- PeerPerformance:::adjustPi(edge, n = 99, lambda = 0.5, fast = FALSE)
  e1 <- PeerPerformance:::adjustPi(edge, n = 99, lambda = 0.5, fast = TRUE)
  expect_equal(is.na(e0), is.na(e1))                    # same NA pattern
  expect_equal(e0[!is.na(e0)], e1[!is.na(e1)], tolerance = 1e-3)
  ## NaN can reach adjustPi through an all-NA p-value row; must not error
  pv <- matrix(c(NA, NA, NA, 0.2, 0.6, 0.9), nrow = 2, byrow = TRUE)
  expect_silent(PeerPerformance:::computePizero(pv, lambda = 0.5,
                                                adjust = TRUE, fast = TRUE))

  ## end to end: a screening with a fixed lambda is unchanged by the flag
  rets <- hfdata[, 1:12]
  s0 <- alphaScreening(rets, control = list(nCore = 1, lambda = 0.5))
  s1 <- alphaScreening(rets, control = list(nCore = 1, lambda = 0.5,
                                            fastAdjust = TRUE))
  expect_equal(s0$pizero, s1$pizero, tolerance = 1e-3)
  expect_equal(s0$pipos,  s1$pipos,  tolerance = 1e-3)
  ## the flag is validated like the other logical controls
  expect_error(alphaScreening(rets, control = list(nCore = 1,
                                                   fastAdjust = c(TRUE, FALSE))))

  ## round-trip: the fast path really solves the inversion (this is the strong
  ## check; agreement with the default is capped by uniroot's own tolerance)
  fwd <- function(pi0, n, lambda) {
    nlambda <- pi0 * n * (1 - lambda)
    out <- pi0
    i <- nlambda < n
    s <- sqrt(nlambda[i] * (n - nlambda[i])/(n^3 * (1 - lambda)^2))
    z <- (1 - pi0[i])/s
    out[i] <- pi0[i] + s * (-dnorm(z) + (1 - pnorm(z)) * z)
    out
  }
  for (nn in c(9, 99)) {
    for (lam in c(0.3, 0.5, 0.7)) {
      r <- PeerPerformance:::adjustPi(ph, n = nn, lambda = lam, fast = TRUE)
      ## only entries that were actually inverted: the target must lie inside
      ## the bracket [fwd(1e-5), fwd(1.5)] (otherwise uniroot fails and both
      ## paths fall back to the input), and the result must be unclamped
      brack <- ph >= fwd(1e-05, nn, lam) & ph <= fwd(1.5, nn, lam)
      inner <- brack & r > 0 & r < 1
      expect_true(any(inner))
      expect_equal(fwd(r[inner], nn, lam), ph[inner], tolerance = 1e-9)
    }
  }
})

test_that("asymptotic screening does not depend on the bootstrap block length", {
  ## bBoot is irrelevant when type = 1; an unbalanced panel with short pairs
  ## must not warn or error just because bBoot exceeds some pair length
  X <- hfdata[, 1:6]
  X[1:52, 2] <- NA                            # one pair with only 8 obs
  expect_silent(s1 <- sharpeScreening(X, control = list(nCore = 1, type = 1,
                                                        bBoot = 20, minObs = 5)))
  expect_silent(m1 <- msharpeScreening(X, control = list(nCore = 1, type = 1,
                                                         bBoot = 20, minObs = 5)))
  expect_true(any(!is.na(s1$pval)))
  expect_true(any(!is.na(m1$pval)))
})

test_that("plot methods accept the graphical arguments documented in '...'", {
  pf <- tempfile(fileext = ".pdf"); pdf(pf); on.exit({dev.off(); unlink(pf)})
  ## lambda fixed: only the plotting behaviour is under test here
  ctr <- list(nCore = 1, lambda = 0.5)
  sc  <- alphaScreening(hfdata[, 1:8], control = ctr)
  scb <- alphaScreening(hfdata[, 1:8], factors = hfdata[, 50:51],
                        control = c(ctr, screen_beta = TRUE))
  eh  <- exposureHeterogeneity(scb)
  rl  <- rollScreening(hfdata[, 1:12], width = 40, by = 20, control = ctr)
  one <- alphaScreening(hfdata[, 1], Y = hfdata[, 11:20], control = ctr)
  ## these all used to fail with "matched by multiple actual arguments"
  expect_silent(plot(sc,  main = "screening"))
  expect_silent(plot(one, main = "single fund"))
  expect_silent(plot(eh,  main = "heterogeneity", ylab = "h"))
  expect_silent(plot(rl,  xlab = "month", ylab = "ratio"))
  ## and the defaults still apply when nothing is passed
  expect_silent(plot(sc)); expect_silent(plot(eh)); expect_silent(plot(rl))
})

## Rolling-window peer performance screening (time series of ratios).

## internal: cross-sectional aggregation of a SCREENING result
.rollAgg <- function(sc) {
  pz <- sc$pizero; pp <- sc$pipos; pn <- sc$pineg
  if (is.matrix(pz)) {
    cn <- rownames(pz)
    if (is.null(cn)) cn <- paste0("coef", seq_len(nrow(pz)))
    data.frame(coefficient = cn,
               pizero = as.vector(rowMeans(pz, na.rm = TRUE)),
               pipos  = as.vector(rowMeans(pp, na.rm = TRUE)),
               pineg  = as.vector(rowMeans(pn, na.rm = TRUE)),
               stringsAsFactors = FALSE)
  } else {
    data.frame(coefficient = NA_character_,
               pizero = mean(pz, na.rm = TRUE),
               pipos  = mean(pp, na.rm = TRUE),
               pineg  = mean(pn, na.rm = TRUE),
               stringsAsFactors = FALSE)
  }
}

#' @name rollScreening
#' @title Rolling-window peer performance screening
#' @description Runs a peer performance screening on a rolling time window and
#' returns the time series of cross-sectionally averaged peer performance
#' ratios. This is the design underlying the dynamic analyses of Ardia et al.
#' (2022, 2023): on each window the screening is computed and the ratios are
#' averaged across funds (per coefficient when \code{screen_beta = TRUE}),
#' yielding a series of (factor exposure) heterogeneity measures.
#' @param X Matrix \eqn{(T \times N)}{(TxN)} of \eqn{T} returns for the \eqn{N}
#' funds.
#' @param factors Matrix \eqn{(T \times K)}{(TxK)} of factor returns (for
#' \code{screen = "alpha"}). Default: \code{NULL}.
#' @param Y Optional matrix \eqn{(T \times M)}{(TxM)} of returns for a peer group
#' (cross-group screening, see \code{\link{alphaScreening}}). Default:
#' \code{NULL}.
#' @param screen Performance measure to screen on: \code{"alpha"} (default),
#' \code{"sharpe"}, or \code{"msharpe"}.
#' @param width Window length (number of observations). Default: \code{36}.
#' @param by Step (in observations) between successive windows. Default:
#' \code{1}.
#' @param level Modified Value-at-Risk level (for \code{screen = "msharpe"}).
#' Default: \code{0.9}.
#' @param na.neg A logical value passed to \code{\link{msharpeScreening}}
#' (for \code{screen = "msharpe"}) indicating whether a negative modified
#' Value-at-Risk yields \code{NA}. Default: \code{TRUE}.
#' @param control Control parameters passed to the screening function (see
#' \code{\link{alphaScreening}}); set \code{control = list(screen_beta = TRUE)}
#' to obtain per-factor heterogeneity series. With \code{nCore > 1} each
#' window spins up its own PSOCK cluster, which is wasteful for many small
#' windows; the default \code{nCore = 1} runs serially without any cluster.
#' @param dates Optional vector of length \eqn{T} (e.g. \code{Date}) used to
#' label the windows by their end date. Default: \code{NULL}.
#' @return A \code{data.frame} of class \code{rollScreening} with one row per
#' window (or per window/coefficient when \code{screen_beta = TRUE}), with
#' columns \code{window}, \code{index} (end-of-window row), optionally
#' \code{date}, \code{coefficient}, the averaged \code{pizero}/\code{pipos}/
#' \code{pineg}, and \code{heterogeneity} (\eqn{1-\bar\pi^0}).
#' @author David Ardia and Kris Boudt.
#' @references
#' Ardia, D., Bluteau, K., Tran, D. (2022).
#' How easy is it for investment managers to deploy their talent in green and brown stocks?
#' \emph{Finance Research Letters} \bold{48}, 102992.
#'
#' Ardia, D., Bluteau, K., Lortie-Cloutier, G., Tran, D. (2023).
#' Factor exposure heterogeneity in green and brown stocks.
#' \emph{Finance Research Letters} \bold{55}, Part A, 103900.
#' @seealso \code{\link{alphaScreening}}, \code{\link{exposureHeterogeneity}}.
#' @keywords htest
#' @examples
#' \donttest{
#' data("hfdata")
#' set.seed(1234)
#' ## a subset keeps the example quick; drop the column index to screen the
#' ## whole universe
#' roll <- rollScreening(hfdata[, 1:20], screen = "alpha", width = 36, by = 6,
#'                       control = list(nCore = 1))
#' plot(roll)
#' }
#' @export
rollScreening <- function(X, factors = NULL, Y = NULL,
                          screen = c("alpha", "sharpe", "msharpe"),
                          width = 36L, by = 1L, level = 0.9, na.neg = TRUE,
                          control = list(), dates = NULL) {
  screen <- match.arg(screen)
  X <- as.matrix(X)
  T <- nrow(X)
  if (width < 2L || width > T) {
    stop("'width' must be between 2 and nrow(X)")
  }
  if (length(by) != 1L || !is.finite(by) || by < 1L) {
    stop("'by' must be a single positive integer")
  }
  by <- as.integer(by)
  if (!is.null(dates) && length(dates) != T) {
    stop("'dates' must have length nrow(X)")
  }
  ctr <- processControl(control)
  # screen_beta only applies to the alpha screening with factors
  screen_beta <- isTRUE(ctr$screen_beta) && screen == "alpha" && !is.null(factors)
  Fmat <- if (!is.null(factors)) as.matrix(factors) else NULL
  Ymat <- if (!is.null(Y)) as.matrix(Y) else NULL

  starts <- seq.int(1L, T - width + 1L, by = by)
  frags <- vector("list", length(starts))
  for (k in seq_along(starts)) {
    idx <- starts[k]:(starts[k] + width - 1L)
    Xw <- X[idx, , drop = FALSE]
    Fw <- if (!is.null(Fmat)) Fmat[idx, , drop = FALSE] else NULL
    Yw <- if (!is.null(Ymat)) Ymat[idx, , drop = FALSE] else NULL
    sc <- switch(screen,
                 alpha   = alphaScreening(Xw, factors = Fw, control = control, Y = Yw),
                 sharpe  = sharpeScreening(Xw, control = control, Y = Yw),
                 msharpe = msharpeScreening(Xw, level = level, na.neg = na.neg,
                                            control = control, Y = Yw))
    frag <- .rollAgg(sc)
    frag <- cbind(window = k, index = idx[length(idx)], frag)
    frags[[k]] <- frag
  }
  out <- do.call(rbind, frags)
  if (!is.null(dates)) {
    out <- cbind(out[, 1:2, drop = FALSE], date = dates[out$index],
                 out[, -(1:2), drop = FALSE])
  }
  out$heterogeneity <- 1 - out$pizero
  rownames(out) <- NULL
  class(out) <- c("rollScreening", "data.frame")
  attr(out, "screen") <- screen
  attr(out, "screen_beta") <- screen_beta
  out
}

#' @name plot.rollScreening
#' @title Plot a rolling peer performance screening
#' @description Time-series plot of the rolling peer performance ratios produced
#' by \code{\link{rollScreening}}.
#' @param x A \code{rollScreening} object.
#' @param what Either \code{"heterogeneity"} (\eqn{1-\bar\pi^0}) or
#' \code{"ratios"} (the averaged \eqn{\bar\pi^+,\bar\pi^0,\bar\pi^-}). Default:
#' \code{"heterogeneity"} for a beta screening, \code{"ratios"} otherwise.
#' @param legend A logical value indicating whether to draw a legend. Default:
#' \code{TRUE}.
#' @param ... Further graphical arguments.
#' @return Invisibly returns \code{x}.
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{rollScreening}}.
#' @keywords hplot
#' @examples
#' \donttest{
#'   data("hfdata")
#'   roll <- rollScreening(hfdata[, 1:15], width = 36, by = 12,
#'                         control = list(nCore = 1))
#'   plot(roll)
#' }
#' @export
#' @importFrom graphics matplot legend lines
#' @importFrom grDevices hcl.colors
plot.rollScreening <- function(x, what = NULL, legend = TRUE, ...) {
  beta <- isTRUE(attr(x, "screen_beta"))
  if (is.null(what)) {
    what <- if (beta) "heterogeneity" else "ratios"
  }
  hasdate <- !is.null(x$date)
  # without dates the x axis carries window-end indices, which need a label
  xlab0 <- if (hasdate) "" else "window end (observation index)"

  if (what == "ratios" && !beta) {
    ord <- order(x$index)
    xs <- (if (hasdate) x$date else x$index)[ord]
    do.call(graphics::matplot,
            .gpar(list(x = xs,
                       y = cbind(x$pipos[ord], x$pizero[ord], x$pineg[ord]),
                       type = "l", lty = 1, lwd = 2,
                       col = c("black", "grey60", "grey40"),
                       xlab = xlab0, ylab = "average ratio",
                       ylim = c(0, 1)), list(...)))
    if (legend) {
      graphics::legend("topright", bty = "n", lwd = 2,
                       col = c("black", "grey60", "grey40"),
                       legend = c(expression(bar(pi)^"+"),
                                  expression(bar(pi)^0),
                                  expression(bar(pi)^"-")))
    }
  } else {
    coefs <- if (beta) unique(x$coefficient) else NA
    cols <- grDevices::hcl.colors(max(length(coefs), 1), "Dark3")
    ylab <- if (what == "heterogeneity") expression(1 - bar(pi)^0) else what
    first <- TRUE
    for (i in seq_along(coefs)) {
      sub <- if (beta) x[x$coefficient == coefs[i], , drop = FALSE] else x
      ord <- order(sub$index)
      xs <- (if (hasdate) sub$date else sub$index)[ord]
      ys <- sub[[what]][ord]
      if (first) {
        do.call(graphics::plot,
                .gpar(list(x = xs, y = ys, type = "l", lwd = 2, col = cols[i],
                           ylim = c(0, 1), xlab = xlab0, ylab = ylab),
                      list(...)))
        first <- FALSE
      } else {
        graphics::lines(xs, ys, lwd = 2, col = cols[i])
      }
    }
    if (legend && beta) {
      graphics::legend("topright", bty = "n", lwd = 2, col = cols, legend = coefs)
    }
  }
  invisible(x)
}

## Set of R functions for Sharpe ratio testing

# #' @name .sharpeTesting
# #' @import compiler
.sharpeTesting <- function(x, y, control = list()) {

  x <- as.matrix(x)
  y <- as.matrix(y)

  # process control parameters
  ctr <- processControl(control)

  # check if enough data are available for testing
  dxy <- x - y
  idx <- (!is.nan(dxy) & !is.na(dxy))
  rets <- cbind(x[idx], y[idx])
  T <- sum(idx)
  if (T < ctr$minObs) {
    stop("intersection of 'x' and 'y' is shorter than 'minObs'")
  }

  # sharpe testing
  if (ctr$type == 1) {
    # ==> asymptotic approach
    tmp <- sharpeTestAsymptotic(rets, ctr$hac, ctr$ttype)
  } else {
    # ==> bootstrap approach (iid and circular block bootstrap)
    if (ctr$bBoot == 0) {
      ctr$bBoot <- sharpeBlockSize(x, y, ctr)
    }
    bsids <- bootIndices(T, ctr$nBoot, ctr$bBoot)
    tmp <- sharpeTestBootstrap(rets, bsids, ctr$bBoot, ctr$ttype, ctr$pBoot)
  }

  # info on the funds
  info <- infoFund(rets)

  ## form output
  out <- list(n = T, sharpe = info$sharpe, dsharpe = -diff(info$sharpe),
              tstat = as.vector(tmp$tstat), pval = as.vector(tmp$pval))
  class(out) <- "TESTING"
  return(out)
}

#' @name sharpeTesting
#' @title Testing the difference of Sharpe ratios
#' @description Function which performs the testing of the difference of Sharpe ratios.
#' @details The Sharpe ratio (Sharpe 1994) is one industry standard for measuring the
#' absolute risk adjusted performance of hedge funds. This function performs
#' the testing of Sharpe ratio difference for two funds using the approach by
#' Ledoit and Wolf (2002).
#'
#' For the testing, only the intersection of non-\code{NA} observations for the
#' two funds are used.
#'
#' The argument \code{control} is a list that can supply any of the following
#' components:
#' \itemize{
#' \item \code{'type'} Asymptotic approach (\code{type = 1}) or
#' studentized circular bootstrap approach (\code{type = 2}). Default:
#' \code{type = 1}.
#' \item \code{'ttype'} Test based on ratio (\code{type = 1})
#' or product (\code{type = 2}). Default: \code{type = 2}.
#' \item \code{'hac'} Heteroscedastic-autocorrelation consistent standard
#' errors. Default: \code{hac = FALSE}.
#' \item \code{'nBoot'} Number of bootstrap replications for computing the p-value. Default: \code{nBoot =
#' 499}.
#' \item \code{'bBoot'} Block length in the circular bootstrap. Default:
#' \code{bBoot = 1}, i.e. iid bootstrap. \code{bBoot = 0} uses optimal
#' block-length.
#' \item \code{'pBoot'} Symmetric p-value (\code{pBoot = 1}) or
#' asymmetric p-value (\code{pBoot = 2}). Default: \code{pBoot = 1}.
#' }
#' @param x Vector (of length \eqn{T}) of returns for the first fund. \code{NA}
#' values are allowed.
#' @param y Vector (of length \eqn{T}) returns for the second fund. \code{NA}
#' values are allowed.
#' @param control Control parameters (see *Details*).
#' @return A list with the following components:\cr
#'
#' \code{n}: Number of non-\code{NA} concordant observations.\cr
#'
#' \code{sharpe}: Vector (of length 2) of unconditional Sharpe ratios.\cr
#'
#' \code{dsharpe}: Sharpe ratios difference.\cr
#'
#' \code{tstat}: t-stat of Sharpe ratios differences.\cr
#'
#' \code{pval}: pvalues of test of Sharpe ratios differences.
#' @note Further details on the methodology with an application to the hedge
#' fund industry is given in Ardia and Boudt (2018).
#'
#' Some internal functions where adapted from Michael Wolf MATLAB code.
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{sharpe}}, \code{\link{sharpeScreening}} and
#' \code{\link{msharpeTesting}}.
#' @references
#' Ardia, D., Boudt, K. (2015).
#' Testing equality of modified Sharpe ratios.
#' \emph{Finance Research Letters} \bold{13}, 97--104.
#'
#' Ardia, D., Boudt, K. (2018).
#' The peer performance ratios of hedge funds.
#' \emph{Journal of Banking and Finance} \bold{87}, 351--368.
#'
#' Barras, L., Scaillet, O., Wermers, R. (2010).
#' False discoveries in mutual fund performance: Measuring luck in estimated alphas.
#' \emph{Journal of Finance} \bold{65}(1), 179--216.
#'
#' Sharpe, W.F. (1994).
#' The Sharpe ratio.
#' \emph{Journal of Portfolio Management} \bold{21}(1), 49--58.
#'
#' Ledoit, O., Wolf, M. (2008).
#' Robust performance hypothesis testing with the Sharpe ratio.
#' \emph{Journal of Empirical Finance} \bold{15}(5), 850--859.
#'
#' Storey, J. (2002).
#' A direct approach to false discovery rates.
#' \emph{Journal of the Royal Statistical Society B} \bold{64}(3), 479--498.
#' @keywords htest
#' @examples
#' ## Load the data (randomized data of monthly hedge fund returns)
#' data("hfdata")
#' x = hfdata[,1]
#' y = hfdata[,2]
#'
#' ## Run Sharpe testing (asymptotic)
#' ctr = list(type = 1)
#' out = sharpeTesting(x, y, control = ctr)
#' print(out)
#'
#' ## Run Sharpe testing (asymptotic hac)
#' ctr = list(type = 1, hac = TRUE)
#' out = sharpeTesting(x, y, control = ctr)
#' print(out)
#'
#' ## Run Sharpe testing (iid bootstrap)
#' set.seed(1234)
#' ctr = list(type = 2, nBoot = 100)
#' out = sharpeTesting(x, y, control = ctr)
#' print(out)
#'
#' ## Run Sharpe testing (circular bootstrap)
#' set.seed(1234)
#' ctr = list(type = 2, nBoot = 100, bBoot = 5)
#' out = sharpeTesting(x, y, control = ctr)
#' print(out)
#' @export
#' @importFrom compiler cmpfun
sharpeTesting <- compiler::cmpfun(.sharpeTesting)

#@name .sharpe.ratio.diff
#@title Difference of sharpe ratios
.sharpe.ratio.diff <- function(X, Y, ttype) {

  if (is.null(Y)) {
    Y <- X[, 2, drop = FALSE]
    X <- X[, 1, drop = FALSE]
  }
  n <- nrow(X)
  mu1.hat <- colMeans(X)
  mu2.hat <- colMeans(Y)
  X_ <- sweep(x = X, MARGIN = 2, STATS = mu1.hat, FUN = "-")
  Y_ <- sweep(x = Y, MARGIN = 2, STATS = mu2.hat, FUN = "-")
  sig1.hat <- sqrt(colSums(X_^2)/(n - 1))
  sig2.hat <- sqrt(colSums(Y_^2)/(n - 1))

  if (ttype == 1) {
    SR1.hat <- mu1.hat/sig1.hat
    SR2.hat <- mu2.hat/sig2.hat
  } else {
    SR1.hat <- mu1.hat * sig2.hat
    SR2.hat <- mu2.hat * sig1.hat
  }
  diff <- SR1.hat - SR2.hat
  return(diff)
}
sharpe.ratio.diff <- compiler::cmpfun(.sharpe.ratio.diff)

# #' @name .sharpeTestAsymptotic
# #' @title Asymptotic Sharpe test
# #' @importFrom stats pnorm
# #' @import compiler
.sharpeTestAsymptotic <- function(rets, hac, ttype) {

  dsharpe <- sharpe.ratio.diff(rets, Y = NULL, ttype)
  se <- se.sharpe.asymptotic(rets, hac, ttype)
  tstat <- dsharpe/se
  pval <- 2 * stats::pnorm(-abs(tstat))  # asymptotic normal p-value
  out <- list(dsharpe = dsharpe, tstat = tstat, se = se, pval = pval)
  return(out)
}
sharpeTestAsymptotic <- compiler::cmpfun(.sharpeTestAsymptotic)

# #' @name .se.sharpe.asymptotic
# #' @title Asymptotic standard error
# #' @importFrom stats cov ar
# #' @import compiler
.se.sharpe.asymptotic <- function(X, hac, ttype) {

  # estimation of (robust) Psi function; see Ledoit Wolf paper
  compute.Psi.hat <- function(V.hat, hac) {

    if (hac) {
      T <- length(V.hat[, 1])
      alpha.hat <- compute.alpha.hat(V.hat)
      S.star <- 2.6614 * (alpha.hat * T)^0.2
      Psi.hat <- compute.Gamma.hat(V.hat, 0)
      j <- 1
      while (j < S.star) {
        Gamma.hat <- compute.Gamma.hat(V.hat, j)
        Psi.hat <- Psi.hat + kernel.Parzen(j/S.star) * (Gamma.hat +
                                                          t(Gamma.hat))
        j <- j + 1
      }
      Psi.hat <- (T/(T - 4)) * Psi.hat
    } else {
      Psi.hat <- stats::cov(V.hat)
    }
    return(Psi.hat)

  }

  # Parzen kernel
  kernel.Parzen <- function(x) {

    if (abs(x) <= 0.5)
      result <- 1 - 6 * x^2 + 6 * abs(x)^3 else if (abs(x) <= 1)
        result <- 2 * (1 - abs(x))^3 else result <- 0
        return(result)

  }

  compute.alpha.hat <- function(V.hat) {

    p <- ncol(V.hat)
    num <- den <- 0
    for (i in 1:p) {
      fit <- stats::ar(V.hat[, i], 0, 1, method = "ols")
      rho.hat <- as.numeric(fit[2])
      sig.hat <- sqrt(as.numeric(fit[3]))
      num <- num + 4 * rho.hat^2 * sig.hat^4/(1 - rho.hat)^8
      den <- den + sig.hat^4/(1 - rho.hat)^4
    }
    return(num/den)

  }

  compute.Gamma.hat <- function(V.hat, j) {

    T <- nrow(V.hat)
    p <- ncol(V.hat)
    Gamma.hat <- matrix(0, p, p)
    if (j >= T)
      stop("j must be smaller than the row dimension!")
    for (i in ((j + 1):T)) {
      Gamma.hat <- Gamma.hat + tcrossprod(V.hat[i, ], V.hat[i - j,
                                                            ])
    }
    Gamma.hat <- Gamma.hat/T
    return(Gamma.hat)

  }

  T <- nrow(X)

  if (ttype == 1) {
    mu.hat <- colMeans(X)
    gamma.hat <- colMeans(X^2)
    gradient <- vector("double", 4)
    gradient[1] <- gamma.hat[1]/(gamma.hat[1] - mu.hat[1]^2)^1.5
    gradient[2] <- -gamma.hat[2]/(gamma.hat[2] - mu.hat[2]^2)^1.5
    gradient[3] <- -0.5 * mu.hat[1]/(gamma.hat[1] - mu.hat[1]^2)^1.5
    gradient[4] <- 0.5 * mu.hat[2]/(gamma.hat[2] - mu.hat[2]^2)^1.5
    V.hat <- matrix(NA, T, 4)
    V.hat[, 1:2] <- sweep(x = X, MARGIN = 2, STATS = mu.hat, FUN = "-")
    V.hat[, 3:4] <- sweep(x = X^2, MARGIN = 2, STATS = gamma.hat, FUN = "-")
  } else {
    m1 <- colMeans(X)
    X_ <- sweep(x = X, MARGIN = 2, STATS = m1, FUN = "-")
    m2 <- colMeans(X_^2)
    g2 <- m2 + m1^2
    dm1i <- c(1, 0, 0, 0)
    dm1j <- c(0, 0, 1, 0)
    dsigi <- 1/(2 * sqrt(m2[1])) * c(-2 * m1[1], 1, 0, 0)
    dsigj <- 1/(2 * sqrt(m2[2])) * c(0, 0, -2 * m1[2], 1)
    tmp1 <- dm1i * sqrt(m2[2]) + dsigj * m1[1]
    tmp2 <- dm1j * sqrt(m2[1]) + dsigi * m1[2]
    gradient <- tmp1 - tmp2
    V.hat <- matrix(NA, T, 4)
    V.hat[, c(1, 3)] <- sweep(x = X, MARGIN = 2, STATS = m1, FUN = "-")
    V.hat[, c(2, 4)] <- sweep(x = X^2, MARGIN = 2, STATS = g2, FUN = "-")
  }

  Psi.hat <- compute.Psi.hat(V.hat, hac)
  se <- as.numeric(sqrt(crossprod(gradient, Psi.hat %*% gradient)/T))
  return(se)

}
se.sharpe.asymptotic <- compiler::cmpfun(.se.sharpe.asymptotic)

# #' @name .sharpeTestBootstrap
# #' @import compiler
.sharpeTestBootstrap <- function(rets, bsids, b, ttype, pBoot, d = 0) {

  T <- nrow(rets)
  x <- rets[, 1, drop = FALSE]
  y <- rets[, 2, drop = FALSE]
  dsharpe <- as.numeric(sharpe.ratio.diff(x, y, ttype) - d)
  se <- se.sharpe.bootstrap(x, y, b, ttype)
  # se = se.sharpe.asymptotic(X = cbind(x, y), hac = TRUE, ttype = ttype)

  # bootstrap indices
  nBoot <- ncol(bsids)
  bsidx <- 1 + bsids%%T  # ensure that the bootstrap indices match the length of the time series
  bsX <- matrix(x[bsidx], T, nBoot)
  bsY <- matrix(y[bsidx], T, nBoot)

  bsdsharpe <- sharpe.ratio.diff(bsX, bsY, ttype)
  bsse <- se.sharpe.bootstrap(bsX, bsY, b, ttype)
  tstat <- dsharpe/se

  if (pBoot == 1) {
    # first type p-value calculation. '>=' matches the modified Sharpe test and
    # is the conservative convention: the bootstrap statistics live on a finite
    # resample space, so exact ties with |tstat| are possible and are counted
    # against the null rather than ignored.
    bststat <- abs(bsdsharpe - dsharpe)/bsse
    pval <- (sum(bststat >= abs(tstat)) + 1)/(nBoot + 1)
    # pval = sum(bststat >= abs(tstat)) / nBoot
  } else {
    # second type p-value calculation (as in Barras)
    bststat <- (bsdsharpe - dsharpe)/bsse
    pval <- 2 * min(sum(bststat > tstat) + 1, sum(bststat < tstat) +
                      1)/(nBoot + 1)
    # pval = 2 * min(sum(bststat > tstat), sum(bststat < tstat)) / nBoot
  }

  out <- list(dsharpe = dsharpe, tstat = tstat, se = se, bststat = bststat,
              pval = pval)
  return(out)
}
sharpeTestBootstrap <- compiler::cmpfun(.sharpeTestBootstrap)

# #' @name .se.sharpe.bootstrap
# #' @title Bootstrap standard error
# #' @importFrom stats cov
# #' @import compiler
.se.sharpe.bootstrap <- function(X, Y, b, ttype) {

  ## Compute Psi with two approaches: 1) iid bootstrap, 2) circular block
  ## bootstrap
  compute.Psi.hat <- function(V.hat, b) {

    T <- length(V.hat[, 1])
    if (b == 1) {
      # ==> standard estimation
      Psi.hat <- stats::cov(V.hat)
    } else {
      # ==> block estimation
      l <- floor(T/b)
      Psi.hat <- matrix(0, 4, 4)
      for (j in (1:l)) {
        zeta <- b^0.5 * colMeans(V.hat[((j - 1) * b + 1):(j * b),
                                       , drop = FALSE])
        Psi.hat <- Psi.hat + tcrossprod(zeta)
      }
      Psi.hat <- Psi.hat/l
    }
    return(Psi.hat)

  }

  T <- nrow(X)
  N <- ncol(Y)
  if (ttype == 1) {
    mu1.hat <- colMeans(X)
    mu2.hat <- colMeans(Y)
    gamma1.hat <- colMeans(X^2)
    gamma2.hat <- colMeans(Y^2)
    gradient <- array(NA, c(4, 1, N))
    gradient[1, 1, ] <- gamma1.hat/(gamma1.hat - mu1.hat^2)^1.5
    gradient[2, 1, ] <- -gamma2.hat/(gamma2.hat - mu2.hat^2)^1.5
    gradient[3, 1, ] <- -0.5 * mu1.hat/(gamma1.hat - mu1.hat^2)^1.5
    gradient[4, 1, ] <- 0.5 * mu2.hat/(gamma2.hat - mu2.hat^2)^1.5
    V.hat <- array(NA, c(T, 4, N))
    V.hat[, 1, ] <- sweep(x = X, MARGIN = 2, STATS = mu1.hat, FUN = "-")
    V.hat[, 2, ] <- sweep(x = Y, MARGIN = 2, STATS = mu2.hat, FUN = "-")
    V.hat[, 3, ] <- sweep(x = X^2, MARGIN = 2, STATS = gamma1.hat,
                          FUN = "-")
    V.hat[, 4, ] <- sweep(x = Y^2, MARGIN = 2, STATS = gamma2.hat,
                          FUN = "-")
  } else {
    m1X <- colMeans(X)
    m1Y <- colMeans(Y)
    X_ <- sweep(x = X, MARGIN = 2, STATS = m1X, FUN = "-")
    Y_ <- sweep(x = Y, MARGIN = 2, STATS = m1Y, FUN = "-")
    m2X <- colMeans(X_^2)
    m2Y <- colMeans(Y_^2)
    g2X <- m2X + m1X^2
    g2Y <- m2Y + m1Y^2

    cst1X <- 1/(2 * sqrt(m2X))
    cst1Y <- 1/(2 * sqrt(m2Y))

    dm1X <- matrix(rep(c(1, 0, 0, 0), N), 4, N, FALSE)
    dm1Y <- matrix(rep(c(0, 0, 1, 0), N), 4, N, FALSE)
    dsigX <- rbind(-2 * cst1X * m1X, cst1X, 0, 0)
    dsigY <- rbind(0, 0, -2 * cst1Y * m1Y, cst1Y)

    # matrix form
    m1X_ <- matrix(m1X, nrow = 4, ncol = N, byrow = TRUE)
    m1Y_ <- matrix(m1Y, nrow = 4, ncol = N, byrow = TRUE)
    m2X_ <- matrix(m2X, nrow = 4, ncol = N, byrow = TRUE)
    m2Y_ <- matrix(m2Y, nrow = 4, ncol = N, byrow = TRUE)

    dm1X_ <- matrix(dm1X, nrow = 4, ncol = N, byrow = FALSE)
    dm1Y_ <- matrix(dm1Y, nrow = 4, ncol = N, byrow = FALSE)
    dsigX_ <- matrix(dsigX, nrow = 4, ncol = N, byrow = FALSE)
    dsigY_ <- matrix(dsigY, nrow = 4, ncol = N, byrow = FALSE)

    cst2X_ <- sqrt(m2X_)
    cst2Y_ <- sqrt(m2Y_)

    # gradient
    tmp1 <- dm1X_ * cst2Y_ + dsigY_ * m1X_
    tmp2 <- dm1Y_ * cst2X_ + dsigX_ * m1Y_

    # =======
    gradient <- array(NA, c(4, 1, N))
    gradient[1:4, 1, ] <- tmp1 - tmp2
    V.hat <- array(NA, c(T, 4, N))
    V.hat[, 1, ] <- sweep(x = X, MARGIN = 2, STATS = m1X, FUN = "-")
    V.hat[, 3, ] <- sweep(x = Y, MARGIN = 2, STATS = m1Y, FUN = "-")
    V.hat[, 2, ] <- sweep(x = X^2, MARGIN = 2, STATS = g2X, FUN = "-")
    V.hat[, 4, ] <- sweep(x = Y^2, MARGIN = 2, STATS = g2Y, FUN = "-")
  }
  Psi.hat <- array(apply(X = V.hat, MARGIN = 3, FUN = compute.Psi.hat,
                         b = b), c(4, 4, N))
  se <- vector("double", N)
  for (i in 1:N) {
    se[i] <- sqrt(crossprod(gradient[, , i], Psi.hat[, , i] %*% gradient[,
                                                                         , i])/T)
  }
  return(se)

}
se.sharpe.bootstrap <- compiler::cmpfun(.se.sharpe.bootstrap)

## Set of R functions for Sharpe screening

# #' @name .sharpeScreening
# #' @import compiler
.sharpeScreening <- function(X, control = list(), Y = NULL) {

  # cross-group screening: each fund in X against every fund in group Y
  if (!is.null(Y)) {
    return(.sharpeScreeningXY(X, Y, control = control))
  }

  # process control
  ctr <- processControl(control)
  if (ctr$bBoot == 0) {
    stop("'bBoot = 0' (data-driven block length) is not supported in screening; ",
         "set an explicit block length 'bBoot >= 1'.")
  }

  # size of inputs and outputs
  X <- as.matrix(X)
  T <- nrow(X)
  N <- ncol(X)
  if (N < 2L) {
    stop("within-group screening needs at least two funds in 'X'; supply 'Y' to screen a single fund against a peer group")
  }
  pval <- dsharpe <- tstat <- matrix(data = NA, N, N)

  # determine which pairs can be compared (in a matrix way)
  Y <- 1 * (!is.nan(X) & !is.na(X))
  YY <- crossprod(Y)  #YY = t(Y) %*% Y # row i indicates how many observations in common with column k
  pairLens <- YY[upper.tri(YY)]  # pairwise complete-case lengths (before thresholding)
  YY[YY < ctr$minObs] <- 0
  YY[YY > 0] <- 1
  liststocks <- c(1:nrow(YY))[rowSums(YY) > ctr$minObsPi]

  # pre-generate the bootstrap indices in the master, one matrix per distinct
  # pair length (workers select by length; see bootIndicesByLen). Only needed
  # for the bootstrap test, and only for pairs that pass 'minObs', so that the
  # asymptotic path never depends on 'bBoot'.
  bsids <- NULL
  if (ctr$type == 2) {
    bsids <- bootIndicesByLen(pairLens[pairLens >= ctr$minObs],
                              ctr$nBoot, ctr$bBoot)
  }

  if (length(liststocks) > 1) {
    liststocks <- liststocks[1:(length(liststocks) - 1)]

    if (ctr$nCore == 1) {
      # serial path: no PSOCK cluster (avoids the per-call cluster overhead)
      z <- lapply(as.list(liststocks), sharpeScreeningi,
                  rdata = X, T = T, N = N, nBoot = ctr$nBoot, bsids = bsids,
                  minObs = ctr$minObs, type = ctr$type, hac = ctr$hac,
                  b = ctr$bBoot, ttype = ctr$ttype, pBoot = ctr$pBoot)
    } else {
      cl <- parallel::makeCluster(ctr$nCore)
      on.exit(parallel::stopCluster(cl), add = TRUE)
      z <- parallel::clusterApplyLB(cl = cl, x = as.list(liststocks), fun = sharpeScreeningi,
                                    rdata = X, T = T, N = N, nBoot = ctr$nBoot, bsids = bsids,
                                    minObs = ctr$minObs, type = ctr$type, hac = ctr$hac, b = ctr$bBoot,
                                    ttype = ctr$ttype, pBoot = ctr$pBoot)
    }

    for (i in 1:length(liststocks)) {
      out <- z[[i]]
      id <- liststocks[i]
      pval[id, id:N] <- pval[id:N, id] <- out[[2]][id:N]
      dsharpe[id, id:N] <- out[[1]][id:N]
      dsharpe[id:N, id] <- -out[[1]][id:N]
      tstat[id, id:N] <- out[[3]][id:N]
      tstat[id:N, id] <- -out[[3]][id:N]
    }
  }

  # pi
  pi <- computePi(pval = pval, dalpha = dsharpe, tstat = tstat, lambda = ctr$lambda,
                  nBoot = ctr$nBoot, bpos = ctr$gammaPos, bneg = ctr$gammaNeg,
                  fast = ctr$fastAdjust)

  # info on the funds
  info <- infoFund(X)

  # form output
  out <- list(n = info$nObs, npeer = colSums(!is.na(pval)), sharpe = info$sharpe,
              dsharpe = dsharpe, pval = pval, tstat = tstat, lambda = pi$lambda,
              pizero = pi$pizero, pipos = pi$pipos, pineg = pi$pineg)
  class(out) <- "SCREENING"

  return(out)
}

#' @name sharpeScreening
#' @title Screening using the Sharpe outperformance ratio
#' @description Function which performs the screening of a universe of returns, and
#' computes the Sharpe outperformance ratio.
#' @details The Sharpe ratio (Sharpe 1994) is one industry standard for measuring the
#' absolute risk adjusted performance of hedge funds. We propose to complement
#' the Sharpe ratio with the fund's outperformance ratio, defined as the
#' percentage number of funds that have a significantly lower Sharpe ratio. In
#' a pairwise testing framework, a fund can have a significantly higher Sharpe
#' ratio because of luck. We correct for this by applying the false discovery
#' rate approach by Storey (2002).
#'
#' For the testing, only the intersection of non-\code{NA} observations for the
#' two funds are used.
#'
#' The methodology proceeds as follows:
#' \itemize{
#' \item (1) compute all
#' pairwise tests of Sharpe differences using the bootstrap approach of Ledoit
#' and Wolf (2002). This means that for a universe of \eqn{N} funds, we perform
#' \eqn{N(N-1)/2}{N*(N-1)/2} tests. The algorithm has been parallelized and the
#' computational burden can be split across several cores. The number of cores
#' can be defined in \code{control}, see below.
#' \item (2) for each fund, the
#' false discovery rate approach by Storey (2002) is used to determine the
#' proportions over, equal, and underperforming funds, in terms of Sharpe ratio,
#' in the database.
#' }
#' The argument \code{control} is a list that can supply any of the following
#' components:
#' \itemize{
#' \item \code{'type'} Asymptotic approach (\code{type = 1}) or
#' studentized circular bootstrap approach (\code{type = 2}). Default:
#' \code{type = 1}.
#' \item \code{'ttype'} Test based on ratio (\code{type = 1})
#' or product (\code{type = 2}). Default: \code{type = 2}.
#' \item \code{'hac'} Heteroscedastic-autocorrelation consistent standard
#' errors. Default: \code{hac = FALSE}.
#' \item \code{'nBoot'} Number of bootstrap replications for computing the p-value. Default: \code{nBoot =
#' 499}.
#' \item \code{'bBoot'} Block length in the circular bootstrap. Default:
#' \code{bBoot = 1}, i.e. iid bootstrap. (The data-driven choice
#' \code{bBoot = 0} is only available in \code{\link{sharpeTesting}}, not in
#' screening.)
#' \item \code{'pBoot'} Symmetric p-value (\code{pBoot = 1}) or
#' asymmetric p-value (\code{pBoot = 2}). Default: \code{pBoot = 1}.
#' \item \code{'nCore'} Number of cores to be used. Default: \code{nCore = 1}.
#' \item \code{'minObs'} Minimum number of concordant observations to compute
#' the ratios. Default: \code{minObs = 10}.
#' \item \code{'minObsPi'} Minimum
#' number of observations to compute pi0. Default: \code{minObsPi = 1}.
#' \item \code{'lambda'} Threshold value to compute pi0. Default: \code{lambda
#' = NULL}, i.e. data driven choice.
#' \item \code{'gammaPos'} One-sided quantile level (of the standard Normal
#' distribution) used as the critical value for counting outperformed peers:
#' a peer counts as outperformed when the pairwise t-statistic exceeds
#' \code{qnorm(gammaPos)} (a \emph{negative} threshold for
#' \code{gammaPos < 0.5}), and the expected fraction \code{1 - gammaPos} of
#' false positives among the equal-performing peers is then subtracted.
#' Default: \code{gammaPos = 0.4} (the value recommended in Ardia and Boudt, 2018).
#' \item \code{'gammaNeg'} Mirror image of \code{gammaPos} for the peers that
#' outperform the focal fund: the count uses \code{tstat <= qnorm(gammaNeg)}
#' and subtracts the expected fraction \code{gammaNeg} of false positives.
#' Default: \code{gammaNeg = 0.6}.
#' \item \code{'fastAdjust'} Use a fast vectorised inversion in the
#' truncated-normal bias correction of \eqn{\pi^0}{pi0} instead of one
#' \code{uniroot} call per value. This is the dominant cost when \code{lambda}
#' is data driven and gives a large speed-up on big universes. The bisection
#' locates the root to about 1e-12; since \code{uniroot} stops at its own
#' tolerance (about 1.2e-4), the two paths typically differ by a few 1e-5, the
#' fast path being the more accurate. Default: \code{fastAdjust = FALSE},
#' i.e. the original code path, kept as default so that published results
#' reproduce exactly.
#' }
#' @param X Matrix \eqn{(T \times N)}{(TxN)} of \eqn{T} returns for the \eqn{N}
#' funds. \code{NA} values are allowed.
#' @param control Control parameters (see *Details*).
#' @param Y Optional matrix \eqn{(T \times M)}{(TxM)} of returns for a second
#' (peer) group of \eqn{M} funds. When supplied, the ratios are computed for
#' each fund in \code{X} against the funds in \code{Y} (cross-group screening)
#' instead of against the other funds in \code{X}; a single focal fund versus a
#' peer group corresponds to \code{X} being a vector. Columns of \code{Y}
#' identical to the focal fund are automatically excluded. Default:
#' \code{Y = NULL}, i.e. within-group screening.
#' @return A list with the following components:\cr
#'
#' \code{n}: Vector (of length \eqn{N}) of number of non-\code{NA}
#' observations.\cr
#'
#' \code{npeer}: Vector (of length \eqn{N}) of number of available peers.\cr
#'
#' \code{sharpe}: Vector (of length \eqn{N}) of unconditional Sharpe ratios.\cr
#'
#' \code{dsharpe}: Matrix (of size \eqn{N \times N}{NxN}) of Sharpe ratios
#' differences.\cr
#'
#' \code{tstat}: Matrix (of size \eqn{N \times N}{NxN}) of t-statistics.\cr
#'
#' \code{pval}: Matrix (of size \eqn{N \times N}{NxN}) of pvalues of test for Sharpe
#' ratios differences.\cr
#'
#' \code{lambda}: vector (of length \eqn{N}) of lambda values.\cr
#'
#' \code{pizero}: vector (of length \eqn{N}) of probability of equal
#' performance.\cr
#'
#' \code{pipos}: vector (of length \eqn{N}) of probability of outperformance
#' performance.\cr
#'
#' \code{pineg}: Vector (of length \eqn{N}) of probability of underperformance
#' performance.
#' @note Further details on the methodology with an application to the hedge
#' fund industry is given in Ardia and Boudt (2018).
#'
#' Some internal functions where adapted from Michael Wolf MATLAB code.
#'
#' Application of the false discovery rate approach applied to the mutual fund
#' industry has been presented in Barras, Scaillet and Wermers (2010).
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{sharpe}}, \code{\link{sharpeTesting}},
#' \code{\link{msharpeScreening}} and \code{\link{alphaScreening}}.
#' @references
#' Ardia, D., Boudt, K. (2015).
#' Testing equality of modified Sharpe ratios.
#' \emph{Finance Research Letters} \bold{13}, 97--104.
#'
#' Ardia, D., Boudt, K. (2018).
#' The peer performance ratios of hedge funds.
#' \emph{Journal of Banking and Finance} \bold{87}, 351--368.
#'
#' Barras, L., Scaillet, O., Wermers, R. (2010).
#' False discoveries in mutual fund performance: Measuring luck in estimated alphas.
#' \emph{Journal of Finance} \bold{65}(1), 179--216.
#'
#' Sharpe, W.F. (1994).
#' The Sharpe ratio.
#' \emph{Journal of Portfolio Management} \bold{21}(1), 49--58.
#'
#' Ledoit, O., Wolf, M. (2008).
#' Robust performance hypothesis testing with the Sharpe ratio.
#' \emph{Journal of Empirical Finance} \bold{15}(5), 850--859.
#'
#' Storey, J. (2002).
#' A direct approach to false discovery rates.
#' \emph{Journal of the Royal Statistical Society B} \bold{64}(3), 479--498.
#' @keywords htest
#' @examples
#' ## Load the data (randomized data of monthly hedge fund returns)
#' data("hfdata")
#' rets = hfdata[,1:4]
#'
#' ## Sharpe screening
#' sharpeScreening(rets, control = list(nCore = 1))
#'
#' ## Sharpe screening with HAC standard errors (asymptotic test)
#' sharpeScreening(rets, control = list(nCore = 1, hac = TRUE))
#'
#' ## Sharpe screening with the studentized circular bootstrap
#' ## (the 'hac' flag is ignored when type = 2)
#' sharpeScreening(rets, control = list(nCore = 1, type = 2))
#' @export
#' @importFrom compiler cmpfun
sharpeScreening <- compiler::cmpfun(.sharpeScreening)

#@name .sharpeScreeningi
#@title Sharpe ratio screening for fund i again its peers
.sharpeScreeningi <- function(i, rdata, T, N, nBoot, bsids, minObs, type,
                              hac, b, ttype, pBoot) {

  nPeer <- N - i
  X <- matrix(rdata[, i], nrow = T, ncol = nPeer)
  Y <- matrix(rdata[, (i + 1):N], nrow = T, ncol = nPeer)

  dXY <- X - Y
  idx <- (!is.nan(dXY) & !is.na(dXY))
  X[!idx] <- NA
  Y[!idx] <- NA
  nObs <- colSums(idx)

  pvali <- dsharpei <- tstati <- rep(NA, N)

  k <- 0
  for (j in (i + 1):N) {
    k <- k + 1
    if (nObs[k] < minObs) {
      next
    }
    rets <- cbind(X[idx[, k], k], Y[idx[, k], k])

    if (type == 1) {
      tmp <- sharpeTestAsymptotic(rets, hac, ttype)
    } else {
      # indices pre-generated in the master for this pair's length
      bs <- bsids[[as.character(nObs[k])]]
      if (is.null(bs)) {
        next  # block length exceeds this pair's sample size
      }
      tmp <- sharpeTestBootstrap(rets, bs, b, ttype, pBoot)
    }
    if (!is.finite(tmp$tstat)) {
      next  # degenerate pair (e.g. zero-variance series)
    }

    dsharpei[j] <- tmp$dsharpe
    pvali[j] <- tmp$pval
    tstati[j] <- tmp$tstat
  }

  out <- list(dsharpei = dsharpei, pvali = pvali, tstati = tstati)
  return(out)
}
sharpeScreeningi <- compiler::cmpfun(.sharpeScreeningi)

## S3 methods for the 'TESTING' and 'SCREENING' objects

## internal helper: identify the performance measure carried by a SCREENING
## object and return its label and the (possibly matrix) values
.screeningPerf <- function(x) {
  if (!is.null(x$alpha)) {
    return(list(value = x$alpha, label = "alpha"))
  } else if (!is.null(x$sharpe)) {
    return(list(value = x$sharpe, label = "Sharpe ratio"))
  } else if (!is.null(x$msharpe)) {
    return(list(value = x$msharpe, label = "modified Sharpe ratio"))
  }
  stop("unknown performance measure in the 'SCREENING' object")
}

## internal: merge caller-supplied graphical arguments over the method's
## defaults, so that the documented "further graphical arguments" in '...'
## really are accepted. Passing e.g. main= or xlab= to a plot method used to
## raise "formal argument matched by multiple actual arguments".
.gpar <- function(defaults, dots) {
  if (length(dots) > 0L) defaults[names(dots)] <- dots
  defaults
}

## internal: single focal fund (one stacked bar of pi+/pi0/pi-)
.plotScreeningSingle <- function(pipos, pizero, pineg, ny, colorset, label, ...) {
  mat <- matrix(100 * c(pipos, pizero, pineg), ncol = 1)
  op <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(op))
  graphics::par(mar = c(4, 1, 4, 1))
  main <- if (is.null(ny)) {
    paste0("Peer performance (", label, ")")
  } else {
    paste0("Focal fund vs peer group of ", ny, " (", label, ")")
  }
  do.call(graphics::barplot,
          .gpar(list(height = mat, horiz = TRUE, xlim = c(0, 100),
                     col = colorset, border = NA, names.arg = "",
                     axes = FALSE, main = main), list(...)))
  graphics::axis(side = 1, at = seq(0, 100, by = 25),
                 labels = paste0(seq(0, 100, by = 25), "%"))
  graphics::box()
  graphics::legend("top", horiz = TRUE, bty = "n", inset = c(0, -0.16),
                   xpd = TRUE, fill = colorset,
                   legend = c(expression(hat(pi)^"+"), expression(hat(pi)^0),
                              expression(hat(pi)^"-")))
  invisible(t(mat / 100))
}

#' @name print.TESTING
#' @title Print method for the 'TESTING' object
#' @description Print method for the \code{TESTING} object returned by
#' \code{\link{sharpeTesting}}, \code{\link{msharpeTesting}} and
#' \code{\link{alphaTesting}}.
#' @param x A \code{TESTING} object.
#' @param ... Further arguments passed to or from other methods.
#' @return Invisibly returns \code{x}.
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{sharpeTesting}}, \code{\link{msharpeTesting}} and
#' \code{\link{alphaTesting}}.
#' @keywords htest
#' @export
print.TESTING <- function(x, ...) {
  if (!is.null(x$sharpe)) {
    label <- "Sharpe ratio"; vals <- x$sharpe; dval <- x$dsharpe
  } else if (!is.null(x$msharpe)) {
    label <- "modified Sharpe ratio"; vals <- x$msharpe; dval <- x$dmsharpe
  } else {
    label <- "alpha"; vals <- x$alpha; dval <- x$dalpha
  }
  # for a screen_beta alpha test, x$alpha is a (K+1) x 2 matrix (rows =
  # coefficients, cols = fund x / fund y) -- keep the alpha row, not a beta
  if (is.matrix(vals)) vals <- vals[1, ] else vals <- as.vector(vals)
  dval <- as.vector(dval)[1]
  cat("\nPeer performance test (", label, ")\n", sep = "")
  cat("  Number of (concordant) observations: ", x$n[1], "\n", sep = "")
  cat("  Measure (fund x, fund y) : ",
      formatC(vals[1], format = "f", digits = 4), ", ",
      formatC(vals[2], format = "f", digits = 4), "\n", sep = "")
  cat("  Difference (x - y)       : ",
      formatC(dval, format = "f", digits = 4), "\n", sep = "")
  cat("  t-statistic              : ",
      formatC(as.vector(x$tstat)[1], format = "f", digits = 4), "\n", sep = "")
  cat("  p-value                  : ",
      formatC(as.vector(x$pval)[1], format = "f", digits = 4), "\n", sep = "")
  invisible(x)
}

#' @name print.SCREENING
#' @title Print method for the 'SCREENING' object
#' @description Print method for the \code{SCREENING} object returned by
#' \code{\link{alphaScreening}}, \code{\link{sharpeScreening}} and
#' \code{\link{msharpeScreening}}.
#' @param x A \code{SCREENING} object.
#' @param ... Further arguments passed to or from other methods.
#' @return Invisibly returns \code{x}.
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{alphaScreening}}, \code{\link{sharpeScreening}},
#' \code{\link{msharpeScreening}} and \code{\link{plot.SCREENING}}.
#' @keywords htest
#' @export
print.SCREENING <- function(x, ...) {
  perf <- .screeningPerf(x)
  pipos <- x$pipos; pizero <- x$pizero; pineg <- x$pineg
  if (is.matrix(pipos)) {
    pipos <- pipos[1, ]; pizero <- pizero[1, ]; pineg <- pineg[1, ]
  }
  N <- length(pipos)
  cat("\nPeer performance screening (", perf$label, ")\n", sep = "")
  cat("  Number of funds : ", N, "\n", sep = "")
  cat("  Average ratios  : ",
      "pi+ = ", formatC(100 * mean(pipos,  na.rm = TRUE), format = "f", digits = 1), "%",
      "  pi0 = ", formatC(100 * mean(pizero, na.rm = TRUE), format = "f", digits = 1), "%",
      "  pi- = ", formatC(100 * mean(pineg,  na.rm = TRUE), format = "f", digits = 1), "%",
      "\n", sep = "")
  cat("  Use plot() to display the peer performance screening plot.\n")
  invisible(x)
}

#' @name as.data.frame.SCREENING
#' @title Coerce a 'SCREENING' object to a data frame
#' @description Returns a tidy \code{data.frame} with one row per fund (or, when
#' the screening was run with \code{screen_beta = TRUE}, one row per
#' fund/coefficient), suitable for downstream filtering, ranking, and plotting.
#' @param x A \code{SCREENING} object.
#' @param row.names,optional Currently ignored.
#' @param ... Further arguments (ignored).
#' @return A \code{data.frame} with columns \code{fund}, the performance measure
#' (\code{alpha}, \code{sharpe}, or \code{msharpe}), \code{pipos}, \code{pizero},
#' \code{pineg}, \code{npeer}, and \code{lambda}; a \code{coefficient} column is
#' added when \code{screen_beta = TRUE}.
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{alphaScreening}}.
#' @keywords manip
#' @examples
#' \donttest{
#'   data("hfdata")
#'   sc <- alphaScreening(hfdata[, 1:10], control = list(nCore = 1))
#'   head(as.data.frame(sc))
#' }
#' @export
as.data.frame.SCREENING <- function(x, row.names = NULL, optional = FALSE, ...) {
  short <- if (!is.null(x$alpha)) "alpha" else if (!is.null(x$sharpe)) "sharpe" else "msharpe"
  measure <- .screeningPerf(x)$value
  pipos <- x$pipos; pizero <- x$pizero; pineg <- x$pineg
  lambda <- x$lambda; npeer <- x$npeer

  if (!is.matrix(pizero)) {
    fund_lbl <- names(pizero)
    if (is.null(fund_lbl)) fund_lbl <- names(x$n)
    if (is.null(fund_lbl)) fund_lbl <- seq_along(pizero)
    df <- data.frame(fund = fund_lbl, stringsAsFactors = FALSE)
    df[[short]] <- as.vector(measure)
    df$pipos <- as.vector(pipos); df$pizero <- as.vector(pizero)
    df$pineg <- as.vector(pineg)
    df$npeer <- as.vector(npeer)
    df$lambda <- if (is.null(lambda)) NA_real_ else as.vector(lambda)
  } else {
    ncoef <- nrow(pizero); N <- ncol(pizero)
    cn <- rownames(pizero); if (is.null(cn)) cn <- paste0("coef", seq_len(ncoef))
    fcols <- colnames(pizero)
    if (is.null(fcols)) fcols <- names(x$n)
    if (is.null(fcols)) fcols <- seq_len(N)
    df <- data.frame(coefficient = rep(cn, times = N),
                     fund = rep(fcols, each = ncoef),
                     stringsAsFactors = FALSE)
    df[[short]] <- as.vector(measure)
    df$pipos <- as.vector(pipos); df$pizero <- as.vector(pizero)
    df$pineg <- as.vector(pineg)
    df$npeer <- as.vector(npeer)
    df$lambda <- if (is.null(lambda)) NA_real_ else as.vector(lambda)
  }
  rownames(df) <- NULL
  df
}

#' @name summary.SCREENING
#' @title Summary method for the 'SCREENING' object
#' @description Produces a ranked summary of a screening: distribution of the
#' performance measure, the peer performance ratios, optional win/loss counts
#' from the pairwise tests, and the top/bottom funds (overall and by
#' out-/underperformance ratio). The win/loss counts require the square
#' (within-group) pairwise matrices and are therefore omitted for cross-group
#' screenings (a \code{Y} peer group or \code{\link{targetPeerPerformance}}).
#' Originally contributed by Murilo Andre Peres Pereira.
#' @param object A \code{SCREENING} object.
#' @param coef For a \code{screen_beta} screening, which coefficient row to use
#' (default 1, the alpha).
#' @param top Number of best/worst funds to display. Default: 5.
#' @param p_level Significance level for the win/loss counts (used only when the
#' object carries square \code{pval}/\code{tstat} matrices). Default: 0.05.
#' @param ... Further arguments (ignored).
#' @return An object of class \code{summary.SCREENING}.
#' @author Murilo Andre Peres Pereira, David Ardia and Kris Boudt.
#' @seealso \code{\link{alphaScreening}}.
#' @keywords htest
#' @examples
#' \donttest{
#'   data("hfdata")
#'   sc <- alphaScreening(hfdata[, 1:10], control = list(nCore = 1))
#'   summary(sc)
#' }
#' @export
#' @importFrom stats quantile median
#' @importFrom utils head tail
summary.SCREENING <- function(object, coef = 1L, top = 5L, p_level = 0.05, ...) {
  x <- object
  measure <- if (!is.null(x$alpha)) "alpha" else if (!is.null(x$sharpe)) "sharpe" else if (!is.null(x$msharpe)) "msharpe" else stop("no recognized measure in the 'SCREENING' object")

  est <- x[[measure]]
  if (is.matrix(est)) {
    ok_coef <- (is.numeric(coef) && length(coef) == 1L && coef >= 1 && coef <= nrow(est)) ||
               (is.character(coef) && length(coef) == 1L && coef %in% rownames(est))
    if (!ok_coef) stop("'coef' must be a valid coefficient row index or name")
    est <- est[coef, ]
  }
  est <- as.numeric(est)

  funds <- names(x$pizero)
  if (is.null(funds)) funds <- names(x$n)
  if (is.null(funds) || length(funds) != length(est)) {
    funds <- paste0("Fund ", seq_along(est))
  }
  pick <- function(v) {
    if (is.null(v)) return(rep(NA_real_, length(est)))
    if (is.matrix(v)) v <- v[coef, ]
    as.numeric(v)
  }

  tab <- data.frame(fund = funds, n = pick(x$n), npeer = pick(x$npeer),
                    estimate = est, lambda = pick(x$lambda),
                    pizero = pick(x$pizero), pipos = pick(x$pipos),
                    pineg = pick(x$pineg), stringsAsFactors = FALSE)

  ## win/loss counts only when pval/tstat are square (within-group screening)
  if (is.matrix(x$pval) && is.matrix(x$tstat) &&
      nrow(x$pval) == length(funds) && ncol(x$pval) == length(funds)) {
    tab$wins   <- rowSums((x$pval < p_level) & (x$tstat > 0), na.rm = TRUE)
    tab$losses <- rowSums((x$pval < p_level) & (x$tstat < 0), na.rm = TRUE)
    tab$net    <- tab$wins - tab$losses
  }

  tab$rank <- rank(-tab$estimate, ties.method = "first")
  tab <- tab[order(tab$rank), , drop = FALSE]

  est_stats <- c(min  = min(tab$estimate, na.rm = TRUE),
                 q25  = unname(stats::quantile(tab$estimate, 0.25, na.rm = TRUE)),
                 med  = stats::median(tab$estimate, na.rm = TRUE),
                 mean = mean(tab$estimate, na.rm = TRUE),
                 q75  = unname(stats::quantile(tab$estimate, 0.75, na.rm = TRUE)),
                 max  = max(tab$estimate, na.rm = TRUE))

  top <- max(1L, min(as.integer(top), nrow(tab)))
  ord_plus  <- order(-tab$pipos, -tab$estimate)
  ord_minus <- order(-tab$pineg, -tab$estimate)

  res <- list(measure = measure, coef = coef, p_level = p_level,
              n_funds = nrow(tab), stats = est_stats, table = tab,
              top = utils::head(tab, top), bottom = utils::tail(tab, top),
              top_pi_plus  = utils::head(tab[ord_plus, , drop = FALSE], top),
              top_pi_minus = utils::head(tab[ord_minus, , drop = FALSE], top))
  class(res) <- "summary.SCREENING"
  res
}

#' @name print.summary.SCREENING
#' @title Print method for the 'summary.SCREENING' object
#' @description Print method for the object returned by
#' \code{\link{summary.SCREENING}}.
#' @param x A \code{summary.SCREENING} object.
#' @param ... Further arguments (ignored).
#' @return Invisibly returns \code{x}.
#' @author Murilo Andre Peres Pereira, David Ardia and Kris Boudt.
#' @keywords htest
#' @export
print.summary.SCREENING <- function(x, ...) {
  fmt <- function(z) formatC(as.numeric(z), format = "f", digits = 4)
  round_df <- function(df) {
    cols <- intersect(c("estimate", "pizero", "pipos", "pineg", "lambda"), names(df))
    for (cc in cols) df[[cc]] <- fmt(df[[cc]])
    df
  }
  cat("\nPeer performance screening summary (", x$measure, ")\n", sep = "")
  cat("  Funds: ", x$n_funds, "\n\n", sep = "")
  s <- x$stats
  # split over two lines: a single line overruns an 80-column console (and the
  # text block of a typeset document) once the measure name is prepended
  cat("  ", x$measure, " distribution:\n", sep = "")
  cat("    min ", fmt(s["min"]), " | 25% ", fmt(s["q25"]), " | med ", fmt(s["med"]),
      "\n    mean ", fmt(s["mean"]), " | 75% ", fmt(s["q75"]), " | max ",
      fmt(s["max"]), "\n\n", sep = "")
  cat("Top funds (by ", x$measure, "):\n", sep = "")
  print(round_df(x$top), row.names = FALSE, right = TRUE)
  cat("\nTop funds (by outperformance ratio pi+):\n")
  print(round_df(x$top_pi_plus), row.names = FALSE, right = TRUE)
  invisible(x)
}

#' @name confint.SCREENING
#' @title Bootstrap confidence intervals for the peer performance ratios
#' @description Computes confidence intervals for one of the peer performance
#' ratios (\eqn{\hat\pi^+}, \eqn{\hat\pi^0}, or \eqn{\hat\pi^-}) of a
#' \code{SCREENING} object, by a nonparametric peer (pairwise) bootstrap: for
#' each focal fund its peers are resampled with replacement and the ratio is
#' recomputed from the stored pairwise p-values and test statistics. The
#' interval therefore reflects the peer-composition and aggregation uncertainty,
#' holding the individual pairwise test statistics fixed.
#' @param object A \code{SCREENING} object (within-group or cross-group; not
#' \code{screen_beta}).
#' @param parm Which ratio to bound: \code{"pipos"} (default), \code{"pizero"},
#' or \code{"pineg"}.
#' @param level Confidence level. Default: \code{0.95}.
#' @param ... Unused.
#' @param nBoot Number of bootstrap resamples. Default: \code{499}.
#' @param gammaPos,gammaNeg One-sided thresholds used to recompute the ratios
#' (should match those of the original screening). Defaults: \code{0.4},
#' \code{0.6}.
#' @return A matrix with one row per fund and two columns (lower and upper
#' bounds); the point estimate is attached as attribute \code{"estimate"}.
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{alphaScreening}}.
#' @keywords htest
#' @examples
#' \donttest{
#' data("hfdata")
#' set.seed(1234)
#' sc <- alphaScreening(hfdata[, 1:15], control = list(nCore = 1))
#' confint(sc, parm = "pipos")
#' }
#' @export
#' @importFrom stats quantile
confint.SCREENING <- function(object, parm = c("pipos", "pizero", "pineg"),
                              level = 0.95, ..., nBoot = 499L,
                              gammaPos = 0.4, gammaNeg = 0.6) {
  parm <- match.arg(parm)
  est <- object[[parm]]
  if (is.matrix(est) || (!is.null(dim(est)) && length(dim(est)) > 1L)) {
    stop("confint is not supported for 'screen_beta' screenings")
  }
  pval  <- object$pval
  tstat <- object$tstat
  ddiff <- if (!is.null(object$dalpha)) object$dalpha
           else if (!is.null(object$dsharpe)) object$dsharpe
           else object$dmsharpe
  if (is.null(pval) || !is.matrix(pval) || is.null(tstat) || is.null(ddiff)) {
    stop("confint requires the pairwise 'pval'/'tstat'/difference matrices")
  }
  lambda <- object$lambda
  m <- nrow(pval)
  a <- (1 - level) / 2
  lo <- hi <- rep(NA_real_, m)
  for (i in seq_len(m)) {
    pv <- pval[i, ]; dd <- ddiff[i, ]; ts <- tstat[i, ]
    ok <- !is.na(pv) & !is.na(ts) & !is.na(dd)
    n <- sum(ok)
    if (n < 2L) next
    pv <- pv[ok]; dd <- dd[ok]; ts <- ts[ok]
    lam <- if (!is.null(lambda) && length(lambda) >= i && !is.na(lambda[i])) lambda[i] else 0.5
    bs <- rep(NA_real_, nBoot)
    for (b in seq_len(nBoot)) {
      idx <- sample.int(n, n, replace = TRUE)
      pp <- computePi(pv[idx], dd[idx], ts[idx], lambda = lam,
                      bpos = gammaPos, bneg = gammaNeg)
      bs[b] <- pp[[parm]]
    }
    qs <- stats::quantile(bs, c(a, 1 - a), na.rm = TRUE, names = FALSE)
    lo[i] <- qs[1]; hi[i] <- qs[2]
  }
  out <- cbind(lower = lo, upper = hi)
  colnames(out) <- paste0(formatC(100 * c(a, 1 - a), format = "g"), "%")
  rn <- names(est)
  if (is.null(rn)) rn <- names(object$n)
  rownames(out) <- if (!is.null(rn)) rn else seq_len(m)
  attr(out, "estimate") <- as.vector(est)
  out
}

#' @name exposureHeterogeneity
#' @title Factor exposure heterogeneity from a beta screening
#' @description Aggregates the per-fund equal-performance ratios of a
#' \code{screen_beta = TRUE} \code{\link{alphaScreening}} into the
#' factor-by-factor heterogeneity measure of Ardia et al. (2023). For each
#' coefficient \eqn{k} (the alpha and each factor beta) it reports the average
#' equal-exposure ratio \eqn{\pi^0_k = \frac1N\sum_i \pi^0_{i,k}} and the
#' heterogeneity \eqn{1-\pi^0_k}: the share of peers that are significantly
#' differentiated on coefficient \eqn{k}. A value close to one indicates large
#' heterogeneity (much room to differentiate); close to zero, homogeneity.
#' @param object A \code{SCREENING} object produced with
#' \code{screen_beta = TRUE}.
#' @return A \code{data.frame} of class \code{exposureHeterogeneity} with columns
#' \code{coefficient}, \code{equalExposure} (\eqn{\pi^0_k}), and
#' \code{heterogeneity} (\eqn{1-\pi^0_k}).
#' @author David Ardia and Kris Boudt.
#' @references
#' Ardia, D., Bluteau, K., Lortie-Cloutier, G., Tran, D. (2023).
#' Factor exposure heterogeneity in green and brown stocks.
#' \emph{Finance Research Letters} \bold{55}, Part A, 103900.
#' @seealso \code{\link{alphaScreening}}.
#' @keywords htest
#' @examples
#' \donttest{
#' data("hfdata")
#' set.seed(1234)
#' fac <- matrix(rnorm(nrow(hfdata) * 2), ncol = 2,
#'               dimnames = list(NULL, c("MKT", "SMB")))
#' sc <- alphaScreening(hfdata[, 1:20], factors = fac, screen_beta = TRUE,
#'                      control = list(nCore = 1))
#' exposureHeterogeneity(sc)
#' }
#' @export
exposureHeterogeneity <- function(object) {
  if (!inherits(object, "SCREENING")) {
    stop("'object' must be a 'SCREENING' object")
  }
  if (!is.matrix(object$pizero)) {
    stop("'exposureHeterogeneity' requires a screening run with screen_beta = TRUE")
  }
  pizero <- object$pizero
  cn <- rownames(pizero)
  if (is.null(cn)) cn <- paste0("coef", seq_len(nrow(pizero)))
  eq <- rowMeans(pizero, na.rm = TRUE)
  out <- data.frame(coefficient = cn,
                    equalExposure = as.vector(eq),
                    heterogeneity = as.vector(1 - eq),
                    stringsAsFactors = FALSE)
  rownames(out) <- NULL
  class(out) <- c("exposureHeterogeneity", "data.frame")
  out
}

#' @name plot.exposureHeterogeneity
#' @title Plot factor exposure heterogeneity
#' @description Bar plot of the heterogeneity measure \eqn{1-\pi^0_k} returned by
#' \code{\link{exposureHeterogeneity}}.
#' @param x An \code{exposureHeterogeneity} object.
#' @param ... Further arguments passed to \code{\link[graphics]{barplot}}.
#' @return Invisibly returns \code{x}.
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{exposureHeterogeneity}}.
#' @keywords hplot
#' @examples
#' \donttest{
#'   data("hfdata")
#'   sc <- alphaScreening(hfdata[, 1:10], factors = hfdata[, 50:51],
#'                        control = list(nCore = 1, screen_beta = TRUE))
#'   plot(exposureHeterogeneity(sc))
#' }
#' @export
#' @importFrom graphics barplot
plot.exposureHeterogeneity <- function(x, ...) {
  h <- x$heterogeneity
  names(h) <- x$coefficient
  do.call(graphics::barplot,
          .gpar(list(height = h, ylim = c(0, 1),
                     ylab = expression(1 - hat(pi)^0),
                     main = "Factor exposure heterogeneity"), list(...)))
  invisible(x)
}

#' @name plot.SCREENING
#' @title Peer performance screening plot
#' @description Plot method for the \code{SCREENING} object returned by
#' \code{\link{alphaScreening}}, \code{\link{sharpeScreening}} and
#' \code{\link{msharpeScreening}}. It reproduces the peer performance
#' screening plot of Ardia and Boudt (2018): funds are sorted by their
#' performance measure and, for each fund, the estimated outperformance
#' (\eqn{\hat\pi^+}), equal-performance (\eqn{\hat\pi^0}) and underperformance
#' (\eqn{\hat\pi^-}) ratios are displayed as a horizontal stacked bar. The
#' dashed diagonal lines depict the naive percentile-rank benchmark
#' (\eqn{\hat\pi^0 = 0}); the gap between the black \eqn{\hat\pi^+} area and the
#' diagonal visualizes the luck correction.
#' @param x A \code{SCREENING} object.
#' @param nblock Optional number of equally-sized blocks into which the funds
#' (sorted by performance) are aggregated by averaging, as in Ardia and Boudt
#' (2018). Default: \code{nblock = NULL}, i.e. one bar per fund.
#' @param reference A logical value indicating whether the dashed
#' percentile-rank reference lines should be drawn. Default:
#' \code{reference = NULL}, i.e. drawn for within-group screening and omitted
#' for cross-group (\code{Y}-based) screening, where the within-group
#' percentile-rank benchmark does not apply. A single focal fund (cross-group
#' with one fund in \code{X}) is shown as a single stacked bar.
#' @param band Half-width (in percentage points) of the reference band drawn
#' around the central diagonal. Default: \code{band = 27.5}.
#' @param colorset Vector of three colors for the \eqn{\hat\pi^+},
#' \eqn{\hat\pi^0} and \eqn{\hat\pi^-} areas. Default:
#' \code{c(gray(0), gray(0.8), gray(0.5))}.
#' @param ... Further graphical arguments passed to \code{\link[graphics]{barplot}}.
#' @return Invisibly returns the (sorted, possibly aggregated) matrix of
#' ratios that is plotted.
#' @details If the screening was run with \code{screen_beta = TRUE}, the alpha
#' (first) coefficient is used.
#' @author David Ardia and Kris Boudt.
#' @references
#' Ardia, D., Boudt, K. (2018).
#' The peer performance ratios of hedge funds.
#' \emph{Journal of Banking and Finance} \bold{87}, pp.351--368.
#' \doi{10.1016/j.jbankfin.2017.10.014}
#' @seealso \code{\link{alphaScreening}}, \code{\link{sharpeScreening}} and
#' \code{\link{msharpeScreening}}.
#' @keywords hplot
#' @examples
#' \donttest{
#' data("hfdata")
#' set.seed(1234)
#' sc <- alphaScreening(hfdata[, 1:30], control = list(nCore = 1))
#' plot(sc)
#' }
#' @export
#' @importFrom graphics barplot axis box grid par plot.default
#' @importFrom grDevices gray
plot.SCREENING <- function(x, nblock = NULL, reference = NULL, band = 27.5,
                           colorset = c(grDevices::gray(0), grDevices::gray(0.8),
                                        grDevices::gray(0.5)), ...) {
  cross <- isTRUE(x$cross)
  # the percentile-rank diagonal is a within-group benchmark: off for cross-group
  if (is.null(reference)) {
    reference <- !cross
  }
  perf  <- .screeningPerf(x)
  value <- perf$value
  pipos <- x$pipos; pizero <- x$pizero; pineg <- x$pineg

  # screen_beta = TRUE => matrices; use the alpha (first) coefficient
  if (is.matrix(pipos)) {
    pipos <- pipos[1, ]; pizero <- pizero[1, ]; pineg <- pineg[1, ]
  }
  if (is.matrix(value)) {
    value <- value[1, ]
  }
  value <- as.vector(value)

  # keep funds with complete information
  ok <- !is.na(pipos) & !is.na(pizero) & !is.na(pineg) & !is.na(value)
  value <- value[ok]; pipos <- pipos[ok]; pizero <- pizero[ok]; pineg <- pineg[ok]
  if (length(value) == 0) {
    stop("no fund with non-missing ratios to plot")
  }

  # a single focal fund (typically X-vs-Y with one fund): one stacked bar
  if (length(value) == 1) {
    return(.plotScreeningSingle(pipos, pizero, pineg, x$ny, colorset,
                                perf$label, ...))
  }

  # sort by performance (decreasing): best fund first (drawn at the bottom)
  ord <- order(value, decreasing = TRUE)
  value <- value[ord]; pipos <- pipos[ord]; pizero <- pizero[ord]; pineg <- pineg[ord]
  n <- length(value)

  # optional aggregation into 'nblock' equally-sized blocks
  if (!is.null(nblock) && nblock >= 1 && nblock < n) {
    grp <- ceiling(seq_len(n) / (n / nblock))
    grp[grp > nblock] <- nblock
    value <- as.vector(tapply(value, grp, mean))
    m <- cbind(as.vector(tapply(pipos,  grp, mean)),
               as.vector(tapply(pizero, grp, mean)),
               as.vector(tapply(pineg,  grp, mean)))
    m <- m / rowSums(m)  # renormalize so the triple sums to one
    pipos <- m[, 1]; pizero <- m[, 2]; pineg <- m[, 3]
    n <- length(value)
  }

  mat <- rbind(pipos, pizero, pineg)

  # two-panel layout: performance | peer performance ratios
  op <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(op))
  graphics::par(mfrow = c(1, 2))

  # left panel: performance measure, best fund at the bottom
  graphics::plot.default(value, seq_len(n), type = "b", pch = 20, las = 1,
                         axes = FALSE, xlab = "", ylab = "fund (sorted by performance)",
                         main = perf$label)
  graphics::box(); graphics::grid()
  graphics::axis(side = 1); graphics::axis(side = 2, las = 1)

  # right panel: stacked outperformance / equal / underperformance ratios
  mainstr <- expression(hat(pi)^"+" * " / " * hat(pi)^0 * " / " * hat(pi)^"-")
  do.call(graphics::barplot,
          .gpar(list(height = 100 * mat, horiz = TRUE, space = 0,
                     names.arg = NULL, col = colorset, border = NA,
                     xlim = c(0, 100), axes = FALSE, main = mainstr),
                list(...)))
  graphics::axis(side = 1, at = seq(0, 100, by = 25),
                 labels = paste0(seq(0, 100, by = 25), "%"))
  graphics::box()

  # dashed percentile-rank reference lines (naive benchmark, pi0 = 0)
  if (reference) {
    yy <- seq(0, 100, length.out = 200)
    xx <- seq(104, -4, length.out = 200)
    for (shift in c(0, -band, band)) {
      graphics::par(new = TRUE)
      graphics::plot.default(xx + shift, yy, type = "l", lty = "dashed", lwd = 1,
                             xlim = c(0, 100), ylim = c(0, 100), axes = FALSE,
                             xlab = "", ylab = "")
    }
  }

  invisible(t(mat))
}

## Set of R function for alpha screening

#@name .alphaScreening
#@description See alphaScreening
.alphaScreening <- function(X, factors = NULL, control = list(),
                            screen_beta = NULL, Y = NULL) {

  # process control
  ctr <- processControl(control)

  # screen_beta can be passed either as an argument (takes precedence, for
  # backward compatibility) or via the control list
  if (is.null(screen_beta)) {
    screen_beta <- ctr$screen_beta
  }
  # screen_beta only makes sense with factors; coerce to FALSE otherwise
  # (this avoids a downstream crash in infoFund() when factors is NULL)
  if (isTRUE(screen_beta) && is.null(factors)) {
    warning("'screen_beta = TRUE' requires 'factors'; it is ignored.")
    screen_beta <- FALSE
  }

  # cross-group screening: each fund in X against every fund in group Y
  if (!is.null(Y)) {
    return(.alphaScreeningXY(X, Y, factors = factors, control = control,
                             screen_beta = screen_beta))
  }

  X <- as.matrix(X)
  T <- nrow(X)
  N <- ncol(X)
  if (N < 2L) {
    stop("within-group screening needs at least two funds in 'X'; supply 'Y' to screen a single fund against a peer group")
  }

  if (screen_beta & !is.null(factors)) {
    row_return <- 1:(1 + ncol(factors))
    pval <- dalpha <- tstat <- array(rep(NA, N * N * (1 + ncol(factors))), 
                                     dim = c((1 + ncol(factors)), N, N))
  } else {
    row_return <- 1
    pval <- dalpha <- tstat <- array(rep(NA, N*N), dim = c(1, N, N))
  }
  # pval <- dalpha <- tstat <- matrix(data = NA, N, N)

  # determine which pairs can be compared (in a matrix way)
  Y <- 1 * (!is.nan(X) & !is.na(X))
  YY <- crossprod(Y)  #YY = t(Y) %*% Y # row i indicates how many observations in common with column k
  YY[YY < ctr$minObs] <- 0
  YY[YY > 0] <- 1
  liststocks <- c(1:nrow(YY))[rowSums(YY) > ctr$minObsPi]

  if (length(liststocks) > 1) {
    liststocks <- liststocks[1:(length(liststocks) - 1)]

    if (ctr$nCore == 1) {
      # serial path: no PSOCK cluster (avoids the per-call cluster overhead,
      # e.g. inside rollScreening's window loop)
      z <- lapply(as.list(liststocks), alphaScreeningi,
                  rdata = X, factors = factors, T = T, N = N,
                  hac = ctr$hac, screen_beta = screen_beta,
                  minObs = ctr$minObs)
    } else {
      cl <- parallel::makeCluster(ctr$nCore)
      on.exit(parallel::stopCluster(cl), add = TRUE)
      z <- parallel::clusterApplyLB(cl = cl, x = as.list(liststocks),
                                    fun = alphaScreeningi,
                                    rdata = X, factors = factors, T = T, N = N,
                                    hac = ctr$hac, screen_beta = screen_beta,
                                    minObs = ctr$minObs)
    }

    for (i in 1:length(liststocks)) {
      out <- z[[i]]
      id <- liststocks[i]
      pval[row_return, id, id:N] <- pval[row_return, id:N, id] <- out[[2]][row_return, id:N]
      dalpha[row_return, id, id:N] <- out[[1]][row_return, id:N]
      dalpha[row_return, id:N, id] <- -out[[1]][row_return, id:N]
      tstat[row_return, id, id:N] <- out[[3]][row_return, id:N]
      tstat[row_return, id:N, id] <- -out[[3]][row_return, id:N]
    }
  }

  # pi
  pi <- computePi(pval = pval, dalpha = dalpha, tstat = tstat, lambda = ctr$lambda,
                  nBoot = ctr$nBoot, bpos = ctr$gammaPos, bneg = ctr$gammaNeg,
                  fast = ctr$fastAdjust)


  # info on the funds
  info <- infoFund(X, factors = factors, screen_beta = screen_beta)

  if (screen_beta == FALSE) {
  	pval <- pval[1, , ]
  	dalpha <- dalpha[1, , ]
  	tstat <- tstat[1, , ]
  	npeer <- colSums(!is.na(pval))
  } else {
    npeer <- apply(!is.na(pval), c(1, 3), sum)
    # label the coefficient rows (alpha + factor betas)
    cn <- .coefNames(factors)
    rownames(pi$pizero) <- rownames(pi$pipos) <- rownames(pi$pineg) <- cn
    rownames(pi$lambda) <- cn
    rownames(info$alpha) <- cn
    rownames(npeer) <- cn
  }

  # form output
  out <- list(n = info$nObs, npeer = npeer, alpha = info$alpha,
              dalpha = dalpha, pval = pval, tstat = tstat, lambda = pi$lambda,
              pizero = pi$pizero, pipos = pi$pipos, pineg = pi$pineg)
  class(out) <- "SCREENING"

  return(out)
}

#' @name alphaScreening
#' @title Screening using the alpha outperformance ratio
#' @description Function which performs the screening of a universe of returns, and
#' computes the alpha outperformance ratio.
#' @details The alpha measure (Treynor and Black 1973, Carhart 1997, Fung and Hsieh
#' 2004) is one industry standard for measuring the absolute risk adjusted
#' performance of hedge funds. We propose to complement the alpha measure with
#' the fund's alpha outperformance ratio, defined as the percentage number of
#' funds that have a significantly lower alpha. In a pairwise testing
#' framework, a fund can have a significantly higher alpha because of luck. We
#' correct for this by applying the false discovery rate approach by Storey (2002).
#'
#' The methodology proceeds as follows:
#' \itemize{
#' \item (1) compute all pairwise tests of alpha differences. This means that for a universe of
#' \eqn{N} funds, we perform \eqn{N(N-1)/2}{N*(N-1)/2} tests. The algorithm has
#' been parallelized and the computational burden can be split across several
#' cores. The number of cores can be defined in \code{control}, see below.
#' \item (2) for each fund, the false discovery rate approach by Storey (2002)
#' is used to determine the proportions of over, equal, and underperforming
#' funds, in terms of alpha, in the database.}
#' The argument \code{control} is a list that can supply any of the following
#' components:
#' \itemize{
#' \item \code{'hac'} Heteroscedastic-autocorrelation consistent
#' standard errors. Default: \code{hac = FALSE}.
#' \item \code{'minObs'} Minimum number of concordant observations to compute the ratios. Default:
#' \code{minObs = 10}.
#' \item \code{'minObsPi'} Minimum number of observations
#' for computing the p-values). Default: \code{minObsPi = 1}.
#' \item \code{'nCore'} Number of cores used to perform the screening. Default:
#' \code{nCore = 1}.
#' \item \code{'lambda'} Threshold value to compute pi0.
#' Default: \code{lambda = NULL}, i.e. data driven choice.
#' \item \code{'gammaPos'} One-sided quantile level (of the standard Normal
#' distribution) used as the critical value for counting outperformed peers:
#' a peer counts as outperformed when the pairwise t-statistic exceeds
#' \code{qnorm(gammaPos)} (a \emph{negative} threshold for
#' \code{gammaPos < 0.5}; e.g., \code{qnorm(0.4)} is about \code{-0.25}), and
#' the expected fraction \code{1 - gammaPos} of false positives among the
#' equal-performing peers is then subtracted. Smaller values count more peers
#' before the correction. Default: \code{gammaPos = 0.4} (the value
#' recommended in Ardia and Boudt, 2018).
#' \item \code{'gammaNeg'} Mirror image of \code{gammaPos} for the peers that
#' outperform the focal fund: the count uses \code{tstat <= qnorm(gammaNeg)}
#' and subtracts the expected fraction \code{gammaNeg} of false positives.
#' Default: \code{gammaNeg = 0.6}.
#' \item \code{'fastAdjust'} Use a fast vectorised inversion in the
#' truncated-normal bias correction of \eqn{\pi^0}{pi0} instead of one
#' \code{uniroot} call per value. This is the dominant cost when \code{lambda}
#' is data driven and gives a large speed-up on big universes. The bisection
#' locates the root to about 1e-12; since \code{uniroot} stops at its own
#' tolerance (about 1.2e-4), the two paths typically differ by a few 1e-5, the
#' fast path being the more accurate. Default: \code{fastAdjust = FALSE},
#' i.e. the original code path, kept as default so that published results
#' reproduce exactly.
#' \item \code{'screen_beta'} Screen the factor exposures (betas) in addition to
#' the alpha; see the \code{screen_beta} argument. Default:
#' \code{screen_beta = FALSE}.
#' }
#' @param X Matrix \eqn{(T \times N)}{(TxN)} of \eqn{T} returns for the \eqn{N}
#' funds. \code{NA} values are allowed.
#' @param factors Matrix \eqn{(T \times K)}{(TxK)} of \eqn{T} returns for the
#' \eqn{K} factors. \code{NA} values are allowed.
#' @param control Control parameters (see *Details*).
#' @param screen_beta Boolean to screen all factors' coefficients (beta).
#' Default: \code{screen_beta = NULL}, in which case the value is taken from
#' \code{control$screen_beta} (itself defaulting to \code{FALSE}, i.e. only the
#' alpha is screened). When supplied directly, the argument takes precedence
#' over the control list. If \code{TRUE}, each element of the returned list will
#' have a new first dimension representing each coefficient (the first one being
#' alpha).
#' @param Y Optional matrix \eqn{(T \times M)}{(TxM)} of returns for a second
#' (peer) group of \eqn{M} funds. When supplied, the ratios are computed for
#' each fund in \code{X} \emph{against the funds in \code{Y}} (cross-group
#' screening) instead of against the other funds in \code{X}. A single focal
#' fund versus a peer group corresponds to \code{X} being a vector (or a
#' one-column matrix). Columns of \code{Y} identical to the focal fund (e.g.
#' when \code{X} is a subset of \code{Y}) are automatically excluded. Default:
#' \code{Y = NULL}, i.e. within-group screening.
#' @return A list with the following components:\cr
#'
#' \code{n}: Vector (of length \eqn{N}) of number of non-\code{NA}
#' observations.\cr
#'
#' \code{npeer}: Vector (of length \eqn{N}) of number of available peers.\cr
#'
#' \code{alpha}: Vector (of length \eqn{N}) of unconditional alpha.\cr
#'
#' \code{dalpha}: Matrix (of size \eqn{N \times N}{NxN}) of alpha
#' differences.\cr
#'
#' \code{tstat}: Matrix (of size \eqn{N \times N}{NxN}) of t-statistics.\cr
#'
#' \code{pval}: Matrix (of size \eqn{N \times N}) of p-values of test for alpha
#' differences.\cr
#'
#' \code{lambda}: Vector (of length \eqn{N}) of lambda values.\cr
#'
#' \code{pizero}: Vector (of length \eqn{N}) of probability of equal
#' performance.\cr
#'
#' \code{pipos}: Vector (of length \eqn{N}) of probability of outperformance
#' performance.\cr
#'
#' \code{pineg}: Vector (of length \eqn{N}) of probability of underperformance
#' performance.
#' @note Further details on the methodology with an application to the hedge
#' fund industry is given in Ardia and Boudt (2018).
#'
#' Application of the false discovery rate approach applied to the mutual fund
#' industry has been presented in Barras, Scaillet and Wermers (2010).
#'
#' HAC standard errors are available via \code{control = list(hac = TRUE)}
#' (computed with \pkg{sandwich}/\pkg{lmtest}). The studentized circular block
#' bootstrap of Ledoit and Wolf (2008) applies to the Sharpe-ratio routines
#' (\code{\link{sharpeScreening}}, \code{\link{msharpeScreening}}) and is not
#' used by \code{alphaScreening}.
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{sharpeScreening}} and \code{\link{msharpeScreening}}.
#' @references
#' Ardia, D., Boudt, K. (2015).
#' Testing equality of modified Sharpe ratios.
#' \emph{Finance Research Letters} \bold{13}, pp.97--104.
#' \doi{10.1016/j.frl.2015.02.008}
#'
#' Ardia, D., Boudt, K. (2018).
#' The peer performance ratios of hedge funds.
#' \emph{Journal of Banking and Finance} \bold{87}, pp.351--368.
#' \doi{10.1016/j.jbankfin.2017.10.014}
#'
#' Barras, L., Scaillet, O., Wermers, R. (2010).
#' False discoveries in mutual fund performance: Measuring luck in estimated alphas.
#' \emph{Journal of Finance} \bold{65}(1), pp.179--216.
#'
#' Carhart, M. (1997).
#' On persistence in mutual fund performance.
#' \emph{Journal of Finance} \bold{52}(1), pp.57--82.
#'
#' Fama, E., French, K. (2010).
#' Luck versus skill in the cross-section of mutual fund returns.
#' \emph{Journal of Finance} \bold{65}(5), pp.1915--1947.
#'
#' Fung, W., Hsieh, D. (2004).
#' Hedge fund benchmarks: A risk based approach.
#' \emph{Financial Analysts Journal} \bold{60}(5), pp.65--80.
#'
#' Storey, J. (2002).
#' A direct approach to false discovery rates.
#' \emph{Journal of the Royal Statistical Society B} \bold{64}(3), pp.479--498.
#'
#' Treynor, J. L., Black, F. (1973).
#' How to use security analysis to improve portfolio selection.
#' \emph{Journal of Business} \bold{46}(1), pp.66--86.
#' @keywords htest
#' @examples
#' ## Load the data (randomized data of monthly hedge fund returns)
#' data("hfdata")
#' rets = hfdata[,1:4]
#'
#' ## Run alpha screening
#' ctr = list(nCore = 1)
#' alphaScreening(rets, control = ctr)
#'
#' ## Run alpha screening with HAC standard deviation
#' ctr = list(nCore = 1, hac = TRUE)
#' alphaScreening(rets, control = ctr)
#'
#' ## Cross-group screening: a single focal fund against a peer group
#' alphaScreening(hfdata[, 1], Y = hfdata[, 11:20], control = list(nCore = 1))
#'
#' ## Cross-group screening: peer group X against peer group Y
#' alphaScreening(hfdata[, 1:5], Y = hfdata[, 11:20], control = list(nCore = 1))
#' @export
#' @importFrom parallel makeCluster clusterApplyLB stopCluster
#' @importFrom compiler cmpfun
alphaScreening <- compiler::cmpfun(.alphaScreening)

# #' @name .alphaScreeningi
# #' @title Screening for fund i again its peers
# #' @importFrom stats lm na.omit
# #' @importFrom lmtest coeftest
# #' @importFrom sandwich vcovHAC
.alphaScreeningi <- function(i, rdata, factors, T, N, hac, screen_beta=FALSE,
                             minObs = 10) {


  if(screen_beta & !is.null(factors)){
    row_return <- 1:(1+ncol(factors))
    pvali <- dalphai <- tstati <- matrix(rep(NA, N*(1+ncol(factors))), ncol=N)
  }else{
    row_return <- 1
    pvali <- dalphai <- tstati <- matrix(rep(NA, N), ncol=N)
  }

  nPeer <- N - i
  X <- matrix(rdata[, i], nrow = T, ncol = nPeer)
  Y <- matrix(rdata[, (i + 1):N], nrow = T, ncol = nPeer)
  dXY <- X - Y

  # Per-pair complete-case availability. Count only rows where the return
  # differential AND the factors (when supplied) are observed, so a pair is
  # tested only when at least 'minObs' usable observations remain. (Previously
  # the factor NAs were ignored and 'minObs' was not enforced at the pair level.)
  if (is.null(factors)) {
    avail <- colSums(!is.na(dXY))
  } else {
    fok   <- stats::complete.cases(factors)          # length T, recycled per column
    avail <- colSums(!is.na(dXY) & fok)
  }
  selId.in  <- which(avail >= minObs)
  selId.out <- selId.in + i

  # k selects the column in dXY (matches selId.in); j places the result in the
  # full N-vector (matches selId.out). seq_along guards the empty case.
  for (idx in seq_along(selId.in)) {
    k <- selId.in[idx]
    j <- selId.out[idx]

    if (is.null(factors)) {
      fit <- stats::lm(dXY[, k] ~ 1, na.action = stats::na.omit)
    } else {
      fit <- stats::lm(dXY[, k] ~ 1 + factors, na.action = stats::na.omit)
    } # end of factors/no factors

    # skip (near) deterministic differentials (zero residual variance)
    sfit_lm <- summary(fit)
    if (!is.finite(sfit_lm$sigma) || sfit_lm$sigma < sqrt(.Machine$double.eps)) {
      next
    }

    # HAC within loop.
    if (!hac) {
      pvali[row_return, j] <- sfit_lm$coef[row_return, 4]
      dalphai[row_return, j] <- sfit_lm$coef[row_return, 1]
      tstati[row_return, j] <- sfit_lm$coef[row_return, 3]
    } else {
      sumfit <- lmtest::coeftest(fit, vcov. = sandwich::vcovHAC(fit))
      pvali[row_return, j] <- sumfit[row_return, 4]
      dalphai[row_return, j] <- sumfit[row_return, 1]
      tstati[row_return, j] <- sumfit[row_return, 3]
    }
  }

  out <- list(dalphai = dalphai, pvali = pvali, tstati = tstati)
  return(out)
}
alphaScreeningi <- compiler::cmpfun(.alphaScreeningi)

## Set of R functions for the modified Sharpe ratio testing

# #' @name .msharpeTesting
# #' @import compiler
.msharpeTesting <- function(x, y, level = 0.9, na.neg = TRUE, control = list()) {

  x <- as.matrix(x)
  y <- as.matrix(y)

  # process control parameters
  ctr <- processControl(control)

  # check if enough data are available for testing
  dxy <- x - y
  idx <- (!is.nan(dxy) & !is.na(dxy))
  rets <- cbind(x[idx], y[idx])
  T <- sum(idx)
  if (T < ctr$minObs) {
    stop("intersection of 'x' and 'y' is shorter than 'minObs'")
  }

  # msharpe testing
  if (ctr$type == 1) {
    # ==> asymptotic approach
    tmp <- msharpeTestAsymptotic(rets, level, na.neg, ctr$hac, ctr$ttype)
  } else {
    # ==> bootstrap approach (iid and circular block bootstrap)
    if (ctr$bBoot == 0) {
      ctr$bBoot <- msharpeBlockSize(x, y, level, na.neg, ctr)
    }
    bsids <- bootIndices(T, ctr$nBoot, ctr$bBoot)
    tmp <- msharpeTestBootstrap(rets, level, na.neg, bsids, ctr$bBoot,
                                ctr$ttype, ctr$pBoot)
  }

  # info on the funds
  info <- infoFund(rets, level = level, na.neg = na.neg)

  ## form output
  out <- list(n = T, msharpe = info$msharpe, dmsharpe = -diff(info$msharpe),
              tstat = as.vector(tmp$tstat), pval = as.vector(tmp$pval))
  class(out) <- "TESTING"
  return(out)

}

#' @name msharpeTesting
#' @title Testing the difference of modified Sharpe ratios
#' @description Function which performs the testing of the difference of modified Sharpe
#' ratios.
#' @details The modified Sharpe ratio (Favre and Galeano 2002) is one industry
#' standard for measuring the absolute risk adjusted performance of hedge
#' funds. This function performs the testing of modified Sharpe ratio
#' difference for two funds using a similar approach than Ledoit and Wolf
#' (2002). See also Gregoriou and Gueyie (2003).
#'
#' For the testing, only the intersection of non-\code{NA} observations for the
#' two funds are used.
#'
#' The argument \code{control} is a list that can supply any of the following
#' components:
#' \itemize{
#' \item \code{'type'} Asymptotic approach (\code{type = 1}) or
#' studentized circular bootstrap approach (\code{type = 2}). Default:
#' \code{type = 1}.
#' \item \code{'ttype'} Test based on ratio (\code{type = 1})
#' or product (\code{type = 2}). Default: \code{type = 2}.
#' \item \code{'hac'} Heteroscedastic-autocorrelation consistent standard
#' errors. Default: \code{hac = FALSE}.
#' \item \code{'minObs'} Minimum number of concordant observations to compute the ratios. Default: \code{minObs =
#' 10}.
#' \item \code{'nBoot'} Number of bootstrap replications for computing the
#' p-value. Default: \code{nBoot = 499}.
#' \item \code{'bBoot'} Block length in
#' the circular bootstrap. Default: \code{bBoot = 1}, i.e. iid bootstrap.
#' \code{bBoot = 0} uses optimal block-length.
#' \item \code{'pBoot'} Symmetric
#' p-value (\code{pBoot = 1}) or asymmetric p-value (\code{pBoot = 2}).
#' Default: \code{pBoot = 1}.
#' }
#' @param x Vector (of length \eqn{T}) of returns for the first fund. \code{NA}
#' values are allowed.
#' @param y Vector (of length \eqn{T}) of returns for the second fund. \code{NA}
#' values are allowed.
#' @param level Modified Value-at-Risk level. Default: \code{level = 0.90}.
#' @param na.neg A logical value indicating whether \code{NA} values should be
#' returned if a negative modified Value-at-Risk is obtained.  Default
#' \code{na.neg = TRUE}.
#' @param control Control parameters (see *Details*).
#' @return A list with the following components:\cr
#'
#' \code{n}: Number of non-\code{NA} concordant observations.\cr
#'
#' \code{msharpe}: Vector (of length 2) of unconditional modified Sharpe
#' ratios.\cr
#'
#' \code{dmsharpe}: Modified Sharpe ratios difference.\cr
#'
#' \code{tstat}: t-stat of modified Sharpe ratios differences.\cr
#'
#' \code{pval}: pvalues of test of modified Sharpe ratios differences.
#' @note Further details on the methodology with an application to the hedge
#' fund industry is given in Ardia and Boudt (2018).
#'
#' Some internal functions where adapted from Michael Wolf MATLAB code.
#'
#' @author David Ardia and Kris Boudt.
#' @seealso \code{\link{msharpe}}, \code{\link{msharpeScreening}} and
#' \code{\link{sharpeTesting}}.
#' @references
#' Ardia, D., Boudt, K. (2015).
#' Testing equality of modified Sharpe ratios.
#' \emph{Finance Research Letters} \bold{13}, pp.97--104.
#' \doi{10.1016/j.frl.2015.02.008}
#'
#' Ardia, D., Boudt, K. (2018).
#' The peer performance ratios of hedge funds.
#' \emph{Journal of Banking and Finance} \bold{87}, pp.351--368.
#' \doi{10.1016/j.jbankfin.2017.10.014}
#'
#' Barras, L., Scaillet, O., Wermers, R. (2010).
#' False discoveries in mutual fund performance: Measuring luck in estimated alphas.
#' \emph{Journal of Finance} \bold{65}(1), pp.179--216.
#'
#' Favre, L., Galeano, J.A. (2002).
#' Mean-modified Value-at-Risk Optimization with Hedge Funds.
#' \emph{Journal of Alternative Investments} \bold{5}(2), pp.21--25.
#'
#' Gregoriou, G. N., Gueyie, J.-P. (2003).
#' Risk-adjusted performance of funds of hedge funds using a modified Sharpe ratio.
#' \emph{Journal of Wealth Management} \bold{6}(3), pp.77--83.
#'
#' Ledoit, O., Wolf, M. (2008).
#' Robust performance hypothesis testing with the Sharpe ratio.
#' \emph{Journal of Empirical Finance} \bold{15}(5), pp.850--859.
#'
#' Storey, J. (2002).
#' A direct approach to false discovery rates.
#' \emph{Journal of the Royal Statistical Society B} \bold{64}(3), pp.479--498.
#' @keywords htest
#' @examples
#' ## Load the data (randomized data of monthly hedge fund returns)
#' data("hfdata")
#' x = hfdata[,1]
#' y = hfdata[,2]
#'
#' ## Run modified Sharpe testing (asymptotic)
#' ctr = list(type = 1)
#' out = msharpeTesting(x, y, level = 0.95, control = ctr)
#' print(out)
#'
#' ## Run modified Sharpe testing (asymptotic hac)
#' ctr = list(type = 1, hac = TRUE)
#' out = msharpeTesting(x, y, level = 0.95, control = ctr)
#' print(out)
#'
#' ## Run modified Sharpe testing (iid bootstrap)
#' set.seed(1234)
#' ctr = list(type = 2, nBoot = 250)
#' out = msharpeTesting(x, y, level = 0.95, control = ctr)
#' print(out)
#'
#' ## Run modified Sharpe testing (circular bootstrap)
#' set.seed(1234)
#' ctr = list(type = 2, nBoot = 250, bBoot = 5)
#' out = msharpeTesting(x, y, level = 0.95, control = ctr)
#' print(out)
#' @export
#' @importFrom compiler cmpfun
msharpeTesting <- compiler::cmpfun(.msharpeTesting)

# #' @name .msharpe.ratio.diff
# #' @title Difference of sharpe ratios
# #' @importFrom stats qnorm
# #' @import compiler
.msharpe.ratio.diff <- function(X, Y = NULL, level, na.neg, ttype) {
  if (is.null(Y)) {
    Y <- X[, 2, drop = FALSE]
    X <- X[, 1, drop = FALSE]
  }
  m1X <- colMeans(X)
  m1Y <- colMeans(Y)
  X_ <- sweep(x = X, MARGIN = 2, STATS = m1X, FUN = "-")
  Y_ <- sweep(x = Y, MARGIN = 2, STATS = m1Y, FUN = "-")
  m2X <- colMeans(X_^2)
  m2Y <- colMeans(Y_^2)
  m3X <- colMeans(X_^3)
  m3Y <- colMeans(Y_^3)
  m4X <- colMeans(X_^4)
  m4Y <- colMeans(Y_^4)
  za <- stats::qnorm(1 - level)
  skewX <- m3X/m2X^(3/2)
  skewY <- m3Y/m2Y^(3/2)
  kurtX <- (m4X/m2X^2) - 3
  kurtY <- (m4Y/m2Y^2) - 3
  mVaRX <- -m1X + sqrt(m2X) * (-za - (1/6) * (za^2 - 1) * skewX - (1/24) *
                                 (za^3 - 3 * za) * kurtX + (1/36) * (2 * za^3 - 5 * za) * skewX^2)
  mVaRY <- -m1Y + sqrt(m2Y) * (-za - (1/6) * (za^2 - 1) * skewY - (1/24) *
                                 (za^3 - 3 * za) * kurtY + (1/36) * (2 * za^3 - 5 * za) * skewY^2)
  if (na.neg) {
    mVaRX[mVaRX < 0] <- NA
    mVaRY[mVaRY < 0] <- NA
  }
  if (ttype == 1) {
    # test based on quotient
    mSR1 <- m1X/mVaRX
    mSR2 <- m1Y/mVaRY
  } else {
    # test based on product
    mSR1 <- m1X * mVaRY
    mSR2 <- m1Y * mVaRX
  }
  diff <- mSR1 - mSR2
  return(diff)
}
msharpe.ratio.diff <- compiler::cmpfun(.msharpe.ratio.diff)

# #' @name .msharpeTestAsymptotic
# #' @title Asymptotic Sharpe testing
# #' @importFrom stats pnorm
# #' @import compiler
.msharpeTestAsymptotic <- function(rets, level, na.neg, hac, ttype) {

  dmsharpe <- msharpe.ratio.diff(rets, Y = NULL, level, na.neg, ttype)
  if (is.na(dmsharpe)) {
    out <- list(dmsharpe = NA, tstat = NA, se = NA, pval = NA)
    return(out)
  }
  se <- se.msharpe.asymptotic(rets, level, hac, ttype)
  tstat <- dmsharpe/se
  pval <- 2 * stats::pnorm(-abs(tstat))  # asymptotic normal p-value
  out <- list(dmsharpe = dmsharpe, tstat = tstat, se = se, pval = pval)
  return(out)
}
msharpeTestAsymptotic <- compiler::cmpfun(.msharpeTestAsymptotic)

# #' @name .se.msharpe.asymptotic
# #' @title Asymptotic standard error
# #' @importFrom stats cov ar qnorm
# #' @import compiler
.se.msharpe.asymptotic <- function(X, level, hac, ttype) {

  # estimation of (robust) Psi function; see Ledoit Wolf paper
  compute.Psi.hat <- function(V.hat, hac) {
    if (hac) {
      T <- length(V.hat[, 1])
      alpha.hat <- compute.alpha.hat(V.hat)
      S.star <- min(2.6614 * (alpha.hat * T)^0.2, T)  # DA fix here to avoid >T values
      Psi.hat <- compute.Gamma.hat(V.hat, 0)
      j <- 1
      while (j < S.star) {
        Gamma.hat <- compute.Gamma.hat(V.hat, j)
        Psi.hat <- Psi.hat + kernel.Parzen(j/S.star) * (Gamma.hat +
                                                          t(Gamma.hat))
        j <- j + 1
      }
      Psi.hat <- (T/(T - 4)) * Psi.hat
    } else {
      Psi.hat <- stats::cov(V.hat)
    }
    return(Psi.hat)

  }

  # Parzen kernel
  kernel.Parzen <- function(x) {
    if (abs(x) <= 0.5)
      result <- 1 - 6 * x^2 + 6 * abs(x)^3 else if (abs(x) <= 1)
        result <- 2 * (1 - abs(x))^3 else result <- 0
        return(result)
  }

  compute.alpha.hat <- function(V.hat) {
    p <- ncol(V.hat)
    num <- den <- 0
    for (i in 1:p) {
      fit <- stats::ar(V.hat[, i], 0, 1, method = "ols")
      rho.hat <- as.numeric(fit[2])
      sig.hat <- sqrt(as.numeric(fit[3]))
      num <- num + 4 * rho.hat^2 * sig.hat^4/(1 - rho.hat)^8
      den <- den + sig.hat^4/(1 - rho.hat)^4
    }
    return(num/den)

  }

  compute.Gamma.hat <- function(V.hat, j) {
    T <- nrow(V.hat)
    p <- ncol(V.hat)
    Gamma.hat <- matrix(0, p, p)
    if (j >= T)
      stop("j must be smaller than the row dimension!")
    for (i in ((j + 1):T)) {
      Gamma.hat <- Gamma.hat + tcrossprod(V.hat[i, ], V.hat[i - j,
                                                            ])
    }
    Gamma.hat <- Gamma.hat/T
    return(Gamma.hat)

  }

  T <- nrow(X)
  m1 <- colMeans(X)
  X_ <- sweep(x = X, MARGIN = 2, STATS = m1, FUN = "-")
  m2 <- colMeans(X_^2)
  m3 <- colMeans(X_^3)
  m4 <- colMeans(X_^4)
  g2 <- m2 + m1^2
  g3 <- m3 + 3 * m1 * g2 - 2 * m1^3
  g4 <- m4 + 4 * m1 * g3 - 6 * m1^2 * g2 + 3 * m1^4
  skew <- m3/m2^(3/2)
  kurt <- (m4/m2^2) - 3

  # gradient for underlying moments
  dm1i <- c(1, 0, 0, 0, 0, 0, 0, 0)
  dm1j <- c(0, 0, 0, 0, 1, 0, 0, 0)
  dm2i <- c(-2 * m1[1], 1, 0, 0, 0, 0, 0, 0)
  dm2j <- c(0, 0, 0, 0, -2 * m1[2], 1, 0, 0)
  dm3i <- c(-3 * g2[1] + 6 * m1[1]^2, -3 * m1[1], 1, 0, 0, 0, 0, 0)
  dm3j <- c(0, 0, 0, 0, -3 * g2[2] + 6 * m1[2]^2, -3 * m1[2], 1, 0)
  dm4i <- c(-4 * g3[1] + 12 * m1[1] * g2[1] - 12 * m1[1]^3, 6 * m1[1]^2,
            -4 * m1[1], 1, 0, 0, 0, 0)
  dm4j <- c(0, 0, 0, 0, -4 * g3[2] + 12 * m1[2] * g2[2] - 12 * m1[2]^3,
            6 * m1[2]^2, -4 * m1[2], 1)

  dsi <- (m2[1]^(3/2) * dm3i - 1.5 * m3[1] * m2[1]^(1/2) * dm2i)/m2[1]^3
  dsj <- (m2[2]^(3/2) * dm3j - 1.5 * m3[2] * m2[2]^(1/2) * dm2j)/m2[2]^3

  dki <- (m2[1]^2 * dm4i - 2 * m4[1] * m2[1] * dm2i)/m2[1]^4
  dkj <- (m2[2]^2 * dm4j - 2 * m4[2] * m2[2] * dm2j)/m2[2]^4

  za <- qnorm(1 - level)

  # constants for speedup
  cst1i <- 1/(2 * sqrt(m2[1]))
  cst1j <- 1/(2 * sqrt(m2[2]))
  cst2i <- sqrt(m2[1])
  cst2j <- sqrt(m2[2])

  # dmVaRi
  dmVaRi <- -dm1i - cst1i * dm2i * za
  dmVaRi <- dmVaRi - (1/6) * (za^2 - 1) * (cst1i * skew[1] * dm2i + cst2i *
                                             dsi)
  dmVaRi <- dmVaRi + (1/36) * (2 * za^3 - 5 * za) * (cst1i * skew[1]^2 *
                                                       dm2i + 2 * cst2i * skew[1] * dsi)
  dmVaRi <- dmVaRi - (1/24) * (za^3 - 3 * za) * (cst1i * kurt[1] * dm2i +
                                                   cst2i * dki)

  # dmVaRj
  dmVaRj <- -dm1j - cst1j * dm2j * za
  dmVaRj <- dmVaRj - (1/6) * (za^2 - 1) * (cst1j * skew[2] * dm2j + cst2j *
                                             dsj)
  dmVaRj <- dmVaRj + (1/36) * (2 * za^3 - 5 * za) * (cst1j * skew[2]^2 *
                                                       dm2j + 2 * cst2j * skew[2] * dsj)
  dmVaRj <- dmVaRj - (1/24) * (za^3 - 3 * za) * (cst1j * kurt[2] * dm2j +
                                                   cst2j * dkj)

  # mVaR
  mVaR <- -m1 + sqrt(m2) * (-za - (1/6) * (za^2 - 1) * skew - (1/24) *
                              (za^3 - 3 * za) * kurt + (1/36) * (2 * za^3 - 5 * za) * skew^2)

  # gradient
  if (ttype == 1) {
    # test based on quotient
    tmp1 <- (mVaR[1] * dm1i - m1[1] * dmVaRi)/mVaR[1]^2
    tmp2 <- (mVaR[2] * dm1j - m1[2] * dmVaRj)/mVaR[2]^2
  } else {
    # test based on product
    tmp1 <- (mVaR[2] * dm1i + m1[1] * dmVaRj)
    tmp2 <- (mVaR[1] * dm1j + m1[2] * dmVaRi)
  }

  gradient <- tmp1 - tmp2
  V.hat <- matrix(NA, T, 8)
  V.hat[, c(1, 5)] <- sweep(x = X, MARGIN = 2, STATS = m1, FUN = "-")
  V.hat[, c(2, 6)] <- sweep(x = X^2, MARGIN = 2, STATS = g2, FUN = "-")
  V.hat[, c(3, 7)] <- sweep(x = X^3, MARGIN = 2, STATS = g3, FUN = "-")
  V.hat[, c(4, 8)] <- sweep(x = X^4, MARGIN = 2, STATS = g4, FUN = "-")
  Psi.hat <- compute.Psi.hat(V.hat, hac)
  se <- as.numeric(sqrt(crossprod(gradient, Psi.hat %*% gradient)/T))
  return(se)
}
se.msharpe.asymptotic <- compiler::cmpfun(.se.msharpe.asymptotic)

# #' @name .msharpeTestBootstrap
# #' @import compiler
.msharpeTestBootstrap <- function(rets, level, na.neg, bsids, b, ttype,
                                  pBoot, d = 0) {

  T <- nrow(rets)
  x <- rets[, 1, drop = FALSE]
  y <- rets[, 2, drop = FALSE]

  dmsharpe <- as.numeric(msharpe.ratio.diff(x, y, level, na.neg, ttype) -
                           d)
  if (is.na(dmsharpe)) {
    out <- list(dmsharpe = NA, tstat = NA, se = NA, bststat = NA, pval = NA)
    return(out)
  }
  se <- se.msharpe.bootstrap(x, y, level, b, ttype)
  # se = se.msharpe.asymptotic(X = cbind(x, y), level = level, hac =
  # TRUE, ttype = ttype)

  # bootstrap indices
  nBoot <- ncol(bsids)
  bsidx <- 1 + bsids%%T  # ensure that the bootstrap indices match the length of the time series
  bsX <- matrix(x[bsidx], T, nBoot)
  bsY <- matrix(y[bsidx], T, nBoot)

  bsdmsharpe <- msharpe.ratio.diff(bsX, bsY, level, na.neg = FALSE, ttype)  # DA consider negative mVaR as well in the bootstrap
  bsse <- se.msharpe.bootstrap(bsX, bsY, level, b, ttype)
  tstat <- dmsharpe/se

  if (pBoot == 1) {
    # first type p-value calculation
    bststat <- abs(bsdmsharpe - dmsharpe)/bsse
    pval <- (sum(bststat >= abs(tstat)) + 1)/(nBoot + 1)
    # pval = sum(bststat >= abs(tstat)) / nBoot
  } else {
    # second type p-value calculation (as in Barras)
    bststat <- (bsdmsharpe - dmsharpe)/bsse
    pval <- 2 * min(sum(bststat > tstat) + 1, sum(bststat < tstat) +
                      1)/(nBoot + 1)
    # pval = 2 * min(sum(bststat > tstat), sum(bststat < tstat)) / nBoot
  }

  out <- list(dmsharpe = dmsharpe, tstat = tstat, se = se, bststat = bststat,
              pval = pval)
  return(out)
}
msharpeTestBootstrap <- compiler::cmpfun(.msharpeTestBootstrap)

# #' @name .se.msharpe.bootstrap
# #' @importFrom stats cov qnorm
# #' @import compiler
.se.msharpe.bootstrap <- function(X, Y, level, b, ttype) {

  ## Compute Psi with two approaches: 1) iid bootstrap, 2) circular block
  ## bootstrap
  compute.Psi.hat <- function(V.hat, b) {
    T <- length(V.hat[, 1])
    if (b == 1) {
      # ==> standard estimation
      Psi.hat <- stats::cov(V.hat)
    } else {
      # ==> block estimation
      l <- floor(T/b)
      Psi.hat <- matrix(0, 8, 8)
      for (j in (1:l)) {
        zeta <- b^0.5 * colMeans(V.hat[((j - 1) * b + 1):(j * b),
                                       , drop = FALSE])
        Psi.hat <- Psi.hat + tcrossprod(zeta)
      }
      Psi.hat <- Psi.hat/l
    }
    return(Psi.hat)
  }

  T <- nrow(X)
  N <- ncol(Y)
  m1X <- colMeans(X)
  m1Y <- colMeans(Y)
  X_ <- sweep(x = X, MARGIN = 2, STATS = m1X, FUN = "-")
  Y_ <- sweep(x = Y, MARGIN = 2, STATS = m1Y, FUN = "-")
  m2X <- colMeans(X_^2)
  m2Y <- colMeans(Y_^2)
  m3X <- colMeans(X_^3)
  m3Y <- colMeans(Y_^3)
  m4X <- colMeans(X_^4)
  m4Y <- colMeans(Y_^4)
  g2X <- m2X + m1X^2
  g2Y <- m2Y + m1Y^2
  g3X <- m3X + 3 * m1X * g2X - 2 * m1X^3
  g3Y <- m3Y + 3 * m1Y * g2Y - 2 * m1Y^3
  g4X <- m4X + 4 * m1X * g3X - 6 * m1X^2 * g2X + 3 * m1X^4
  g4Y <- m4Y + 4 * m1Y * g3Y - 6 * m1Y^2 * g2Y + 3 * m1Y^4

  dm1X <- matrix(rep(c(1, 0, 0, 0, 0, 0, 0, 0), N), 8, N, FALSE)
  dm1Y <- matrix(rep(c(0, 0, 0, 0, 1, 0, 0, 0), N), 8, N, FALSE)
  dm2X <- rbind(-2 * m1X, 1, 0, 0, 0, 0, 0, 0)
  dm2Y <- rbind(0, 0, 0, 0, -2 * m1Y, 1, 0, 0)
  dm3X <- rbind(-3 * g2X + 6 * m1X^2, -3 * m1X, 1, 0, 0, 0, 0, 0)
  dm3Y <- rbind(0, 0, 0, 0, -3 * g2Y + 6 * m1Y^2, -3 * m1Y, 1, 0)
  dm4X <- rbind(-4 * g3X + 12 * m1X * g2X - 12 * m1X^3, 6 * m1X^2, -4 *
                  m1X, 1, 0, 0, 0, 0)
  dm4Y <- rbind(0, 0, 0, 0, -4 * g3Y + 12 * m1Y * g2Y - 12 * m1Y^3, 6 *
                  m1Y^2, -4 * m1Y, 1)

  # matrix form
  m1X_ <- matrix(m1X, nrow = 8, ncol = N, byrow = TRUE)
  m1Y_ <- matrix(m1Y, nrow = 8, ncol = N, byrow = TRUE)
  m2X_ <- matrix(m2X, nrow = 8, ncol = N, byrow = TRUE)
  m2Y_ <- matrix(m2Y, nrow = 8, ncol = N, byrow = TRUE)
  m3X_ <- matrix(m3X, nrow = 8, ncol = N, byrow = TRUE)
  m3Y_ <- matrix(m3Y, nrow = 8, ncol = N, byrow = TRUE)
  m4X_ <- matrix(m4X, nrow = 8, ncol = N, byrow = TRUE)
  m4Y_ <- matrix(m4Y, nrow = 8, ncol = N, byrow = TRUE)

  dm1X_ <- matrix(dm1X, nrow = 8, ncol = N, byrow = FALSE)
  dm1Y_ <- matrix(dm1Y, nrow = 8, ncol = N, byrow = FALSE)
  dm2X_ <- matrix(dm2X, nrow = 8, ncol = N, byrow = FALSE)
  dm2Y_ <- matrix(dm2Y, nrow = 8, ncol = N, byrow = FALSE)
  dm3X_ <- matrix(dm3X, nrow = 8, ncol = N, byrow = FALSE)
  dm3Y_ <- matrix(dm3Y, nrow = 8, ncol = N, byrow = FALSE)
  dm4X_ <- matrix(dm4X, nrow = 8, ncol = N, byrow = FALSE)
  dm4Y_ <- matrix(dm4Y, nrow = 8, ncol = N, byrow = FALSE)

  dsX_ <- (m2X_^(3/2) * dm3X_ - 1.5 * m3X_ * m2X_^(1/2) * dm2X_)/m2X_^3
  dsY_ <- (m2Y_^(3/2) * dm3Y_ - 1.5 * m3Y_ * m2Y_^(1/2) * dm2Y_)/m2Y_^3
  dkX_ <- (m2X_^2 * dm4X_ - 2 * m4X_ * m2X_ * dm2X_)/m2X_^4
  dkY_ <- (m2Y_^2 * dm4Y_ - 2 * m4Y_ * m2Y_ * dm2Y_)/m2Y_^4

  skewX_ <- m3X_/m2X_^(3/2)
  skewY_ <- m3Y_/m2Y_^(3/2)
  kurtX_ <- (m4X_/m2X_^2) - 3
  kurtY_ <- (m4Y_/m2Y_^2) - 3

  za <- qnorm(1 - level)

  # constants for speedup
  cst1X_ <- 1/(2 * sqrt(m2X_))
  cst1Y_ <- 1/(2 * sqrt(m2Y_))
  cst2X_ <- sqrt(m2X_)
  cst2Y_ <- sqrt(m2Y_)

  dmVaRX_ <- -dm1X_ - za * cst1X_ * dm2X_
  dmVaRX_ <- dmVaRX_ - (1/6) * (za^2 - 1) * (cst1X_ * skewX_ * dm2X_ +
                                               cst2X_ * dsX_)
  dmVaRX_ <- dmVaRX_ + (1/36) * (2 * za^3 - 5 * za) * (cst1X_ * skewX_^2 *
                                                         dm2X_ + 2 * cst2X_ * skewX_ * dsX_)
  dmVaRX_ <- dmVaRX_ - (1/24) * (za^3 - 3 * za) * (cst1X_ * kurtX_ *
                                                     dm2X_ + cst2X_ * dkX_)

  # dmVaRY
  dmVaRY_ <- -dm1Y_ - za * cst1Y_ * dm2Y_
  dmVaRY_ <- dmVaRY_ - (1/6) * (za^2 - 1) * (cst1Y_ * skewY_ * dm2Y_ +
                                               cst2Y_ * dsY_)
  dmVaRY_ <- dmVaRY_ + (1/36) * (2 * za^3 - 5 * za) * (cst1Y_ * skewY_^2 *
                                                         dm2Y_ + 2 * cst2Y_ * skewY_ * dsY_)
  dmVaRY_ <- dmVaRY_ - (1/24) * (za^3 - 3 * za) * (cst1Y_ * kurtY_ *
                                                     dm2Y_ + cst2Y_ * dkY_)

  # mVaR
  mVaRX_ <- -m1X_ + sqrt(m2X_) * (-za - (1/6) * (za^2 - 1) * skewX_ -
                                    (1/24) * (za^3 - 3 * za) * kurtX_ + (1/36) * (2 * za^3 - 5 * za) *
                                    skewX_^2)
  mVaRY_ <- -m1Y_ + sqrt(m2Y_) * (-za - (1/6) * (za^2 - 1) * skewY_ -
                                    (1/24) * (za^3 - 3 * za) * kurtY_ + (1/36) * (2 * za^3 - 5 * za) *
                                    skewY_^2)

  # gradient
  if (ttype == 1) {
    # test based on quotient
    tmp1 <- (mVaRX_ * dm1X_ - m1X_ * dmVaRX_)/mVaRX_^2
    tmp2 <- (mVaRY_ * dm1Y_ - m1Y_ * dmVaRY_)/mVaRY_^2
  } else {
    # test based on product
    tmp1 <- (mVaRY_ * dm1X_ + m1X_ * dmVaRY_)
    tmp2 <- (mVaRX_ * dm1Y_ + m1Y_ * dmVaRX_)
  }

  # =======
  gradient <- array(NA, c(8, 1, N))
  gradient[1:8, 1, ] <- tmp1 - tmp2
  V.hat <- array(NA, c(T, 8, N))
  V.hat[, 1, ] <- sweep(x = X, MARGIN = 2, STATS = m1X, FUN = "-")
  V.hat[, 5, ] <- sweep(x = Y, MARGIN = 2, STATS = m1Y, FUN = "-")
  V.hat[, 2, ] <- sweep(x = X^2, MARGIN = 2, STATS = g2X, FUN = "-")
  V.hat[, 6, ] <- sweep(x = Y^2, MARGIN = 2, STATS = g2Y, FUN = "-")
  V.hat[, 3, ] <- sweep(x = X^3, MARGIN = 2, STATS = g3X, FUN = "-")
  V.hat[, 7, ] <- sweep(x = Y^3, MARGIN = 2, STATS = g3Y, FUN = "-")
  V.hat[, 4, ] <- sweep(x = X^4, MARGIN = 2, STATS = g4X, FUN = "-")
  V.hat[, 8, ] <- sweep(x = Y^4, MARGIN = 2, STATS = g4Y, FUN = "-")
  Psi.hat <- array(apply(X = V.hat, MARGIN = 3, FUN = compute.Psi.hat,
                         b = b), c(8, 8, N))
  se <- vector("double", N)
  for (i in 1:N) {
    se[i] <- sqrt(crossprod(gradient[, , i], Psi.hat[, , i] %*% gradient[,
                                                                         , i])/T)
  }
  return(se)
}
se.msharpe.bootstrap <- compiler::cmpfun(.se.msharpe.bootstrap)

#' @name hfdata
#' @title Hedge fund data
#' @description The matrix \code{hfdata} contains randomized and modified monthly returns of hedge funds.
#' @source HFR database
#' @docType data
#' @keywords datasets
#' @usage data(hfdata)
#' @format A matrix of size 60 x 100, containing 60 monthly net returns for 100 hedge funds. Returns have been modified.
NULL

#' @name PeerPerformance
#' @aliases PeerPerformance-package
#' @title PeerPerformance: Luck-corrected peer performance analysis in R
#' @description \code{PeerPerformance} is an \R package for the peer-performance evaluation of financial investments with
#' luck-correction, useful in the financial industry. In particular, it implements the peer performance ratios of Ardia and Boudt
#' (2018) which measure the percentage of peers a focal (hedge) fund outperforms and underperforms, after
#' correction for luck. It is useful for fund or portfolio managers to benchmark their investments or screen a universe of new funds.
#' In addition, the package implements the testing framework for the Sharpe and modified Sharpe ratios, described in
#' Ledoit and Wolf (2008) and Ardia and Boudt (2015). See also Ardia et al. (2022,2023) for applications in sustainable finance.
#' @section Functions:
#' \itemize{
#' \item Alpha: \code{\link{alphaTesting}} and \code{\link{alphaScreening}};
#' \item Sharpe ratio: \code{\link{sharpe}}, \code{\link{sharpeTesting}} and \code{\link{sharpeScreening}};
#' \item Modified Sharpe ratio: \code{\link{msharpe}}, \code{\link{msharpeTesting}} and \code{\link{msharpeScreening}};
#' \item Screening: \code{\link{alphaScreening}}, \code{\link{sharpeScreening}}, and \code{\link{msharpeScreening}}.
#' }
#' @section Update:
#' The latest version of the package is available at \url{https://github.com/ArdiaD/PeerPerformance}
#' @author David Ardia and Kris Boudt.
#' @note By using \code{PeerPerformance} you agree to the following rules: (1) You must cite Ardia and Boudt (2018) in
#' working papers and published papers that use \code{PeerPerformance} (use \code{citation("PeerPerformance")}), (2) you
#' must place the URL \url{https://CRAN.R-project.org/package=PeerPerformance} in a footnote to help
#' others find \code{PeerPerformance}, and (3) you assume all risk for the use of \code{PeerPerformance}.
#' @note Full description of the methodologies implemented in the various functions is available
#' in Ledoit and Wolf (2008) and Ardia and Boudt (2015, 2018).
#' @references
#' Ardia, D., Boudt, K. (2015).
#' Testing equality of modified Sharpe ratios.
#' \emph{Finance Research Letters} \bold{13}, 97--104.
#'
#' Ardia, D., Boudt, K. (2018).
#' The peer performance ratios of hedge Funds.
#' \emph{Journal of Banking and Finance} \bold{87}, 351--368.
#'
#' Ardia, D., Bluteau, K., Tran, D. (2022).
#' How easy is it for investment managers to deploy their talent in green and brown stocks?
#' \emph{Finance Research Letters} \bold{48}, 102992.
#'
#' Ardia, D., Bluteau, K., Lortie-Cloutier, G., Tran, D. (2023).
#' Factor exposure heterogeneity in green and brown stocks.
#' \emph{Finance Research Letters} \bold{55}, Part A, 103900.
#'
#' Barras, L., Scaillet, O., Wermers, R. (2010).
#' False discoveries in mutual fund performance: Measuring luck in estimated alphas.
#' \emph{Journal of Finance} \bold{65}(1), 179--216.
#'
#' Favre, L., Galeano, J.A. (2002).
#' Mean-modified Value-at-Risk Optimization with Hedge Funds.
#' \emph{Journal of Alternative Investments} \bold{5}(2), 21--25.
#'
#' Gregoriou, G. N., Gueyie, J.-P. (2003).
#' Risk-adjusted performance of funds of hedge funds using a modified Sharpe ratio.
#' \emph{Journal of Wealth Management} \bold{6}(3), 77--83.
#'
#' Ledoit, O., Wolf, M. (2008).
#' Robust performance hypothesis testing with the Sharpe ratio.
#' \emph{Journal of Empirical Finance} \bold{15}(5), 850--859.
#'
#' Sharpe, W.F. (1994).
#' The Sharpe ratio.
#' \emph{Journal of Portfolio Management} \bold{21}(1), 49--58.
#'
#' Storey, J. (2002).
#' A direct approach to false discovery rates.
#' \emph{Journal of the Royal Statistical Society B} \bold{64}(3), 479--498.
#' @import lmtest
#' @import sandwich
#' @importFrom compiler cmpfun
NULL

## Targeted peer-performance screening for a selected subset of funds.
## This is a thin convenience wrapper over the cross-group screening
## (`*Screening(X[, funds], Y = X)`): each focal fund is screened against the
## whole universe `X` (its own column is excluded automatically). Originally
## contributed by Murilo Andre Peres Pereira.

.targetPeerPerformance <- function(X, funds,
                                   method = c("alpha", "sharpe", "msharpe"),
                                   factors = NULL, level = 0.9, na.neg = TRUE,
                                   control = list()) {
  method <- match.arg(method)
  X <- as.matrix(X)
  N <- ncol(X)
  if (is.null(colnames(X))) {
    colnames(X) <- paste0("Fund ", seq_len(N))
  }
  fund_names <- colnames(X)

  # resolve 'funds' to column indices (preserving the user-supplied order)
  idx <- if (is.numeric(funds)) {
    if (anyNA(funds) || any(!is.finite(funds)) || any(funds != as.integer(funds))) {
      stop("numeric 'funds' must be whole, finite, positive indices")
    }
    as.integer(funds)
  } else {
    m <- match(funds, fund_names)
    if (anyNA(m)) stop("unknown fund(s): ", paste(funds[is.na(m)], collapse = ", "))
    m
  }
  if (any(idx < 1L | idx > N)) stop("some 'funds' indices are out of range")
  if (anyDuplicated(idx)) stop("duplicate entries in 'funds' are not allowed")
  if (method != "alpha" && !is.null(factors)) {
    stop("'factors' is only used when method = 'alpha'")
  }

  Xf <- X[, idx, drop = FALSE]

  # each focal fund vs the whole universe (self column excluded by the
  # cross-group machinery)
  out <- switch(method,
                alpha   = alphaScreening(Xf, factors = factors, control = control, Y = X),
                sharpe  = sharpeScreening(Xf, control = control, Y = X),
                msharpe = msharpeScreening(Xf, level = level, na.neg = na.neg,
                                           control = control, Y = X))

  # label the focal funds (rows / vectors) and the peers (columns)
  fn <- fund_names[idx]
  dname <- switch(method, alpha = "dalpha", sharpe = "dsharpe", msharpe = "dmsharpe")
  for (el in c("n", "npeer", "lambda", "pizero", "pipos", "pineg", method)) {
    v <- out[[el]]
    if (is.null(v)) next
    if (is.matrix(v)) {
      # screen_beta: coefficient-by-focal-fund matrices
      if (ncol(v) == length(fn)) colnames(out[[el]]) <- fn
    } else if (length(v) == length(fn)) {
      names(out[[el]]) <- fn
    }
  }
  for (el in c("pval", "tstat", dname)) {
    if (is.matrix(out[[el]])) {
      rownames(out[[el]]) <- fn
      colnames(out[[el]]) <- fund_names
    }
  }
  out
}

#' @name targetPeerPerformance
#' @title Targeted peer-performance screening for selected funds
#' @description Performs peer-performance screening for a user-selected subset of
#' funds (the \dQuote{focal} funds) against the whole universe, returning
#' detailed outputs for those focal funds only. It is a convenience wrapper that
#' calls the corresponding screening function with the cross-group argument
#' \code{Y} set to the full matrix \code{X}; equivalently,
#' \code{targetPeerPerformance(X, funds, "alpha")} is
#' \code{alphaScreening(X[, funds], Y = X)}.
#' @param X Matrix \eqn{(T \times N)}{(TxN)} of \eqn{T} returns for \eqn{N}
#' funds. \code{NA} values are allowed.
#' @param funds Integer indices or character column names identifying the focal
#' funds. The output preserves the order provided in \code{funds}.
#' @param method Screening method: \code{"alpha"} (default), \code{"sharpe"}, or
#' \code{"msharpe"}.
#' @param factors Optional matrix of factor returns (only when
#' \code{method = "alpha"}).
#' @param level Modified Value-at-Risk level (only when \code{method = "msharpe"}).
#' @param na.neg Logical; return \code{NA} for a negative modified
#' Value-at-Risk (only when \code{method = "msharpe"}).
#' @param control Control parameters (see \code{\link{alphaScreening}}).
#' @return A \code{SCREENING} object (see \code{\link{alphaScreening}}) whose
#' fund-level components (\code{pizero}, \code{pipos}, \code{pineg}, the
#' performance measure, \code{npeer}, \ldots) cover the focal funds only, and
#' whose \code{pval}/\code{tstat}/difference matrices have one row per focal fund
#' and one column per universe fund. Works with the \code{print}, \code{plot},
#' \code{summary}, and \code{as.data.frame} methods.
#' @author Murilo Andre Peres Pereira, David Ardia and Kris Boudt.
#' @seealso \code{\link{alphaScreening}}, \code{\link{sharpeScreening}},
#' \code{\link{msharpeScreening}}.
#' @keywords htest
#' @examples
#' data("hfdata")
#' rets <- hfdata[, 1:10]
#'
#' ## focal Sharpe screening (funds by index) against the whole universe
#' out <- targetPeerPerformance(rets, funds = c(2, 5, 7), method = "sharpe",
#'                              control = list(nCore = 1))
#' out$sharpe
#' @export
#' @importFrom compiler cmpfun
targetPeerPerformance <- compiler::cmpfun(.targetPeerPerformance)

## Set of R functions for the optimal block length computation for the
## Sharpe ratio

# #' @name .sharpeBlockSize
# #' @title See sharpeBlockSize
# #' @importFrom stats rgeom
# #' @import compiler
.sharpeBlockSize <- function(x, y, control = list(), b.vec = c(1, 3, 6, 10), 
                             alpha = 0.05, M = 199, K = 500, b.av = 5, 
                             T.start = 50) {
  
  sb.sequence <- function(T, b.av, length = T) {
    
    index.sequence <- c(1:T, 1:T)
    sequence <- rep.int(0, length + T)
    current <- 0
    while (current < length) {
      start <- sample(1:T, 1)
      b <- stats::rgeom(1, 1/b.av) + 1
      sequence[(current + 1):(current + b)] <- index.sequence[start:(start + 
                                                                       b - 1)]
      current <- current + b
    }
    out <- sequence[1:length]
    return(out)
    
  }
  
  x <- as.matrix(x)
  y <- as.matrix(y)
  rets <- cbind(x, y)
  
  # process control parameters
  ctr <- processControl(control)
  
  b.len <- length(b.vec)
  emp.reject.probs <- rep.int(0, b.len)
  d <- sharpe.ratio.diff(x, y, ctr$ttype)
  T <- length(x)
  Var.data <- matrix(data = 0, nrow = T.start + T, ncol = 2)
  Var.data[1, ] <- rets[1, ]
  fit1 <- stats::lm(x[2:T] ~ x[1:(T - 1)] + y[1:(T - 1)])
  fit2 <- stats::lm(y[2:T] ~ x[1:(T - 1)] + y[1:(T - 1)])
  coef1 <- as.numeric(fit1$coef)
  coef2 <- as.numeric(fit2$coef)
  resid.mat <- cbind(as.numeric(fit1$resid), as.numeric(fit2$resid))
  for (k in 1:K) {
    ids <- sb.sequence(T - 1, b.av, T.start + T - 1)
    resid.mat.star <- rbind(c(0, 0), resid.mat[ids, ])
    for (t in 2:(T.start + T)) {
      Var.data[t, 1] <- coef1[1] + coef1[2] * Var.data[t - 1, 1] + 
        coef1[3] * Var.data[t - 1, 2] + resid.mat.star[t, 1]
      Var.data[t, 2] <- coef2[1] + coef2[2] * Var.data[t - 1, 1] +
        coef2[3] * Var.data[t - 1, 2] + resid.mat.star[t, 2]
    }
    Var.data.trunc <- Var.data[(T.start + 1):(T.start + T), ]
    for (j in 1:b.len) {
      bsids <- bootIndices(T, M, b.vec[j])
      tmp <- sharpeTestBootstrap(Var.data.trunc, bsids, b.vec[j], 
                                 ctr$ttype, pBoot = 1, d)
      if (tmp$pval <= alpha) {
        emp.reject.probs[j] <- emp.reject.probs[j] + 1
      }
    }
  }
  emp.reject.probs <- emp.reject.probs/K
  b.order <- order(abs(emp.reject.probs - alpha))
  b.opt <- b.vec[b.order[1]]
  return(b.opt)
}
sharpeBlockSize <- compiler::cmpfun(.sharpeBlockSize)

context("Screening")

f.test.pval = function(pval) {
  out = pval >= 0 && pval <= 1
  return(out)
}

data('hfdata')
N = 10
rets = hfdata[,1:10]

test_that("Alpha", {
  tmp = alphaScreening(rets, control = list(nCore = 1))
  expect_equal(length(tmp$n), N)
  expect_equal(length(tmp$alpha), N)
  expect_equal(length(tmp$npeer), N)
  expect_equal(length(tmp$lambda), N)
  expect_equal(dim(tmp$dalpha), c(N,N))
  expect_equal(dim(tmp$tstat), c(N,N))
  expect_equal(dim(tmp$pval), c(N,N))

  tmp = alphaScreening(rets, control = list(nCore = 1, hac = TRUE))
  expect_equal(length(tmp$n), N)
  expect_equal(length(tmp$alpha), N)
  expect_equal(length(tmp$npeer), N)
  expect_equal(length(tmp$lambda), N)
  expect_equal(dim(tmp$dalpha), c(N,N))
  expect_equal(dim(tmp$tstat), c(N,N))
  expect_equal(dim(tmp$pval), c(N,N))
})

test_that("Sharpe", {
  tmp = sharpeScreening(rets, control = list(nCore = 1))
  expect_equal(length(tmp$n), N)
  expect_equal(length(tmp$sharpe), N)
  expect_equal(length(tmp$npeer), N)
  expect_equal(length(tmp$lambda), N)
  expect_equal(dim(tmp$dsharpe), c(N,N))
  expect_equal(dim(tmp$tstat), c(N,N))
  expect_equal(dim(tmp$pval), c(N,N))

  tmp = sharpeScreening(rets, control = list(nCore = 1, hac = TRUE))
  expect_equal(length(tmp$n), N)
  expect_equal(length(tmp$sharpe), N)
  expect_equal(length(tmp$npeer), N)
  expect_equal(length(tmp$lambda), N)
  expect_equal(dim(tmp$dsharpe), c(N,N))
  expect_equal(dim(tmp$tstat), c(N,N))
  expect_equal(dim(tmp$pval), c(N,N))
})

test_that("Modified Sharpe", {
  tmp = msharpeScreening(rets, level = 0.9, control = list(nCore = 1))
  expect_equal(length(tmp$n), N)
  expect_equal(length(tmp$msharpe), N)
  expect_equal(length(tmp$npeer), N)
  expect_equal(length(tmp$lambda), N)
  expect_equal(dim(tmp$dmsharpe), c(N,N))
  expect_equal(dim(tmp$tstat), c(N,N))
  expect_equal(dim(tmp$pval), c(N,N))

  tmp = msharpeScreening(rets, level = 0.95, control = list(nCore = 1, hac = TRUE))
  expect_equal(length(tmp$n), N)
  expect_equal(length(tmp$msharpe), N)
  expect_equal(length(tmp$npeer), N)
  expect_equal(length(tmp$lambda), N)
  expect_equal(dim(tmp$dmsharpe), c(N,N))
  expect_equal(dim(tmp$tstat), c(N,N))
  expect_equal(dim(tmp$pval), c(N,N))
})

---
title: "Luck-Corrected Peer Performance Analysis with PeerPerformance"
author: "David Ardia, Murilo Andre Peres Pereira, Benjamin Seguin"
output: rmarkdown::html_vignette
bibliography: null
vignette: >
  %\VignetteIndexEntry{Luck-Corrected Peer Performance Analysis with PeerPerformance}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", fig.width = 7,
                      fig.height = 3.4)
```

## Overview

`PeerPerformance` evaluates the performance of investment funds *relative to
their peers*, with a correction for luck that is robust to false discoveries.
For each fund $i$ it estimates three **peer performance ratios** that sum to
one:

- $\hat\pi^+_i$ -- the proportion of peers fund $i$ significantly **outperforms**,
- $\hat\pi^0_i$ -- the proportion of peers with **equal** performance,
- $\hat\pi^-_i$ -- the proportion of peers that **outperform** fund $i$.

The estimators implement the methodology of Ardia and Boudt (2018,
*Journal of Banking & Finance*); the modified Sharpe ratio test follows
Ardia and Boudt (2015, *Finance Research Letters*). A naive percentile rank
ignores that many funds may be statistically indistinguishable; the
false-discovery-rate (FDR) correction of Storey (2002) corrects for this.

```{r}
library(PeerPerformance)
data("hfdata")          # 60 monthly returns for 100 anonymized hedge funds
dim(hfdata)
```

## Alpha peer performance screening

`alphaScreening()` runs all pairwise tests and returns the three ratios for
each fund. With `factors = NULL` it compares raw returns; supply a `factors`
matrix to compare risk-adjusted alphas.

```{r}
set.seed(1234)
rets <- hfdata[, 1:15]
sc <- alphaScreening(rets, control = list(nCore = 1))
round(rbind(pipos = sc$pipos, pizero = sc$pizero, pineg = sc$pineg), 3)[, 1:6]
```

The three ratios sum to one for every fund. A `summary()` method ranks the
funds and reports the win/loss counts:

```{r}
summary(sc, top = 3)
```

The ratios are point estimates; `confint()` attaches confidence intervals by a
nonparametric peer (pairwise) bootstrap that resamples each fund's peers:

```{r}
set.seed(1234)
round(confint(sc, parm = "pipos")[1:6, ], 3)
```

The `plot()` method reproduces the Ardia and Boudt (2018) screening plot
(funds sorted by performance; stacked out-/equal-/under-performance bars with
the naive percentile-rank diagonal):

```{r screenfig}
plot(alphaScreening(hfdata[, 1:30], control = list(nCore = 1)))
```

## Sharpe and modified Sharpe ratios

```{r}
round(sharpe(rets[, 1:6]), 3)
round(msharpe(rets[, 1:6], level = 0.95), 3)
```

Equality of two funds' (modified) Sharpe ratios is tested with
`sharpeTesting()` / `msharpeTesting()`. The asymptotic test is the default; a
studentized bootstrap (recommended for short or autocorrelated series) is
requested with `control = list(type = 2)`.

```{r}
res <- msharpeTesting(hfdata[, 1], hfdata[, 2], level = 0.95)
c(dmsharpe = res$dmsharpe, tstat = res$tstat, pval = res$pval)
```

`sharpeScreening()` and `msharpeScreening()` build the peer performance ratios
from Sharpe (resp. modified Sharpe) comparisons instead of alphas.

## Comparing a fund (or group) against a separate peer group

The `Y` argument (or the convenience wrapper `targetPeerPerformance()`) screens
a chosen fund, or subset of funds, against a separate universe.

```{r}
focal <- hfdata[, 1]
peers <- hfdata[, 11:40]
res_xy <- alphaScreening(focal, Y = peers, control = list(nCore = 1))
round(c(pizero = res_xy$pizero, pipos = res_xy$pipos, pineg = res_xy$pineg), 3)
```

## Control parameters

All screening/testing functions share a `control` list. Common fields include
`nCore` (parallel cores), `lambda` (the $\lambda$ threshold for $\hat\pi^0$,
`NULL` = data-driven), `gammaPos`/`gammaNeg` (the one-sided thresholds, default
`0.4`/`0.6`), `hac` (HAC standard errors), and for the Sharpe/mSR routines
`type` (asymptotic/bootstrap), `nBoot`, and `bBoot`. See `?alphaScreening`.

Note that with the data-driven `lambda` (the default) the bias correction of
$\hat\pi^0$ dominates the run time on large universes. Setting
`control = list(fastAdjust = TRUE)` inverts it in vectorised form and is
several times faster, at the price of differing from the published code path
by a few $10^{-5}$; and setting a fixed `lambda` skips the selection entirely.

## References

- Ardia, D., Boudt, K. (2018). The peer performance ratios of hedge funds.
  *Journal of Banking & Finance* **87**, 351--368.
- Ardia, D., Boudt, K. (2015). Testing equality of modified Sharpe ratios.
  *Finance Research Letters* **13**, 97--104.
- Storey, J. (2002). A direct approach to false discovery rates.
  *Journal of the Royal Statistical Society B* **64**(3), 479--498.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/targetPeerPerformance.R
\name{targetPeerPerformance}
\alias{targetPeerPerformance}
\title{Targeted peer-performance screening for selected funds}
\usage{
targetPeerPerformance(
  X,
  funds,
  method = c("alpha", "sharpe", "msharpe"),
  factors = NULL,
  level = 0.9,
  na.neg = TRUE,
  control = list()
)
}
\arguments{
\item{X}{Matrix \eqn{(T \times N)}{(TxN)} of \eqn{T} returns for \eqn{N}
funds. \code{NA} values are allowed.}

\item{funds}{Integer indices or character column names identifying the focal
funds. The output preserves the order provided in \code{funds}.}

\item{method}{Screening method: \code{"alpha"} (default), \code{"sharpe"}, or
\code{"msharpe"}.}

\item{factors}{Optional matrix of factor returns (only when
\code{method = "alpha"}).}

\item{level}{Modified Value-at-Risk level (only when \code{method = "msharpe"}).}

\item{na.neg}{Logical; return \code{NA} for a negative modified
Value-at-Risk (only when \code{method = "msharpe"}).}

\item{control}{Control parameters (see \code{\link{alphaScreening}}).}
}
\value{
A \code{SCREENING} object (see \code{\link{alphaScreening}}) whose
fund-level components (\code{pizero}, \code{pipos}, \code{pineg}, the
performance measure, \code{npeer}, \ldots) cover the focal funds only, and
whose \code{pval}/\code{tstat}/difference matrices have one row per focal fund
and one column per universe fund. Works with the \code{print}, \code{plot},
\code{summary}, and \code{as.data.frame} methods.
}
\description{
Performs peer-performance screening for a user-selected subset of
funds (the \dQuote{focal} funds) against the whole universe, returning
detailed outputs for those focal funds only. It is a convenience wrapper that
calls the corresponding screening function with the cross-group argument
\code{Y} set to the full matrix \code{X}; equivalently,
\code{targetPeerPerformance(X, funds, "alpha")} is
\code{alphaScreening(X[, funds], Y = X)}.
}
\examples{
data("hfdata")
rets <- hfdata[, 1:10]

## focal Sharpe screening (funds by index) against the whole universe
out <- targetPeerPerformance(rets, funds = c(2, 5, 7), method = "sharpe",
                             control = list(nCore = 1))
out$sharpe
}
\seealso{
\code{\link{alphaScreening}}, \code{\link{sharpeScreening}},
\code{\link{msharpeScreening}}.
}
\author{
Murilo Andre Peres Pereira, David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/msharpeTesting.R
\name{msharpeTesting}
\alias{msharpeTesting}
\title{Testing the difference of modified Sharpe ratios}
\usage{
msharpeTesting(x, y, level = 0.9, na.neg = TRUE, control = list())
}
\arguments{
\item{x}{Vector (of length \eqn{T}) of returns for the first fund. \code{NA}
values are allowed.}

\item{y}{Vector (of length \eqn{T}) of returns for the second fund. \code{NA}
values are allowed.}

\item{level}{Modified Value-at-Risk level. Default: \code{level = 0.90}.}

\item{na.neg}{A logical value indicating whether \code{NA} values should be
returned if a negative modified Value-at-Risk is obtained.  Default
\code{na.neg = TRUE}.}

\item{control}{Control parameters (see *Details*).}
}
\value{
A list with the following components:\cr

\code{n}: Number of non-\code{NA} concordant observations.\cr

\code{msharpe}: Vector (of length 2) of unconditional modified Sharpe
ratios.\cr

\code{dmsharpe}: Modified Sharpe ratios difference.\cr

\code{tstat}: t-stat of modified Sharpe ratios differences.\cr

\code{pval}: pvalues of test of modified Sharpe ratios differences.
}
\description{
Function which performs the testing of the difference of modified Sharpe
ratios.
}
\details{
The modified Sharpe ratio (Favre and Galeano 2002) is one industry
standard for measuring the absolute risk adjusted performance of hedge
funds. This function performs the testing of modified Sharpe ratio
difference for two funds using a similar approach than Ledoit and Wolf
(2002). See also Gregoriou and Gueyie (2003).

For the testing, only the intersection of non-\code{NA} observations for the
two funds are used.

The argument \code{control} is a list that can supply any of the following
components:
\itemize{
\item \code{'type'} Asymptotic approach (\code{type = 1}) or
studentized circular bootstrap approach (\code{type = 2}). Default:
\code{type = 1}.
\item \code{'ttype'} Test based on ratio (\code{type = 1})
or product (\code{type = 2}). Default: \code{type = 2}.
\item \code{'hac'} Heteroscedastic-autocorrelation consistent standard
errors. Default: \code{hac = FALSE}.
\item \code{'minObs'} Minimum number of concordant observations to compute the ratios. Default: \code{minObs =
10}.
\item \code{'nBoot'} Number of bootstrap replications for computing the
p-value. Default: \code{nBoot = 499}.
\item \code{'bBoot'} Block length in
the circular bootstrap. Default: \code{bBoot = 1}, i.e. iid bootstrap.
\code{bBoot = 0} uses optimal block-length.
\item \code{'pBoot'} Symmetric
p-value (\code{pBoot = 1}) or asymmetric p-value (\code{pBoot = 2}).
Default: \code{pBoot = 1}.
}
}
\note{
Further details on the methodology with an application to the hedge
fund industry is given in Ardia and Boudt (2018).

Some internal functions where adapted from Michael Wolf MATLAB code.
}
\examples{
## Load the data (randomized data of monthly hedge fund returns)
data("hfdata")
x = hfdata[,1]
y = hfdata[,2]

## Run modified Sharpe testing (asymptotic)
ctr = list(type = 1)
out = msharpeTesting(x, y, level = 0.95, control = ctr)
print(out)

## Run modified Sharpe testing (asymptotic hac)
ctr = list(type = 1, hac = TRUE)
out = msharpeTesting(x, y, level = 0.95, control = ctr)
print(out)

## Run modified Sharpe testing (iid bootstrap)
set.seed(1234)
ctr = list(type = 2, nBoot = 250)
out = msharpeTesting(x, y, level = 0.95, control = ctr)
print(out)

## Run modified Sharpe testing (circular bootstrap)
set.seed(1234)
ctr = list(type = 2, nBoot = 250, bBoot = 5)
out = msharpeTesting(x, y, level = 0.95, control = ctr)
print(out)
}
\references{
Ardia, D., Boudt, K. (2015).
Testing equality of modified Sharpe ratios.
\emph{Finance Research Letters} \bold{13}, pp.97--104.
\doi{10.1016/j.frl.2015.02.008}

Ardia, D., Boudt, K. (2018).
The peer performance ratios of hedge funds.
\emph{Journal of Banking and Finance} \bold{87}, pp.351--368.
\doi{10.1016/j.jbankfin.2017.10.014}

Barras, L., Scaillet, O., Wermers, R. (2010).
False discoveries in mutual fund performance: Measuring luck in estimated alphas.
\emph{Journal of Finance} \bold{65}(1), pp.179--216.

Favre, L., Galeano, J.A. (2002).
Mean-modified Value-at-Risk Optimization with Hedge Funds.
\emph{Journal of Alternative Investments} \bold{5}(2), pp.21--25.

Gregoriou, G. N., Gueyie, J.-P. (2003).
Risk-adjusted performance of funds of hedge funds using a modified Sharpe ratio.
\emph{Journal of Wealth Management} \bold{6}(3), pp.77--83.

Ledoit, O., Wolf, M. (2008).
Robust performance hypothesis testing with the Sharpe ratio.
\emph{Journal of Empirical Finance} \bold{15}(5), pp.850--859.

Storey, J. (2002).
A direct approach to false discovery rates.
\emph{Journal of the Royal Statistical Society B} \bold{64}(3), pp.479--498.
}
\seealso{
\code{\link{msharpe}}, \code{\link{msharpeScreening}} and
\code{\link{sharpeTesting}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{as.data.frame.SCREENING}
\alias{as.data.frame.SCREENING}
\title{Coerce a 'SCREENING' object to a data frame}
\usage{
\method{as.data.frame}{SCREENING}(x, row.names = NULL, optional = FALSE, ...)
}
\arguments{
\item{x}{A \code{SCREENING} object.}

\item{row.names, optional}{Currently ignored.}

\item{...}{Further arguments (ignored).}
}
\value{
A \code{data.frame} with columns \code{fund}, the performance measure
(\code{alpha}, \code{sharpe}, or \code{msharpe}), \code{pipos}, \code{pizero},
\code{pineg}, \code{npeer}, and \code{lambda}; a \code{coefficient} column is
added when \code{screen_beta = TRUE}.
}
\description{
Returns a tidy \code{data.frame} with one row per fund (or, when
the screening was run with \code{screen_beta = TRUE}, one row per
fund/coefficient), suitable for downstream filtering, ranking, and plotting.
}
\examples{
\donttest{
  data("hfdata")
  sc <- alphaScreening(hfdata[, 1:10], control = list(nCore = 1))
  head(as.data.frame(sc))
}
}
\seealso{
\code{\link{alphaScreening}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.summary.SCREENING}
\alias{print.summary.SCREENING}
\title{Print method for the 'summary.SCREENING' object}
\usage{
\method{print}{summary.SCREENING}(x, ...)
}
\arguments{
\item{x}{A \code{summary.SCREENING} object.}

\item{...}{Further arguments (ignored).}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print method for the object returned by
\code{\link{summary.SCREENING}}.
}
\author{
Murilo Andre Peres Pereira, David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PeerPerformance.R
\name{PeerPerformance}
\alias{PeerPerformance}
\alias{PeerPerformance-package}
\title{PeerPerformance: Luck-corrected peer performance analysis in R}
\description{
\code{PeerPerformance} is an \R package for the peer-performance evaluation of financial investments with
luck-correction, useful in the financial industry. In particular, it implements the peer performance ratios of Ardia and Boudt
(2018) which measure the percentage of peers a focal (hedge) fund outperforms and underperforms, after
correction for luck. It is useful for fund or portfolio managers to benchmark their investments or screen a universe of new funds.
In addition, the package implements the testing framework for the Sharpe and modified Sharpe ratios, described in
Ledoit and Wolf (2008) and Ardia and Boudt (2015). See also Ardia et al. (2022,2023) for applications in sustainable finance.
}
\note{
By using \code{PeerPerformance} you agree to the following rules: (1) You must cite Ardia and Boudt (2018) in
working papers and published papers that use \code{PeerPerformance} (use \code{citation("PeerPerformance")}), (2) you
must place the URL \url{https://CRAN.R-project.org/package=PeerPerformance} in a footnote to help
others find \code{PeerPerformance}, and (3) you assume all risk for the use of \code{PeerPerformance}.

Full description of the methodologies implemented in the various functions is available
in Ledoit and Wolf (2008) and Ardia and Boudt (2015, 2018).
}
\section{Functions}{

\itemize{
\item Alpha: \code{\link{alphaTesting}} and \code{\link{alphaScreening}};
\item Sharpe ratio: \code{\link{sharpe}}, \code{\link{sharpeTesting}} and \code{\link{sharpeScreening}};
\item Modified Sharpe ratio: \code{\link{msharpe}}, \code{\link{msharpeTesting}} and \code{\link{msharpeScreening}};
\item Screening: \code{\link{alphaScreening}}, \code{\link{sharpeScreening}}, and \code{\link{msharpeScreening}}.
}
}

\section{Update}{

The latest version of the package is available at \url{https://github.com/ArdiaD/PeerPerformance}
}

\references{
Ardia, D., Boudt, K. (2015).
Testing equality of modified Sharpe ratios.
\emph{Finance Research Letters} \bold{13}, 97--104.

Ardia, D., Boudt, K. (2018).
The peer performance ratios of hedge Funds.
\emph{Journal of Banking and Finance} \bold{87}, 351--368.

Ardia, D., Bluteau, K., Tran, D. (2022).
How easy is it for investment managers to deploy their talent in green and brown stocks?
\emph{Finance Research Letters} \bold{48}, 102992.

Ardia, D., Bluteau, K., Lortie-Cloutier, G., Tran, D. (2023).
Factor exposure heterogeneity in green and brown stocks.
\emph{Finance Research Letters} \bold{55}, Part A, 103900.

Barras, L., Scaillet, O., Wermers, R. (2010).
False discoveries in mutual fund performance: Measuring luck in estimated alphas.
\emph{Journal of Finance} \bold{65}(1), 179--216.

Favre, L., Galeano, J.A. (2002).
Mean-modified Value-at-Risk Optimization with Hedge Funds.
\emph{Journal of Alternative Investments} \bold{5}(2), 21--25.

Gregoriou, G. N., Gueyie, J.-P. (2003).
Risk-adjusted performance of funds of hedge funds using a modified Sharpe ratio.
\emph{Journal of Wealth Management} \bold{6}(3), 77--83.

Ledoit, O., Wolf, M. (2008).
Robust performance hypothesis testing with the Sharpe ratio.
\emph{Journal of Empirical Finance} \bold{15}(5), 850--859.

Sharpe, W.F. (1994).
The Sharpe ratio.
\emph{Journal of Portfolio Management} \bold{21}(1), 49--58.

Storey, J. (2002).
A direct approach to false discovery rates.
\emph{Journal of the Royal Statistical Society B} \bold{64}(3), 479--498.
}
\author{
David Ardia and Kris Boudt.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alphaScreening.R
\name{alphaScreening}
\alias{alphaScreening}
\title{Screening using the alpha outperformance ratio}
\usage{
alphaScreening(
  X,
  factors = NULL,
  control = list(),
  screen_beta = NULL,
  Y = NULL
)
}
\arguments{
\item{X}{Matrix \eqn{(T \times N)}{(TxN)} of \eqn{T} returns for the \eqn{N}
funds. \code{NA} values are allowed.}

\item{factors}{Matrix \eqn{(T \times K)}{(TxK)} of \eqn{T} returns for the
\eqn{K} factors. \code{NA} values are allowed.}

\item{control}{Control parameters (see *Details*).}

\item{screen_beta}{Boolean to screen all factors' coefficients (beta).
Default: \code{screen_beta = NULL}, in which case the value is taken from
\code{control$screen_beta} (itself defaulting to \code{FALSE}, i.e. only the
alpha is screened). When supplied directly, the argument takes precedence
over the control list. If \code{TRUE}, each element of the returned list will
have a new first dimension representing each coefficient (the first one being
alpha).}

\item{Y}{Optional matrix \eqn{(T \times M)}{(TxM)} of returns for a second
(peer) group of \eqn{M} funds. When supplied, the ratios are computed for
each fund in \code{X} \emph{against the funds in \code{Y}} (cross-group
screening) instead of against the other funds in \code{X}. A single focal
fund versus a peer group corresponds to \code{X} being a vector (or a
one-column matrix). Columns of \code{Y} identical to the focal fund (e.g.
when \code{X} is a subset of \code{Y}) are automatically excluded. Default:
\code{Y = NULL}, i.e. within-group screening.}
}
\value{
A list with the following components:\cr

\code{n}: Vector (of length \eqn{N}) of number of non-\code{NA}
observations.\cr

\code{npeer}: Vector (of length \eqn{N}) of number of available peers.\cr

\code{alpha}: Vector (of length \eqn{N}) of unconditional alpha.\cr

\code{dalpha}: Matrix (of size \eqn{N \times N}{NxN}) of alpha
differences.\cr

\code{tstat}: Matrix (of size \eqn{N \times N}{NxN}) of t-statistics.\cr

\code{pval}: Matrix (of size \eqn{N \times N}) of p-values of test for alpha
differences.\cr

\code{lambda}: Vector (of length \eqn{N}) of lambda values.\cr

\code{pizero}: Vector (of length \eqn{N}) of probability of equal
performance.\cr

\code{pipos}: Vector (of length \eqn{N}) of probability of outperformance
performance.\cr

\code{pineg}: Vector (of length \eqn{N}) of probability of underperformance
performance.
}
\description{
Function which performs the screening of a universe of returns, and
computes the alpha outperformance ratio.
}
\details{
The alpha measure (Treynor and Black 1973, Carhart 1997, Fung and Hsieh
2004) is one industry standard for measuring the absolute risk adjusted
performance of hedge funds. We propose to complement the alpha measure with
the fund's alpha outperformance ratio, defined as the percentage number of
funds that have a significantly lower alpha. In a pairwise testing
framework, a fund can have a significantly higher alpha because of luck. We
correct for this by applying the false discovery rate approach by Storey (2002).

The methodology proceeds as follows:
\itemize{
\item (1) compute all pairwise tests of alpha differences. This means that for a universe of
\eqn{N} funds, we perform \eqn{N(N-1)/2}{N*(N-1)/2} tests. The algorithm has
been parallelized and the computational burden can be split across several
cores. The number of cores can be defined in \code{control}, see below.
\item (2) for each fund, the false discovery rate approach by Storey (2002)
is used to determine the proportions of over, equal, and underperforming
funds, in terms of alpha, in the database.}
The argument \code{control} is a list that can supply any of the following
components:
\itemize{
\item \code{'hac'} Heteroscedastic-autocorrelation consistent
standard errors. Default: \code{hac = FALSE}.
\item \code{'minObs'} Minimum number of concordant observations to compute the ratios. Default:
\code{minObs = 10}.
\item \code{'minObsPi'} Minimum number of observations
for computing the p-values). Default: \code{minObsPi = 1}.
\item \code{'nCore'} Number of cores used to perform the screening. Default:
\code{nCore = 1}.
\item \code{'lambda'} Threshold value to compute pi0.
Default: \code{lambda = NULL}, i.e. data driven choice.
\item \code{'gammaPos'} One-sided quantile level (of the standard Normal
distribution) used as the critical value for counting outperformed peers:
a peer counts as outperformed when the pairwise t-statistic exceeds
\code{qnorm(gammaPos)} (a \emph{negative} threshold for
\code{gammaPos < 0.5}; e.g., \code{qnorm(0.4)} is about \code{-0.25}), and
the expected fraction \code{1 - gammaPos} of false positives among the
equal-performing peers is then subtracted. Smaller values count more peers
before the correction. Default: \code{gammaPos = 0.4} (the value
recommended in Ardia and Boudt, 2018).
\item \code{'gammaNeg'} Mirror image of \code{gammaPos} for the peers that
outperform the focal fund: the count uses \code{tstat <= qnorm(gammaNeg)}
and subtracts the expected fraction \code{gammaNeg} of false positives.
Default: \code{gammaNeg = 0.6}.
\item \code{'fastAdjust'} Use a fast vectorised inversion in the
truncated-normal bias correction of \eqn{\pi^0}{pi0} instead of one
\code{uniroot} call per value. This is the dominant cost when \code{lambda}
is data driven and gives a large speed-up on big universes. The bisection
locates the root to about 1e-12; since \code{uniroot} stops at its own
tolerance (about 1.2e-4), the two paths typically differ by a few 1e-5, the
fast path being the more accurate. Default: \code{fastAdjust = FALSE},
i.e. the original code path, kept as default so that published results
reproduce exactly.
\item \code{'screen_beta'} Screen the factor exposures (betas) in addition to
the alpha; see the \code{screen_beta} argument. Default:
\code{screen_beta = FALSE}.
}
}
\note{
Further details on the methodology with an application to the hedge
fund industry is given in Ardia and Boudt (2018).

Application of the false discovery rate approach applied to the mutual fund
industry has been presented in Barras, Scaillet and Wermers (2010).

HAC standard errors are available via \code{control = list(hac = TRUE)}
(computed with \pkg{sandwich}/\pkg{lmtest}). The studentized circular block
bootstrap of Ledoit and Wolf (2008) applies to the Sharpe-ratio routines
(\code{\link{sharpeScreening}}, \code{\link{msharpeScreening}}) and is not
used by \code{alphaScreening}.
}
\examples{
## Load the data (randomized data of monthly hedge fund returns)
data("hfdata")
rets = hfdata[,1:4]

## Run alpha screening
ctr = list(nCore = 1)
alphaScreening(rets, control = ctr)

## Run alpha screening with HAC standard deviation
ctr = list(nCore = 1, hac = TRUE)
alphaScreening(rets, control = ctr)

## Cross-group screening: a single focal fund against a peer group
alphaScreening(hfdata[, 1], Y = hfdata[, 11:20], control = list(nCore = 1))

## Cross-group screening: peer group X against peer group Y
alphaScreening(hfdata[, 1:5], Y = hfdata[, 11:20], control = list(nCore = 1))
}
\references{
Ardia, D., Boudt, K. (2015).
Testing equality of modified Sharpe ratios.
\emph{Finance Research Letters} \bold{13}, pp.97--104.
\doi{10.1016/j.frl.2015.02.008}

Ardia, D., Boudt, K. (2018).
The peer performance ratios of hedge funds.
\emph{Journal of Banking and Finance} \bold{87}, pp.351--368.
\doi{10.1016/j.jbankfin.2017.10.014}

Barras, L., Scaillet, O., Wermers, R. (2010).
False discoveries in mutual fund performance: Measuring luck in estimated alphas.
\emph{Journal of Finance} \bold{65}(1), pp.179--216.

Carhart, M. (1997).
On persistence in mutual fund performance.
\emph{Journal of Finance} \bold{52}(1), pp.57--82.

Fama, E., French, K. (2010).
Luck versus skill in the cross-section of mutual fund returns.
\emph{Journal of Finance} \bold{65}(5), pp.1915--1947.

Fung, W., Hsieh, D. (2004).
Hedge fund benchmarks: A risk based approach.
\emph{Financial Analysts Journal} \bold{60}(5), pp.65--80.

Storey, J. (2002).
A direct approach to false discovery rates.
\emph{Journal of the Royal Statistical Society B} \bold{64}(3), pp.479--498.

Treynor, J. L., Black, F. (1973).
How to use security analysis to improve portfolio selection.
\emph{Journal of Business} \bold{46}(1), pp.66--86.
}
\seealso{
\code{\link{sharpeScreening}} and \code{\link{msharpeScreening}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rollScreening.R
\name{plot.rollScreening}
\alias{plot.rollScreening}
\title{Plot a rolling peer performance screening}
\usage{
\method{plot}{rollScreening}(x, what = NULL, legend = TRUE, ...)
}
\arguments{
\item{x}{A \code{rollScreening} object.}

\item{what}{Either \code{"heterogeneity"} (\eqn{1-\bar\pi^0}) or
\code{"ratios"} (the averaged \eqn{\bar\pi^+,\bar\pi^0,\bar\pi^-}). Default:
\code{"heterogeneity"} for a beta screening, \code{"ratios"} otherwise.}

\item{legend}{A logical value indicating whether to draw a legend. Default:
\code{TRUE}.}

\item{...}{Further graphical arguments.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Time-series plot of the rolling peer performance ratios produced
by \code{\link{rollScreening}}.
}
\examples{
\donttest{
  data("hfdata")
  roll <- rollScreening(hfdata[, 1:15], width = 36, by = 12,
                        control = list(nCore = 1))
  plot(roll)
}
}
\seealso{
\code{\link{rollScreening}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{msharpe}
\alias{msharpe}
\title{Compute modified Sharpe ratio}
\usage{
msharpe(X, level = 0.9, na.rm = TRUE, na.neg = TRUE)
}
\arguments{
\item{X}{Vector (of length \eqn{T}) or matrix (of size \eqn{T \times
N}{TxN}) of returns.  \code{NA} values are allowed.}

\item{level}{Modified Value-at-Risk level. Default: \code{level = 0.90}.}

\item{na.rm}{A logical value indicating whether \code{NA} values should be
stripped before the computation. Default \code{na.rm = TRUE}.}

\item{na.neg}{A logical value indicating whether \code{NA} values should be
returned if a negative modified Value-at-Risk is obtained.  Default
\code{na.neg = TRUE}.}
}
\value{
Scalar or a vector (of size \eqn{N}) with the modified Sharpe
ratios.
}
\description{
Function which computes the modified Sharpe ratio
}
\details{
The modified Sharpe ratio (Favre and Galeano 2002) is one industry
standard for measuring the absolute risk adjusted performance of hedge
funds.
}
\examples{
## Load the data (randomized data of monthly hedge fund returns)
data('hfdata')

out = msharpe(hfdata)
print(out)

out = msharpe(hfdata, na.rm = FALSE)
print(out)
}
\references{
Ardia, D., Boudt, K. (2015).
Testing equality of modified Sharpe ratios.
\emph{Finance Research Letters} \bold{13}, 97--104.

Ardia, D., Boudt, K. (2018).
The peer performance ratios of hedge Funds.
\emph{Journal of Banking and Finance} \bold{87}, 351--368.

Favre, L., Galeano, J.A. (2002).
Mean-modified Value-at-Risk Optimization with Hedge Funds.
\emph{Journal of Alternative Investments} \bold{5}(2), 21--25.

Gregoriou, G. N., Gueyie, J.-P. (2003).
Risk-adjusted performance of funds of hedge funds using a modified Sharpe ratio.
\emph{Journal of Wealth Management} \bold{6}(3), 77--83.
}
\seealso{
\code{\link{msharpeTesting}}, \code{\link{msharpeScreening}} and
\code{\link{sharpe}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hfdata.R
\docType{data}
\name{hfdata}
\alias{hfdata}
\title{Hedge fund data}
\format{
A matrix of size 60 x 100, containing 60 monthly net returns for 100 hedge funds. Returns have been modified.
}
\source{
HFR database
}
\usage{
data(hfdata)
}
\description{
The matrix \code{hfdata} contains randomized and modified monthly returns of hedge funds.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.TESTING}
\alias{print.TESTING}
\title{Print method for the 'TESTING' object}
\usage{
\method{print}{TESTING}(x, ...)
}
\arguments{
\item{x}{A \code{TESTING} object.}

\item{...}{Further arguments passed to or from other methods.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print method for the \code{TESTING} object returned by
\code{\link{sharpeTesting}}, \code{\link{msharpeTesting}} and
\code{\link{alphaTesting}}.
}
\seealso{
\code{\link{sharpeTesting}}, \code{\link{msharpeTesting}} and
\code{\link{alphaTesting}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{plot.SCREENING}
\alias{plot.SCREENING}
\title{Peer performance screening plot}
\usage{
\method{plot}{SCREENING}(
  x,
  nblock = NULL,
  reference = NULL,
  band = 27.5,
  colorset = c(grDevices::gray(0), grDevices::gray(0.8), grDevices::gray(0.5)),
  ...
)
}
\arguments{
\item{x}{A \code{SCREENING} object.}

\item{nblock}{Optional number of equally-sized blocks into which the funds
(sorted by performance) are aggregated by averaging, as in Ardia and Boudt
(2018). Default: \code{nblock = NULL}, i.e. one bar per fund.}

\item{reference}{A logical value indicating whether the dashed
percentile-rank reference lines should be drawn. Default:
\code{reference = NULL}, i.e. drawn for within-group screening and omitted
for cross-group (\code{Y}-based) screening, where the within-group
percentile-rank benchmark does not apply. A single focal fund (cross-group
with one fund in \code{X}) is shown as a single stacked bar.}

\item{band}{Half-width (in percentage points) of the reference band drawn
around the central diagonal. Default: \code{band = 27.5}.}

\item{colorset}{Vector of three colors for the \eqn{\hat\pi^+},
\eqn{\hat\pi^0} and \eqn{\hat\pi^-} areas. Default:
\code{c(gray(0), gray(0.8), gray(0.5))}.}

\item{...}{Further graphical arguments passed to \code{\link[graphics]{barplot}}.}
}
\value{
Invisibly returns the (sorted, possibly aggregated) matrix of
ratios that is plotted.
}
\description{
Plot method for the \code{SCREENING} object returned by
\code{\link{alphaScreening}}, \code{\link{sharpeScreening}} and
\code{\link{msharpeScreening}}. It reproduces the peer performance
screening plot of Ardia and Boudt (2018): funds are sorted by their
performance measure and, for each fund, the estimated outperformance
(\eqn{\hat\pi^+}), equal-performance (\eqn{\hat\pi^0}) and underperformance
(\eqn{\hat\pi^-}) ratios are displayed as a horizontal stacked bar. The
dashed diagonal lines depict the naive percentile-rank benchmark
(\eqn{\hat\pi^0 = 0}); the gap between the black \eqn{\hat\pi^+} area and the
diagonal visualizes the luck correction.
}
\details{
If the screening was run with \code{screen_beta = TRUE}, the alpha
(first) coefficient is used.
}
\examples{
\donttest{
data("hfdata")
set.seed(1234)
sc <- alphaScreening(hfdata[, 1:30], control = list(nCore = 1))
plot(sc)
}
}
\references{
Ardia, D., Boudt, K. (2018).
The peer performance ratios of hedge funds.
\emph{Journal of Banking and Finance} \bold{87}, pp.351--368.
\doi{10.1016/j.jbankfin.2017.10.014}
}
\seealso{
\code{\link{alphaScreening}}, \code{\link{sharpeScreening}} and
\code{\link{msharpeScreening}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{exposureHeterogeneity}
\alias{exposureHeterogeneity}
\title{Factor exposure heterogeneity from a beta screening}
\usage{
exposureHeterogeneity(object)
}
\arguments{
\item{object}{A \code{SCREENING} object produced with
\code{screen_beta = TRUE}.}
}
\value{
A \code{data.frame} of class \code{exposureHeterogeneity} with columns
\code{coefficient}, \code{equalExposure} (\eqn{\pi^0_k}), and
\code{heterogeneity} (\eqn{1-\pi^0_k}).
}
\description{
Aggregates the per-fund equal-performance ratios of a
\code{screen_beta = TRUE} \code{\link{alphaScreening}} into the
factor-by-factor heterogeneity measure of Ardia et al. (2023). For each
coefficient \eqn{k} (the alpha and each factor beta) it reports the average
equal-exposure ratio \eqn{\pi^0_k = \frac1N\sum_i \pi^0_{i,k}} and the
heterogeneity \eqn{1-\pi^0_k}: the share of peers that are significantly
differentiated on coefficient \eqn{k}. A value close to one indicates large
heterogeneity (much room to differentiate); close to zero, homogeneity.
}
\examples{
\donttest{
data("hfdata")
set.seed(1234)
fac <- matrix(rnorm(nrow(hfdata) * 2), ncol = 2,
              dimnames = list(NULL, c("MKT", "SMB")))
sc <- alphaScreening(hfdata[, 1:20], factors = fac, screen_beta = TRUE,
                     control = list(nCore = 1))
exposureHeterogeneity(sc)
}
}
\references{
Ardia, D., Bluteau, K., Lortie-Cloutier, G., Tran, D. (2023).
Factor exposure heterogeneity in green and brown stocks.
\emph{Finance Research Letters} \bold{55}, Part A, 103900.
}
\seealso{
\code{\link{alphaScreening}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.SCREENING}
\alias{summary.SCREENING}
\title{Summary method for the 'SCREENING' object}
\usage{
\method{summary}{SCREENING}(object, coef = 1L, top = 5L, p_level = 0.05, ...)
}
\arguments{
\item{object}{A \code{SCREENING} object.}

\item{coef}{For a \code{screen_beta} screening, which coefficient row to use
(default 1, the alpha).}

\item{top}{Number of best/worst funds to display. Default: 5.}

\item{p_level}{Significance level for the win/loss counts (used only when the
object carries square \code{pval}/\code{tstat} matrices). Default: 0.05.}

\item{...}{Further arguments (ignored).}
}
\value{
An object of class \code{summary.SCREENING}.
}
\description{
Produces a ranked summary of a screening: distribution of the
performance measure, the peer performance ratios, optional win/loss counts
from the pairwise tests, and the top/bottom funds (overall and by
out-/underperformance ratio). The win/loss counts require the square
(within-group) pairwise matrices and are therefore omitted for cross-group
screenings (a \code{Y} peer group or \code{\link{targetPeerPerformance}}).
Originally contributed by Murilo Andre Peres Pereira.
}
\examples{
\donttest{
  data("hfdata")
  sc <- alphaScreening(hfdata[, 1:10], control = list(nCore = 1))
  summary(sc)
}
}
\seealso{
\code{\link{alphaScreening}}.
}
\author{
Murilo Andre Peres Pereira, David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{confint.SCREENING}
\alias{confint.SCREENING}
\title{Bootstrap confidence intervals for the peer performance ratios}
\usage{
\method{confint}{SCREENING}(
  object,
  parm = c("pipos", "pizero", "pineg"),
  level = 0.95,
  ...,
  nBoot = 499L,
  gammaPos = 0.4,
  gammaNeg = 0.6
)
}
\arguments{
\item{object}{A \code{SCREENING} object (within-group or cross-group; not
\code{screen_beta}).}

\item{parm}{Which ratio to bound: \code{"pipos"} (default), \code{"pizero"},
or \code{"pineg"}.}

\item{level}{Confidence level. Default: \code{0.95}.}

\item{...}{Unused.}

\item{nBoot}{Number of bootstrap resamples. Default: \code{499}.}

\item{gammaPos, gammaNeg}{One-sided thresholds used to recompute the ratios
(should match those of the original screening). Defaults: \code{0.4},
\code{0.6}.}
}
\value{
A matrix with one row per fund and two columns (lower and upper
bounds); the point estimate is attached as attribute \code{"estimate"}.
}
\description{
Computes confidence intervals for one of the peer performance
ratios (\eqn{\hat\pi^+}, \eqn{\hat\pi^0}, or \eqn{\hat\pi^-}) of a
\code{SCREENING} object, by a nonparametric peer (pairwise) bootstrap: for
each focal fund its peers are resampled with replacement and the ratio is
recomputed from the stored pairwise p-values and test statistics. The
interval therefore reflects the peer-composition and aggregation uncertainty,
holding the individual pairwise test statistics fixed.
}
\examples{
\donttest{
data("hfdata")
set.seed(1234)
sc <- alphaScreening(hfdata[, 1:15], control = list(nCore = 1))
confint(sc, parm = "pipos")
}
}
\seealso{
\code{\link{alphaScreening}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alphaTesting.R
\name{alphaTesting}
\alias{alphaTesting}
\title{Testing the difference of alpha outperformance ratios}
\usage{
alphaTesting(x, y, factors = NULL, control = list(), screen_beta = NULL)
}
\arguments{
\item{x}{Vector (of length \eqn{T}) of returns for the first fund. \code{NA}
values are allowed.}

\item{y}{Vector (of length \eqn{T}) returns for the second fund. \code{NA}
values are allowed.}

\item{factors}{Matrix \eqn{(T \times K)}{(TxK)} of \eqn{T} returns for the
\eqn{K} factors. \code{NA} values are allowed.}

\item{control}{Control parameters (see *Details*).}

\item{screen_beta}{Boolean to screen all factors' coefficients (beta).
Default: \code{screen_beta = NULL}, in which case the value is taken from
\code{control$screen_beta} (itself defaulting to \code{FALSE}). When supplied
directly, the argument takes precedence. \code{screen_beta = TRUE} requires
\code{factors}.}
}
\value{
A list with the following components:\cr

\code{n}: Number of non-\code{NA} concordant observations.\cr

\code{alpha}: Vector (of length 2) of the two funds' unconditional alphas.\cr

\code{dalpha}: Difference of the two funds' alphas.\cr

\code{tstat}: t-stat of the alpha difference.\cr

\code{pval}: p-value of the test of equal alpha.\cr

When \code{screen_beta = TRUE}, \code{alpha} is a matrix (one row per
coefficient, the first being the alpha) and \code{dalpha}, \code{tstat} and
\code{pval} are vectors with one entry per coefficient; the \code{print}
method reports the alpha (first) coefficient only.
}
\description{
Function which performs the testing of the difference of alpha outperformance ratios.
}
\details{
The alpha measure (Treynor and Black 1973, Carhart 1997, Fung and Hsieh
2004) is one industry standard for measuring the absolute risk adjusted
performance of hedge funds. This function performs
the testing of alpha outperformance ratio difference for two funds.

For the testing, only the intersection of non-\code{NA} observations for the
two funds are used.

The argument \code{control} is a list that can supply any of the following
components:
\itemize{
\item \code{'hac'} Heteroscedastic-autocorrelation consistent
standard errors. Default: \code{hac = FALSE}.
}
}
\note{
Further details on the methodology with an application to the hedge
fund industry is given in Ardia and Boudt (2018).

Some internal functions where adapted from Michael Wolf MATLAB code.
}
\examples{
## Load the data (randomized data of monthly hedge fund returns)
data("hfdata")
x = hfdata[,1]
y = hfdata[,2]

## Run alpha testing
alphaTesting(x, y)
}
\references{
Ardia, D., Boudt, K. (2015).
Testing equality of modified Sharpe ratios.
\emph{Finance Research Letters} \bold{13}, 97--104.

Ardia, D., Boudt, K. (2018).
The peer performance ratios of hedge funds.
\emph{Journal of Banking and Finance} \bold{87}, 351--368.

Barras, L., Scaillet, O., Wermers, R. (2010).
False discoveries in mutual fund performance: Measuring luck in estimated alphas.
\emph{Journal of Finance} \bold{65}(1), 179--216.

Sharpe, W.F. (1994).
The Sharpe ratio.
\emph{Journal of Portfolio Management} \bold{21}(1), 49--58.

Ledoit, O., Wolf, M. (2008).
Robust performance hypothesis testing with the Sharpe ratio.
\emph{Journal of Empirical Finance} \bold{15}(5), 850--859.

Storey, J. (2002).
A direct approach to false discovery rates.
\emph{Journal of the Royal Statistical Society B} \bold{64}(3), 479--498.
}
\seealso{
\code{\link{alphaScreening}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.SCREENING}
\alias{print.SCREENING}
\title{Print method for the 'SCREENING' object}
\usage{
\method{print}{SCREENING}(x, ...)
}
\arguments{
\item{x}{A \code{SCREENING} object.}

\item{...}{Further arguments passed to or from other methods.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print method for the \code{SCREENING} object returned by
\code{\link{alphaScreening}}, \code{\link{sharpeScreening}} and
\code{\link{msharpeScreening}}.
}
\seealso{
\code{\link{alphaScreening}}, \code{\link{sharpeScreening}},
\code{\link{msharpeScreening}} and \code{\link{plot.SCREENING}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rollScreening.R
\name{rollScreening}
\alias{rollScreening}
\title{Rolling-window peer performance screening}
\usage{
rollScreening(
  X,
  factors = NULL,
  Y = NULL,
  screen = c("alpha", "sharpe", "msharpe"),
  width = 36L,
  by = 1L,
  level = 0.9,
  na.neg = TRUE,
  control = list(),
  dates = NULL
)
}
\arguments{
\item{X}{Matrix \eqn{(T \times N)}{(TxN)} of \eqn{T} returns for the \eqn{N}
funds.}

\item{factors}{Matrix \eqn{(T \times K)}{(TxK)} of factor returns (for
\code{screen = "alpha"}). Default: \code{NULL}.}

\item{Y}{Optional matrix \eqn{(T \times M)}{(TxM)} of returns for a peer group
(cross-group screening, see \code{\link{alphaScreening}}). Default:
\code{NULL}.}

\item{screen}{Performance measure to screen on: \code{"alpha"} (default),
\code{"sharpe"}, or \code{"msharpe"}.}

\item{width}{Window length (number of observations). Default: \code{36}.}

\item{by}{Step (in observations) between successive windows. Default:
\code{1}.}

\item{level}{Modified Value-at-Risk level (for \code{screen = "msharpe"}).
Default: \code{0.9}.}

\item{na.neg}{A logical value passed to \code{\link{msharpeScreening}}
(for \code{screen = "msharpe"}) indicating whether a negative modified
Value-at-Risk yields \code{NA}. Default: \code{TRUE}.}

\item{control}{Control parameters passed to the screening function (see
\code{\link{alphaScreening}}); set \code{control = list(screen_beta = TRUE)}
to obtain per-factor heterogeneity series. With \code{nCore > 1} each
window spins up its own PSOCK cluster, which is wasteful for many small
windows; the default \code{nCore = 1} runs serially without any cluster.}

\item{dates}{Optional vector of length \eqn{T} (e.g. \code{Date}) used to
label the windows by their end date. Default: \code{NULL}.}
}
\value{
A \code{data.frame} of class \code{rollScreening} with one row per
window (or per window/coefficient when \code{screen_beta = TRUE}), with
columns \code{window}, \code{index} (end-of-window row), optionally
\code{date}, \code{coefficient}, the averaged \code{pizero}/\code{pipos}/
\code{pineg}, and \code{heterogeneity} (\eqn{1-\bar\pi^0}).
}
\description{
Runs a peer performance screening on a rolling time window and
returns the time series of cross-sectionally averaged peer performance
ratios. This is the design underlying the dynamic analyses of Ardia et al.
(2022, 2023): on each window the screening is computed and the ratios are
averaged across funds (per coefficient when \code{screen_beta = TRUE}),
yielding a series of (factor exposure) heterogeneity measures.
}
\examples{
\donttest{
data("hfdata")
set.seed(1234)
## a subset keeps the example quick; drop the column index to screen the
## whole universe
roll <- rollScreening(hfdata[, 1:20], screen = "alpha", width = 36, by = 6,
                      control = list(nCore = 1))
plot(roll)
}
}
\references{
Ardia, D., Bluteau, K., Tran, D. (2022).
How easy is it for investment managers to deploy their talent in green and brown stocks?
\emph{Finance Research Letters} \bold{48}, 102992.

Ardia, D., Bluteau, K., Lortie-Cloutier, G., Tran, D. (2023).
Factor exposure heterogeneity in green and brown stocks.
\emph{Finance Research Letters} \bold{55}, Part A, 103900.
}
\seealso{
\code{\link{alphaScreening}}, \code{\link{exposureHeterogeneity}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sharpeTesting.R
\name{sharpeTesting}
\alias{sharpeTesting}
\title{Testing the difference of Sharpe ratios}
\usage{
sharpeTesting(x, y, control = list())
}
\arguments{
\item{x}{Vector (of length \eqn{T}) of returns for the first fund. \code{NA}
values are allowed.}

\item{y}{Vector (of length \eqn{T}) returns for the second fund. \code{NA}
values are allowed.}

\item{control}{Control parameters (see *Details*).}
}
\value{
A list with the following components:\cr

\code{n}: Number of non-\code{NA} concordant observations.\cr

\code{sharpe}: Vector (of length 2) of unconditional Sharpe ratios.\cr

\code{dsharpe}: Sharpe ratios difference.\cr

\code{tstat}: t-stat of Sharpe ratios differences.\cr

\code{pval}: pvalues of test of Sharpe ratios differences.
}
\description{
Function which performs the testing of the difference of Sharpe ratios.
}
\details{
The Sharpe ratio (Sharpe 1994) is one industry standard for measuring the
absolute risk adjusted performance of hedge funds. This function performs
the testing of Sharpe ratio difference for two funds using the approach by
Ledoit and Wolf (2002).

For the testing, only the intersection of non-\code{NA} observations for the
two funds are used.

The argument \code{control} is a list that can supply any of the following
components:
\itemize{
\item \code{'type'} Asymptotic approach (\code{type = 1}) or
studentized circular bootstrap approach (\code{type = 2}). Default:
\code{type = 1}.
\item \code{'ttype'} Test based on ratio (\code{type = 1})
or product (\code{type = 2}). Default: \code{type = 2}.
\item \code{'hac'} Heteroscedastic-autocorrelation consistent standard
errors. Default: \code{hac = FALSE}.
\item \code{'nBoot'} Number of bootstrap replications for computing the p-value. Default: \code{nBoot =
499}.
\item \code{'bBoot'} Block length in the circular bootstrap. Default:
\code{bBoot = 1}, i.e. iid bootstrap. \code{bBoot = 0} uses optimal
block-length.
\item \code{'pBoot'} Symmetric p-value (\code{pBoot = 1}) or
asymmetric p-value (\code{pBoot = 2}). Default: \code{pBoot = 1}.
}
}
\note{
Further details on the methodology with an application to the hedge
fund industry is given in Ardia and Boudt (2018).

Some internal functions where adapted from Michael Wolf MATLAB code.
}
\examples{
## Load the data (randomized data of monthly hedge fund returns)
data("hfdata")
x = hfdata[,1]
y = hfdata[,2]

## Run Sharpe testing (asymptotic)
ctr = list(type = 1)
out = sharpeTesting(x, y, control = ctr)
print(out)

## Run Sharpe testing (asymptotic hac)
ctr = list(type = 1, hac = TRUE)
out = sharpeTesting(x, y, control = ctr)
print(out)

## Run Sharpe testing (iid bootstrap)
set.seed(1234)
ctr = list(type = 2, nBoot = 100)
out = sharpeTesting(x, y, control = ctr)
print(out)

## Run Sharpe testing (circular bootstrap)
set.seed(1234)
ctr = list(type = 2, nBoot = 100, bBoot = 5)
out = sharpeTesting(x, y, control = ctr)
print(out)
}
\references{
Ardia, D., Boudt, K. (2015).
Testing equality of modified Sharpe ratios.
\emph{Finance Research Letters} \bold{13}, 97--104.

Ardia, D., Boudt, K. (2018).
The peer performance ratios of hedge funds.
\emph{Journal of Banking and Finance} \bold{87}, 351--368.

Barras, L., Scaillet, O., Wermers, R. (2010).
False discoveries in mutual fund performance: Measuring luck in estimated alphas.
\emph{Journal of Finance} \bold{65}(1), 179--216.

Sharpe, W.F. (1994).
The Sharpe ratio.
\emph{Journal of Portfolio Management} \bold{21}(1), 49--58.

Ledoit, O., Wolf, M. (2008).
Robust performance hypothesis testing with the Sharpe ratio.
\emph{Journal of Empirical Finance} \bold{15}(5), 850--859.

Storey, J. (2002).
A direct approach to false discovery rates.
\emph{Journal of the Royal Statistical Society B} \bold{64}(3), 479--498.
}
\seealso{
\code{\link{sharpe}}, \code{\link{sharpeScreening}} and
\code{\link{msharpeTesting}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{plot.exposureHeterogeneity}
\alias{plot.exposureHeterogeneity}
\title{Plot factor exposure heterogeneity}
\usage{
\method{plot}{exposureHeterogeneity}(x, ...)
}
\arguments{
\item{x}{An \code{exposureHeterogeneity} object.}

\item{...}{Further arguments passed to \code{\link[graphics]{barplot}}.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Bar plot of the heterogeneity measure \eqn{1-\pi^0_k} returned by
\code{\link{exposureHeterogeneity}}.
}
\examples{
\donttest{
  data("hfdata")
  sc <- alphaScreening(hfdata[, 1:10], factors = hfdata[, 50:51],
                       control = list(nCore = 1, screen_beta = TRUE))
  plot(exposureHeterogeneity(sc))
}
}
\seealso{
\code{\link{exposureHeterogeneity}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/msharpeScreening.R
\name{msharpeScreening}
\alias{msharpeScreening}
\title{Screening using the modified Sharpe outperformance ratio}
\usage{
msharpeScreening(X, level = 0.9, na.neg = TRUE, control = list(), Y = NULL)
}
\arguments{
\item{X}{Matrix \eqn{(T \times N)}{(TxN)} of \eqn{T} returns for the \eqn{N}
funds. \code{NA} values are allowed.}

\item{level}{Modified Value-at-Risk level. Default: \code{level = 0.90}.}

\item{na.neg}{A logical value indicating whether \code{NA} values should be
returned if a negative modified Value-at-Risk is obtained.  Default
\code{na.neg = TRUE}.}

\item{control}{Control parameters (see *Details*).}

\item{Y}{Optional matrix \eqn{(T \times M)}{(TxM)} of returns for a second
(peer) group of \eqn{M} funds. When supplied, the ratios are computed for
each fund in \code{X} against the funds in \code{Y} (cross-group screening)
instead of against the other funds in \code{X}; a single focal fund versus a
peer group corresponds to \code{X} being a vector. Columns of \code{Y}
identical to the focal fund are automatically excluded. Default:
\code{Y = NULL}, i.e. within-group screening.}
}
\value{
A list with the following components:\cr

\code{n}: Vector (of length \eqn{N}) of number of non-\code{NA}
observations.\cr

\code{npeer}: Vector (of length \eqn{N}) of number of available peers.\cr

\code{msharpe}: Vector (of length \eqn{N}) of unconditional modified Sharpe
ratios.\cr

\code{dmsharpe}: Matrix (of size \eqn{N \times N}{NxN}) of modified Sharpe
ratios differences.\cr

\code{tstat}: Matrix (of size \eqn{N \times N}{NxN}) of t-statistics.\cr

\code{pval}: Matrix (of size \eqn{N \times N}{NxN}) of p-values of test for
modified Sharpe ratios differences.\cr

\code{lambda}: Vector (of length \eqn{N}) of lambda values.\cr

\code{pizero}: Vector (of length \eqn{N}) of probability of equal
performance.\cr

\code{pipos}: Vector (of length \eqn{N}) of probability of outperformance
performance.\cr

\code{pineg}: Vector (of length \eqn{N}) of probability of underperformance
performance.
}
\description{
Function which performs the screening of a universe of returns, and
computes the modified Sharpe outperformance ratio.
}
\details{
The modified Sharpe ratio (Favre and Galeano 2002, Gregoriou and Gueyie
2003) is one industry standard for measuring the absolute risk adjusted
performance of hedge funds. We propose to complement the modified Sharpe
ratio with the fund's outperformance ratio, defined as the percentage number
of funds that have a significantly lower modified Sharpe ratio. In a
pairwise testing framework, a fund can have a significantly higher modified
Sharpe ratio because of luck. We correct for this by applying the false
discovery rate approach by Storey (2002).

For the testing, only the intersection of non-\code{NA} observations for the
two funds are used.

The argument \code{control} is a list that can supply any of the following
components:
\itemize{
\item \code{'type'} Asymptotic approach (\code{type = 1}) or
studentized circular bootstrap approach (\code{type = 2}). Default:
\code{type = 1}.
\item \code{'ttype'} Test based on ratio (\code{type = 1})
or product (\code{type = 2}). Default: \code{type = 2}.
\item \code{'hac'} heteroscedastic-autocorrelation consistent standard
errors. Default: \code{hac = FALSE}.
\item \code{'nBoot'} Number of
bootstrap replications for computing the p-value. Default: \code{nBoot =
499}.
\item \code{'bBoot'} Block length in the circular bootstrap. Default:
\code{bBoot = 1}, i.e. iid bootstrap. (The data-driven choice
\code{bBoot = 0} is only available in \code{\link{msharpeTesting}}, not in
screening.)
\item \code{'pBoot'} Symmetric p-value (\code{pBoot = 1}) or
asymmetric p-value (\code{pBoot = 2}). Default: \code{pBoot = 1}.
\item \code{'nCore'} Number of cores to be used. Default: \code{nCore = 1}.
\item \code{'minObs'} Minimum number of concordant observations to compute
the ratios. Default: \code{minObs = 10}.
\item \code{'minObsPi'} Minimum number of observations to compute pi0. Default: \code{minObsPi = 1}.
\item \code{'lambda'} Threshold value to compute pi0. Default: \code{lambda
= NULL}, i.e. data driven choice.
\item \code{'gammaPos'} One-sided quantile level (of the standard Normal
distribution) used as the critical value for counting outperformed peers:
a peer counts as outperformed when the pairwise t-statistic exceeds
\code{qnorm(gammaPos)} (a \emph{negative} threshold for
\code{gammaPos < 0.5}), and the expected fraction \code{1 - gammaPos} of
false positives among the equal-performing peers is then subtracted.
Default: \code{gammaPos = 0.4} (the value recommended in Ardia and Boudt, 2018).
\item \code{'gammaNeg'} Mirror image of \code{gammaPos} for the peers that
outperform the focal fund: the count uses \code{tstat <= qnorm(gammaNeg)}
and subtracts the expected fraction \code{gammaNeg} of false positives.
Default: \code{gammaNeg = 0.6}.
\item \code{'fastAdjust'} Use a fast vectorised inversion in the
truncated-normal bias correction of \eqn{\pi^0}{pi0} instead of one
\code{uniroot} call per value. This is the dominant cost when \code{lambda}
is data driven and gives a large speed-up on big universes. The bisection
locates the root to about 1e-12; since \code{uniroot} stops at its own
tolerance (about 1.2e-4), the two paths typically differ by a few 1e-5, the
fast path being the more accurate. Default: \code{fastAdjust = FALSE},
i.e. the original code path, kept as default so that published results
reproduce exactly.
}
}
\note{
Further details on the methodology with an application to the hedge
fund industry is given in Ardia and Boudt (2018).

Some internal functions where adapted from Michael Wolf MATLAB code.

Application of the false discovery rate approach applied to the mutual fund
industry has been presented in Barras, Scaillet and Wermers (2010).
}
\examples{
## Load the data (randomized data of monthly hedge fund returns)
data("hfdata")
rets = hfdata[,1:4]

## Modified Sharpe screening
msharpeScreening(rets, control = list(nCore = 1))

## Modified Sharpe screening with the studentized circular bootstrap
## (the 'hac' flag is ignored when type = 2)
msharpeScreening(rets, control = list(nCore = 1, type = 2))
}
\references{
Ardia, D., Boudt, K. (2015).
Testing equality of modified Sharpe ratios.
\emph{Finance Research Letters} \bold{13}, 97--104.

Ardia, D., Boudt, K. (2018).
The peer performance ratios of hedge funds.
\emph{Journal of Banking and Finance} \bold{87}, 351--368.

Barras, L., Scaillet, O., Wermers, R. (2010).
False discoveries in mutual fund performance: Measuring luck in estimated alphas.
\emph{Journal of Finance} \bold{65}(1), 179--216.

Favre, L., Galeano, J.A. (2002).
Mean-modified Value-at-Risk Optimization with Hedge Funds.
\emph{Journal of Alternative Investments} \bold{5}(2), 21--25.

Gregoriou, G. N., Gueyie, J.-P. (2003).
Risk-adjusted performance of funds of hedge funds using a modified Sharpe ratio.
\emph{Journal of Wealth Management} \bold{6}(3), 77--83.

Ledoit, O., Wolf, M. (2008).
Robust performance hypothesis testing with the Sharpe ratio.
\emph{Journal of Empirical Finance} \bold{15}(5), 850--859.

Storey, J. (2002).
A direct approach to false discovery rates.
\emph{Journal of the Royal Statistical Society B} \bold{64}(3), 479--498.
}
\seealso{
\code{\link{msharpe}}, \code{\link{msharpeTesting}},
\code{\link{sharpeScreening}} and \code{\link{alphaScreening}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{sharpe}
\alias{sharpe}
\title{Compute Sharpe ratio}
\usage{
sharpe(X, na.rm = TRUE)
}
\arguments{
\item{X}{Vector (of length \eqn{T}) or matrix (of size \eqn{T \times
N}{TxN}) of returns for \eqn{N} funds. \code{NA} values are allowed.}

\item{na.rm}{A logical value indicating whether \code{NA} values should be
stripped before the computation. Default \code{na.rm = TRUE}}
}
\value{
A scalar or a vector (of size \eqn{N}) with the Sharpe ratios.
}
\description{
Function which computes the Sharpe ratio.
}
\details{
The Sharpe ratio (Sharpe 1994) is one industry standard for measuring the
absolute risk adjusted performance of hedge funds.
}
\examples{
## Load the data
data('hfdata')

## Compute the Sharpe ratio
out = sharpe(hfdata)
print(out)

out = sharpe(hfdata, na.rm = FALSE)
print(out)
}
\references{
Ardia, D., Boudt, K. (2015).
Testing equality of modified Sharpe ratios.
\emph{Finance Research Letters} \bold{13}, pp.97--104.
\doi{10.1016/j.frl.2015.02.008}

Ardia, D., Boudt, K. (2016).
The Peer Ratios Performance of Hedge Funds.
\emph{Working paper}.
\doi{10.2139/ssrn.2000901}

Sharpe, W.F. (1994).
The Sharpe ratio.
\emph{Journal of Portfolio Management} \bold{21}(1), pp.49--58.
\doi{10.3905/jpm.1994.409501}
}
\seealso{
\code{\link{sharpeTesting}}, \code{\link{sharpeScreening}} and
\code{\link{msharpe}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sharpeScreening.R
\name{sharpeScreening}
\alias{sharpeScreening}
\title{Screening using the Sharpe outperformance ratio}
\usage{
sharpeScreening(X, control = list(), Y = NULL)
}
\arguments{
\item{X}{Matrix \eqn{(T \times N)}{(TxN)} of \eqn{T} returns for the \eqn{N}
funds. \code{NA} values are allowed.}

\item{control}{Control parameters (see *Details*).}

\item{Y}{Optional matrix \eqn{(T \times M)}{(TxM)} of returns for a second
(peer) group of \eqn{M} funds. When supplied, the ratios are computed for
each fund in \code{X} against the funds in \code{Y} (cross-group screening)
instead of against the other funds in \code{X}; a single focal fund versus a
peer group corresponds to \code{X} being a vector. Columns of \code{Y}
identical to the focal fund are automatically excluded. Default:
\code{Y = NULL}, i.e. within-group screening.}
}
\value{
A list with the following components:\cr

\code{n}: Vector (of length \eqn{N}) of number of non-\code{NA}
observations.\cr

\code{npeer}: Vector (of length \eqn{N}) of number of available peers.\cr

\code{sharpe}: Vector (of length \eqn{N}) of unconditional Sharpe ratios.\cr

\code{dsharpe}: Matrix (of size \eqn{N \times N}{NxN}) of Sharpe ratios
differences.\cr

\code{tstat}: Matrix (of size \eqn{N \times N}{NxN}) of t-statistics.\cr

\code{pval}: Matrix (of size \eqn{N \times N}{NxN}) of pvalues of test for Sharpe
ratios differences.\cr

\code{lambda}: vector (of length \eqn{N}) of lambda values.\cr

\code{pizero}: vector (of length \eqn{N}) of probability of equal
performance.\cr

\code{pipos}: vector (of length \eqn{N}) of probability of outperformance
performance.\cr

\code{pineg}: Vector (of length \eqn{N}) of probability of underperformance
performance.
}
\description{
Function which performs the screening of a universe of returns, and
computes the Sharpe outperformance ratio.
}
\details{
The Sharpe ratio (Sharpe 1994) is one industry standard for measuring the
absolute risk adjusted performance of hedge funds. We propose to complement
the Sharpe ratio with the fund's outperformance ratio, defined as the
percentage number of funds that have a significantly lower Sharpe ratio. In
a pairwise testing framework, a fund can have a significantly higher Sharpe
ratio because of luck. We correct for this by applying the false discovery
rate approach by Storey (2002).

For the testing, only the intersection of non-\code{NA} observations for the
two funds are used.

The methodology proceeds as follows:
\itemize{
\item (1) compute all
pairwise tests of Sharpe differences using the bootstrap approach of Ledoit
and Wolf (2002). This means that for a universe of \eqn{N} funds, we perform
\eqn{N(N-1)/2}{N*(N-1)/2} tests. The algorithm has been parallelized and the
computational burden can be split across several cores. The number of cores
can be defined in \code{control}, see below.
\item (2) for each fund, the
false discovery rate approach by Storey (2002) is used to determine the
proportions over, equal, and underperforming funds, in terms of Sharpe ratio,
in the database.
}
The argument \code{control} is a list that can supply any of the following
components:
\itemize{
\item \code{'type'} Asymptotic approach (\code{type = 1}) or
studentized circular bootstrap approach (\code{type = 2}). Default:
\code{type = 1}.
\item \code{'ttype'} Test based on ratio (\code{type = 1})
or product (\code{type = 2}). Default: \code{type = 2}.
\item \code{'hac'} Heteroscedastic-autocorrelation consistent standard
errors. Default: \code{hac = FALSE}.
\item \code{'nBoot'} Number of bootstrap replications for computing the p-value. Default: \code{nBoot =
499}.
\item \code{'bBoot'} Block length in the circular bootstrap. Default:
\code{bBoot = 1}, i.e. iid bootstrap. (The data-driven choice
\code{bBoot = 0} is only available in \code{\link{sharpeTesting}}, not in
screening.)
\item \code{'pBoot'} Symmetric p-value (\code{pBoot = 1}) or
asymmetric p-value (\code{pBoot = 2}). Default: \code{pBoot = 1}.
\item \code{'nCore'} Number of cores to be used. Default: \code{nCore = 1}.
\item \code{'minObs'} Minimum number of concordant observations to compute
the ratios. Default: \code{minObs = 10}.
\item \code{'minObsPi'} Minimum
number of observations to compute pi0. Default: \code{minObsPi = 1}.
\item \code{'lambda'} Threshold value to compute pi0. Default: \code{lambda
= NULL}, i.e. data driven choice.
\item \code{'gammaPos'} One-sided quantile level (of the standard Normal
distribution) used as the critical value for counting outperformed peers:
a peer counts as outperformed when the pairwise t-statistic exceeds
\code{qnorm(gammaPos)} (a \emph{negative} threshold for
\code{gammaPos < 0.5}), and the expected fraction \code{1 - gammaPos} of
false positives among the equal-performing peers is then subtracted.
Default: \code{gammaPos = 0.4} (the value recommended in Ardia and Boudt, 2018).
\item \code{'gammaNeg'} Mirror image of \code{gammaPos} for the peers that
outperform the focal fund: the count uses \code{tstat <= qnorm(gammaNeg)}
and subtracts the expected fraction \code{gammaNeg} of false positives.
Default: \code{gammaNeg = 0.6}.
\item \code{'fastAdjust'} Use a fast vectorised inversion in the
truncated-normal bias correction of \eqn{\pi^0}{pi0} instead of one
\code{uniroot} call per value. This is the dominant cost when \code{lambda}
is data driven and gives a large speed-up on big universes. The bisection
locates the root to about 1e-12; since \code{uniroot} stops at its own
tolerance (about 1.2e-4), the two paths typically differ by a few 1e-5, the
fast path being the more accurate. Default: \code{fastAdjust = FALSE},
i.e. the original code path, kept as default so that published results
reproduce exactly.
}
}
\note{
Further details on the methodology with an application to the hedge
fund industry is given in Ardia and Boudt (2018).

Some internal functions where adapted from Michael Wolf MATLAB code.

Application of the false discovery rate approach applied to the mutual fund
industry has been presented in Barras, Scaillet and Wermers (2010).
}
\examples{
## Load the data (randomized data of monthly hedge fund returns)
data("hfdata")
rets = hfdata[,1:4]

## Sharpe screening
sharpeScreening(rets, control = list(nCore = 1))

## Sharpe screening with HAC standard errors (asymptotic test)
sharpeScreening(rets, control = list(nCore = 1, hac = TRUE))

## Sharpe screening with the studentized circular bootstrap
## (the 'hac' flag is ignored when type = 2)
sharpeScreening(rets, control = list(nCore = 1, type = 2))
}
\references{
Ardia, D., Boudt, K. (2015).
Testing equality of modified Sharpe ratios.
\emph{Finance Research Letters} \bold{13}, 97--104.

Ardia, D., Boudt, K. (2018).
The peer performance ratios of hedge funds.
\emph{Journal of Banking and Finance} \bold{87}, 351--368.

Barras, L., Scaillet, O., Wermers, R. (2010).
False discoveries in mutual fund performance: Measuring luck in estimated alphas.
\emph{Journal of Finance} \bold{65}(1), 179--216.

Sharpe, W.F. (1994).
The Sharpe ratio.
\emph{Journal of Portfolio Management} \bold{21}(1), 49--58.

Ledoit, O., Wolf, M. (2008).
Robust performance hypothesis testing with the Sharpe ratio.
\emph{Journal of Empirical Finance} \bold{15}(5), 850--859.

Storey, J. (2002).
A direct approach to false discovery rates.
\emph{Journal of the Royal Statistical Society B} \bold{64}(3), 479--498.
}
\seealso{
\code{\link{sharpe}}, \code{\link{sharpeTesting}},
\code{\link{msharpeScreening}} and \code{\link{alphaScreening}}.
}
\author{
David Ardia and Kris Boudt.
}
\keyword{htest}
